;;; -*- Mode: Lisp; -*-
;;; YstokHelp - HTML-based help
;;; Copyright (c) 2009-2014 Dr. Dmitry Ivanov.  All rights reserved.
;;; Website: http://lisp.ystok.ru/yhelp/
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Anchor functions
#||
Anchors are defined with the following HTML elements and attributes:
  <A name="anchor_name"></A>
  <DFN id="anchor_name"></DFN>
  <DT id="anchor_name"></DT>
  <Hn id="anchor_name"></Hn>	(ignored by Index compiler by default)

HTML 4.0 Specification
----------------------
Attribute ID is applicable directly to many elements,
serve the same purpose as the A element, and its value shares the same namespace.
The names specified with id="name" for other elements are not treated as anchors
neither by Preprocessor nor Index compiler.

Section 6.2 SGML basic types
- ID and NAME tokens must begin with a letter ([A-Za-z]) and may be followed by
  any number of letters, digits ([0-9]), hyphens ("-"), underscores ("_"),
  colons (":"), and periods (".").

XHTML 1.0 Specification
-----------------------
Id/name atrubute is of Name extending syntax defined by XML:
[4] NameStartChar ::= ":" | [A-Z] | "_" | [a-z]
	| [#xC0-#xD6] | [#xD8-#xF6] | [#xF8-#x2FF]
	| [#x370-#x37D] | [#x37F-#x1FFF] | [#x200C-#x200D] | [#x2070-#x218F]
	| [#x2C00-#x2FEF] | [#x3001-#xD7FF] | [#xF900-#xFDCF] | [#xFDF0-#xFFFD]
	| [#x10000-#xEFFFF]
[4a] NameChar ::= NameStartChar | "-" | "." | [0-9]
	| #xB7 | [#x0300-#x036F] | [#x203F-#x2040] 
[5]  Name ::= NameStartChar (NameChar)* 

Do not use space (#x32), non-break-space, and other whitespace characters.
National characters are allowed in anchors but should be escaped in URI fragment.

RUSSIAN NB:
   UCS-2 -: #x0410-04FF
   Windows-1251    ( .latin-1 ):
	 #xA8	latin-1: Diaeresis
	 #xB8	latin-1: Cedilla
	 #xD7	latin-1: multiplication
	 #xF7	latin-1: division
||#

(in-package :ystok.help)

(defun elem-word-list (elem)
 ;;; Get rid of imbedded lists: tree -> string list
  ;; NB: Is not applicable when there is any template pseudo-tag within the elem content!
  ;; Q: Maybe rename this to elem-text-chunks and do not split on whitespace?
  ;;    What about newlines? Due to :preserve-whitespace :newline, they must have been
  ;;    replaced by :newline already?
  (cond ((stringp elem)
         (split-seq-if #'whitespace-char-p elem))         ;(list elem)
        ((consp elem)
         (mapcan #'elem-word-list (html:elem-cont elem)))))

(defun lhtml-word-list (content)
  (mapcan #'elem-word-list content))

(defun distil-content (cont)
 ;;; Transform LHTML tree: eliminate "excessive" tags and lift their content.
  ;; Args: cont  Anchor element children
  ;; Values: 1) New content as a LHTML list.
  ;;         2) Plist with indicators from *anchor-span-css-class-properties*.
  ;; NB: Whitespace should be already removed by parse-html.
  (let ((plist ()))
    (labels
      ((%transform (cont)
         (let ((acc ()))
           (dolist (elem cont)
             (cond ((stringp elem)
                    ;(nconcf word-list (split-seq-if #'whitespace-char-p elem))
                    (push elem acc))
                   ((atom elem)
                    (when (and elem (not (memq elem '(:br :newline))))
                      (push elem acc)))
                   (t
                    (let ((key (html:elem-key elem))
                          (attrs (html:elem-attrs elem))
                          (cont (html:elem-cont elem)))
                      (if (or (memq key *anchor-inline-tags*)
                              (case key
                                (:span				; text class
                                 ;; Q:    ,   ?
                                 (or (let ((class (getf attrs :class)))
                                       (when (stringp class)
                                         (let ((classes (split-seq-if #'whitespace-char-p
                                                                      class)))
                                           (cond ((intersection classes
                                                                *anchor-span-css-classes*
                                                                :test #'string-equal))
                                                 ((setq class 
                                                        (first (intersection classes
                                                         *anchor-span-css-class-properties*
                                                         :test #'string-equal)))
                                                  (setq plist (list* (if (keywordp class)
                                                                         class
                                                                         (intern-upcase
                                                                          class))
                                                                     cont
                                                                     plist)
                                                        cont nil))))))
                                     (let ((style (getf attrs :style)))
                                       (when (stringp style)		; test style
                                         (get-properties
                                          (phtml:parse-declarations style)
                                          *anchor-span-style-properties*)))))
                                )   )
                          ;; Make copy of the child element and retain it
                          (when-let (cont (%transform cont))
                            (push (html:make-elem key attrs cont) acc))
                          ;; Drop the child element but promote its content one level up
                          (setq acc (nreconc (%transform cont) acc))))) ))
           (nreverse acc))))
      (let ((content (%transform cont)))
        (values content plist)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  PREPROCESS  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun clear-anchors (parsed-html-file)
  (let ((hash-table (anchors (asd:component-system parsed-html-file))))
    (dolist (anchor (anchors parsed-html-file))
      (when-let (aplist (gethash anchor hash-table))
        (setf (gethash anchor hash-table)
              (remove parsed-html-file aplist
                      :key (load-time-value (compose #'source-component #'first))))))
    (setf (anchors parsed-html-file) ())))

(defun register-anchor (anchor elem
                        &optional (attrs (html:elem-attrs elem))
                                  (index t)
                                  (plist (phtml:parse-declarations (getf attrs :help))))
 ;;; Put the anchor into the hash-table for the current book
  ;; Args: index ::= NIL      None
  ;;		  | :keywords Index only explicit keywords but do not content text
  ;;		  | :text     Index either keywords if present or the :text property.
  ;;		  | T         Index either keywords or text if present or content.
  ;; NB: The :text property in the help attribute specifies the title coming to TOC,
  ;;     menu item, or index entry (in case there is no :keywords).
  ;;     If :text is not specified explicitly, extract in from content.
  ;;     The value of :text property of the anchor should not differ from
  ;;     the value of :text for a heading!
  ;(unless (getf plist :text)
    ;(when-let* ((word-list (elem-word-list elem))		 	; not empty
    ;            (text (interleave-seq 'string #\Space word-list)))	; canonicalize
    ;  (setf plist (list* :text text plist))
  (let* ((class (or (getf plist :class) (getf attrs :class)))
         (classes (if (stringp class) (split-seq-if #'whitespace-char-p class) ()))
         (keywords (getf plist :keywords *unbound-value*)))
    (when (or (null index)
              (and (eq index :keywords)				; no explicit keywords
                   (eq keywords *unbound-value*)))
      (setf (getf plist :keywords) nil))
    (when (and (getf plist :class) classes)
      (setq plist (list* :classes classes plist)))
    (unless (getf plist :text)
      (multiple-value-bind (content new-props) (distil-content (html:elem-cont elem))
        (setq plist (list* :content content (if new-props	; implicit props to end
                                                (append plist new-props)
                                                plist)))
        ;; Save parsed CSS class list into the anchor's :classes property only if 
        ;; neither keywords nor text nor explicit classes property is specified
        (when (and classes (eq keywords *unbound-value*) (not (getf plist :class)))
          (when-let (classes (intersection classes *anchor-span-css-classes*
                                           :test #'string-equal))
            (setq plist (list* :classes classes plist))))))
    (when (eq (getf plist :definition *unbound-value*) *unbound-value*)
      (let ((key (html:elem-key elem)))				; default :definition t
        (cond ((memq key *anchor-definition-tags*)		; for certain elements
               (setf plist (list* :definition key plist)))
              ((and classes (intersection classes *anchor-definition-css-classes*
                                          :test #'string-equal))
               (setf plist (list* :definition :classes plist))))))
    (push (cons (last-segment *current-topic*) plist)		; last encoutnered so far
          (gethash anchor (anchors (asd:component-system *current-topic*))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  COMPILE  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun anchor-saplist (anchor hash-table)
 ;;; Collect ((segment . plist) ...) list of anchor definitions.
  ;; Args: anchor If virtual, descend to the definitions of its origins
  ;;		  in depth-first manner.
  (let* ((saplist (gethash anchor hash-table))
         (origins (remove-if-not #'stringp saplist)))
    (if origins
        (let ((path (list anchor))
              (acc ()))				; Q: Remove duplicates?
          (labels ((%push (pair fragment)	; equip plist with the :fragment property
                     (push (list* (car pair) :fragment fragment (cdr pair)) acc))
                   (%recurse (origin path)
                     (if (member origin path :test #'string-equal)
                         (cerror #L"Do not descend ~s again."
                           #L"Cicle detected in definition of virtual anchor ~s,~%path ~s."
                           origin path)
                         (dolist (pair (gethash origin hash-table))
                           (if (stringp pair)
                               (%recurse pair (cons origin path))
                               (%push pair origin))))))
            (dolist (origin origins)
              (%recurse origin path))
            (append (remove-if #'stringp saplist) (nreverse acc))))
        saplist)))
  
(defun translate-uri (uri &optional target referrer)
 ;;; Args: uri       Contains referred anchor, path, and maybe a different book
  ;;                 All components of the uri must be percent-decoded arlready.
  ;;	   referrer  NIL, referencing source component or uri instance.
  ;;	   target    NIL, "_popup", or standard HTML value "_blank", "_parent", etc.
  ;; Values: (1) Fresh instance of uri relative to base of the referrer component or
  ;;		 NIL if the anchor is not found.
  ;;	     (2) Attribute plist to be added to the element, e.g.
  ;;		   :title (title segment)
  ;;	     (3) Plist
  ;;		 - either exaclty from anchor definition inside the source segment,
  ;;		 - or :text (title component) when fragment is not specified in the uri.
  ;; NB1: If some off-the-shelf books have been registered by means of register-book
  ;;      before, links to them by IDs are being translated correctly.
  ;;      Is there any sense in defining/compiling two or more book systems simultaneously?
  ;; NB2:        ,    popup_menu;
  ;;	        popup,  .
  ;; Q: Shall we search for fragment in a case-insensitive manner?
  (cond
    ((eq (uri:uri-scheme uri) :help)
     (let* ((path (uri:uri-parsed-path uri))
            (book-id (uri:uri-host uri))
            ;(book-id (case (first path)
            ;           (:relative (prog1 (second path) (setq path (cddr path))))
            ;           (:absolute (prog1 nil (setq path (cdr path))))))
            (book (cond (book-id
                         (find-book book-id nil))		; an instance of book
                        ((and referrer (not (uri:uri-p referrer)))
                         (asd:component-system referrer))))	; or book-system
            (fragment (uri:uri-fragment uri))
            ;(anchor fragment)
            ;(charset 1251)
	    ;(anchor (uri:percent-decode fragment :windows-1251))
            (referrer-uri (cond ((uri:uri-p referrer) referrer)
                                (referrer (asd:component-output-pathname referrer)))))
       (cond
        ((null book)
         (log-warn #L"Help book is not found~@[: ~s~]." book-id))
        (fragment
         ;; "help:#anchor_name"
         ;; "help://book_id#anchor_name"
         ;; "help://book_id/chapter_name#anchor_name"
         ;; "help:/chapter_name#anchor_name" (current book assumed)
         ;; Open page in a help book (and chapter) identified by the anchor
         ;; within a descendant component of the book (and chapter if given).
	 ;; Explicit target="_popup" limits possible destinations.
         (let ((saplist (anchor-saplist fragment (anchors book))))	; fragment?
           (when path
             (setq saplist (remove path saplist
                                   :test-not #'path-segment-match-p
                                   :key #'car)))
           (when target					; filter distinations by target
             (setq saplist (remove-if (if (string-equal target +target-popup+)
                                          (complement #'popup-p)
                                          #'popup-p)
                                      saplist
                                      :key #'car)))
           (if saplist
               (let* ((saplist (if (rest saplist)
                                   (sort (copy-list saplist) #'segment< :key #'car)
                                   saplist))
                      (segment (caar saplist))
                      (plist (cdar saplist))
                      (u (absolute-uri book segment)))
                 (when referrer-uri
                   (setq u (uri:enough-uri u referrer-uri)))
                 (cond ((rest saplist)
			;; There several possible destinations - insert popup-menu to ask
                        (log-warn
                       "Anchor ~a~@[ target=~s~] is defined in multiple sources:~{~%; ~a~}"
                                  uri target (mapcar #'car saplist))
                        (setf (uri:uri-fragment u) (getf plist :fragment fragment))
                        (uri:clear-cache u)
                        (unless (popup-p *current-topic*)
                          (pushnew :menu (segment-property  ; making popupMenuDIV needed
                                          (nth-segment *current-segment* *current-topic*)
                                          :popups)))
                        (values u
                                (list :onclick (popup-menu-onclick
                                                referrer-uri
                                                (if target	; already filtered
                                                    saplist	; filter off popups
                                                    (remove-if #'popup-p saplist
                                                               :key #'car))
                                                :fragment fragment)
                                      :onkeydown +popup-onkeydown-js+
                                      :onkeyup +popup-onkeyup-js+)
                                (cdar saplist)))
                       ((popup-p segment)
			;; Single target is popup - making popupDIV needed
                        #1=(unless (popup-p *current-topic*)
                             (pushnew :note
                                      (segment-property 
                                       (nth-segment *current-segment* *current-topic*)
                                       :popups)))
                        (values u				; no fragment!
                                (list :onclick +popup-onclick-js+
                                      :onkeydown +popup-onkeydown-js+
                                      :onkeyup +popup-onkeyup-js+)))
                       (t
                        (setf (uri:uri-fragment u) (getf plist :fragment fragment))
                        (uri:clear-cache u)
                        (values u				; ordinary href
                                (list :title (title segment))
                                plist))))
               (progn (log-warn "No source found for anchor: ~a~@[ target=~s~]" uri target)
                      nil))))
        ((typep book 'book)
         ;; "help://book_id/path"
	 ;; Registered book instance but the current book-system
         ;; Return values suitable to be retruned from anchor-a
         (if-bind (location (uri:uri (location book)))		; pathname -> uri
           (let* ((dest (if (and path (not (equal (second path) +null-string+)))
                            (uri:make-uri :path (uri::%render-parsed-path path)
                                          :%parsed-path% path	; path to a file, not dir
                                          :fragment fragment)
                            (uri:uri (start-topic book))))	; help.html by default
                  (path (uri:uri-parsed-path location))		; referenced book base
                  (u (if (and (eq (first path) :relative)
                              (typep referrer 'asd:component))
                         ;; The location is relative to the root of all books,
                         ;; i.e. the run-time value of *default-book-path*. Assume 
                         ;; the current book system is intended to be placed there too.
                         ;; Compute the number of levels of the current component in
                         ;; the physical directory structure and
                         ;; add 1 to go up to the root of all books.
                         ;; The final URI will be relative.
                         (progn (setq path
                                      `(:relative
                                        ,@(make-list
                                           (1+ (- (length (pathname-directory
                                                           referrer-uri))
                                                  (length (pathname-directory
                                                           (asd:component-output-pathname
                                                            (asd:component-system
                                                             referrer))))))
                                           :initial-element "..")
                                        ,@(butlast (rest path)) ; exclude trailing ""
                                        ,@(rest (uri:uri-parsed-path dest))))
                           (uri:copy-uri dest
                                         :path (uri::%render-parsed-path path)
                                         :%parsed-path% path))
                         ;; The location is absolute - sipmly merge dest with it
                         ;; The final URI will be absolute too.
                         (uri:merge-uris dest location))))
             (values u
                     `(:target ,(or target "_blank"))		; attrs
                     `(:text ,(title book))))			; plist 
           (log-warn #L"Location of book~@[ ~s~] is not specified." book)))
        ((or (null path)		 			; book only
             (equal (second path) +null-string+))
         ;; "help://book_id" or "help://book_id/" or "help:/"
         ;; Link to the start topic of this instance of book-system
         (if-bind (segment (first-segment (start-topic book)))
           (let ((u (absolute-uri book segment)))
             (values (if referrer-uri (uri:enough-uri u referrer-uri) u)
                     nil
                     (when-let (title (title segment))
                       (list :text title))))
           (log-warn "Start topic~@[ ~s~] not found in book: ~a"
                     (start-topic-name book) uri)))
        ((equal (first (last path)) +null-string+)		; book and chapter
         ;; "help://book_id/chapter_name/"
         ;; "help:/chapter_name/"
         ;; "help:chapter_name/"
         ;; Directory - produce the link to start topic of a chapter
         (if-bind (chapter (asd:find-component book (butlast path)))
                                                      ;(subseq path 0 (1- (length path)))))
           (if-bind (segment (first-segment (start-topic chapter)))
             (let ((u (absolute-uri book segment)))
               (values (if referrer-uri (uri:enough-uri u referrer-uri) u)
                       nil
                       (when-let (title (title segment))
                         (list :text title))))
             (log-warn "Start topic~@[ ~s~] not found in chapter: ~a"
                       (start-topic-name chapter) uri))
           (log-warn "Chapter not found: ~a" uri)))
        (t							; topic  file
         ;; "help://book_id/path/topic"
         ;; "help://book_id/path/image.png"
         ;; "help:/path/topic"
         ;; "help:path/topic"
         ;; Topic, aux-file, or file in aux folder:
         ;; (:absolute ... "") -> (:absolute ... . path)
         ;; In path, names are interpreted in a descending relevance manner:
         ;;  - first, chapter or component names,
         ;;  - then, directory names,
         ;;  - finally, topic name or file namestring.
         (multiple-value-bind (component module path) (asd:find-component book path)
           (let ((u (if component
                        (absolute-uri book component)
                        (absolute-uri module path))))
             (if referrer-uri (uri:enough-uri u referrer-uri) u)))) )))
    ;((:file nil)		; default
    ; uri)
    ;; Otherwise one of "standard" URI schemes, HTTP is very likely.
    ((and target (string-equal target +target-popup+))
     #1#						; indicate the need of popup DIV
     (values uri
             (list :onclick +popup-onclick-js+
                   :onkeydown +popup-onkeydown-js+ :onkeyup +popup-onkeyup-js+)))
    (t uri)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  API  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Russian NB:
;;;    As a word can have up to six cases in Russian, the function anchor-text and
;;;    relatives are rather useless. They return what is set in the :text property,
;;;    usually the main case (imenitelny padezh).
;;;    In practice, the text argument should be specified explicitly for anchor-a.

(defun anchor-href (link &key target (referrer *current-topic*))
 ;;; Args: link   String or URI containing referred anchor or
  ;;              path to a different book.
  ;;       target Used to filter on/off anchors referring to pop-up notes.
  ;; Value: URI for the href attribute of an A element relative to referrer.
  (translate-uri (help-uri link) target referrer))

(defun %anchor-text (plist)
 ;;; Q: Shall we check against *anchor-span-css-class-properties*?
  (let ((text (getf plist :text)))
    (cond (text						; explicit text property
           (if (and (getf plist :bbcode *bbcode-in-text*)
                    (char-position #\[ text))
               (interleave-seq 'string #\Space
                               (lhtml-word-list (html:bbcode-lhtml text)))
               text))
          ((setq text (getf plist :content))		; concat from content
           (interleave-seq 'string #\Space (lhtml-word-list text))))))

(defun anchor-text (link &key target (referrer *current-topic*))
 ;;; Args: link   String or URI containing referred anchor or
  ;;              path to a different book.
  ;;       target Used to filter on/off anchors referring to pop-up notes.
  ;; Value: Text string or NIL if the anchor is not found or ambiguos.
  ;; Q: Add paramters?
  ;;    chapter           T or string means that the chapter title must precede the text.
  ;;    chapter-separator String or character.
  (multiple-value-bind (uri attrs plist)
      (translate-uri (help-uri link) target referrer)
    (declare (ignore attrs))
    (when uri
      (%anchor-text plist))))

(defun %anchor-content (plist &optional prefix (suffix t))
 ;;; Args: prefix If true, cons :content at the resulting LHTML list.
  ;;       suffix If true, append the caterogy or other span properties to HTML content.
  (let* (value
         (content (cond ((setq value (getf plist :text))
                         (if (and (getf plist :bbcode *bbcode-in-text*)
                                  (char-position #\[ value))
                             (html:bbcode-lhtml value)
                             (list value)))
                        ((setq value (getf plist :content))
                         (multiple-value-bind (span-class span-cont)
                             (if suffix
                                 (get-properties plist *anchor-span-css-class-properties*)
                                 nil)
                           (if span-class
                               (append value
                                       (list #\Space (html:make-elem :span
                                                      (list :class span-class)
                                                      (listify span-cont))))
                               value))))))
    (if prefix
        (cons :content content)
        content)))

(defun anchor-content (link &key target (suffix nil)
                                 (referrer *current-topic*))
 ;;; Args: link   String or URI containing referred anchor or
  ;;              path to a different book.
  ;;       target Used to filter on/off anchors referring to pop-up notes.
  ;; Value: LHTML content of the form (:content . cont), where the cont list is
  ;;        - either distilled content of the anchor element
  ;;        - or bbcode-converted :text property if the :content is empty.
  ;; NB: As can be used everywhere, the default for suffix is false
  (multiple-value-bind (uri attrs plist)
      (translate-uri (help-uri link) target referrer)
    (declare (ignore attrs))
    (when uri
      (%anchor-content plist t suffix))))
        
(defun anchor-a (link &key target (class nil supplied-p) content text (suffix nil)
                           (referrer *current-topic*))
 ;;; Args: link    String or URI containing referred anchor or
  ;;               path to a different book.
  ;;       target  Used to filter on/off anchors referring to pop-up notes.
  ;;	   class   CSS class name of the resulting A element.
  ;;	   content Explicit LHTML content used as an inner content for A element.
  ;;	   text	   Explicit string used as an inner content for A returned.
  ;; Value: LHTML content with A element like
  ;;		((:a :href uri :title title) text-or-content)
  ;; NB: As can be used everywhere, the default for suffix is false
  (multiple-value-bind (uri attrs plist)
      (translate-uri (help-uri link) target referrer)
    (when uri
      (let ((title (getf attrs :title)))
        `(:content ((:a :href ,uri
                     ,@(when (orf class (and (not supplied-p)
                                             (when-let (classes (getf plist :classes))
                                               (interleave-seq 'string #\Space classes))))
                         `(:class ,class))
                     ,@(when (and title (neq title text)) `(:title ,title))
                     ,@(remove-property attrs :title))
                    ,@(cond (content)				; explicit content
                            (text (listify (if (eq text t)	; default text
                                               (%anchor-text plist)
                                               text))) 		; explicit text argument
                            (t (%anchor-content plist nil suffix)))))))))

(defun anchor-topics (&rest args)
 ;;; Links to pages (without fragment) where the anchors reside.
  ;; Args: args  List of anchor strings or uri instances that must contain
  ;;		 a fragment component.
  ;; Value: Sequence of template environments used by TMPL_LOOP of the form:
  ;;  (((item             . (segment . plist))		; target segment and anchor plist
  ;;    (item-topic       . {topic | segment})
  ;;	(item-uri         . relative-uri)
  ;;	(item-class       . CSS-class)			; class for element A
  ;;	(item-title       . segment/topic-title-text)	; or inline-content
  ;;	(item-content     . heading-inline-content))	; inline-content
  ;;	(item-description . segment/topic-description)
  ;;   ...))
  ;; NB: If some anchor is defined in several topics, all of them are collected
  ;;     (including popups!)
  (let* ((base (asd:component-output-pathname (nth-segment *current-segment*
                                                           *current-topic*)))
         (hash-table (anchors *current-book*)))
   (sort
    (loop for link in args
          for fragment = (uri:uri-fragment (help-uri link))
          nconc (loop for pair in (gethash fragment hash-table)
                      ;for segment = (car pair)
                      ;and plist = (cdr pair)
                      collect (item-environment pair base nil nil)))	; no fragment!
    #'segment<
    :key (lambda (alist) (cdr-assoq 'item-topic alist)))))

(defmethod item-a ((pair cons) &key (class nil supplied-p) (suffix t))
 ;;; Args: item ::= (segment . plist)
  ;;       suffix If true, output :category and the like
  ;; NB: As used within iteratrion, the default for suffix is true
  (let* ((referrer (nth-segment *current-segment* *current-topic*))
         (segment (car pair))			; target segment
         (plist (cdr pair))
         (uri (uri:enough-uri (asd:component-output-pathname segment)
                              (asd:component-output-pathname referrer))))
    `(:content ((:a :href ,uri
                 ,@(when (orf class
                              (and (not supplied-p)
                                   (when-let (classes (getf plist :classes))
                                     (interleave-seq 'string #\Space classes))))
                     `(:class ,class))
                 :title ,(or (description segment) (title segment)))
                ,@(%anchor-content plist nil suffix)))))


;; Abbreviated function names
(setf (fdefinition 'aa)       (fdefinition 'anchor-a)
      (fdefinition 'acontent) (fdefinition 'anchor-content)
      (fdefinition 'ahref)    (fdefinition 'anchor-href)
      (fdefinition 'atext)    (fdefinition 'anchor-text)
      (fdefinition 'atopics)  (fdefinition 'anchor-topics))
