;;; -*- Mode: Lisp; -*-
;;; YstokHelp - HTML-based help
;;; Copyright (c) 2009-2014 Dr. Dmitry Ivanov.  All rights reserved.
;;; Website: http://lisp.ystok.ru/yhelp/
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Book component classes and simple methods

(in-package :ystok.help)
 
;;; help-module-mixin - superclass of book-system and chapter

(defclass help-module-mixin ()
 ((start-topic-name :reader start-topic-name :initarg :start :initform nil)
  (start-topic)))		; introduction, TOC, or something else, can be segmented

(defmethod start-topic ((module help-module-mixin))
 ;;; Find by name or take the very first child and cache it
  (if (slot-boundp module 'start-topic)
      (slot-value module 'start-topic)
      (setf (slot-value module 'start-topic)
            (if-bind (start-topic-name (start-topic-name module))
              (or (asd:find-component module start-topic-name)
                  (error #L"Start topic ~s is missing from ~s." start-topic-name module))
              (first (topics module))))))
      
(defmethod reinitialize-instance :after ((self help-module-mixin) &key)
  (slot-makunbound self 'start-topic))

;;; navigation-mixin - superclass of every page loaded into the navigation frame

(defclass navigation-mixin ()
 ((tab :reader tab :initarg :tab))) 			; tab name on toolbar


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  BOOK-SYSTEM, CHAPTER  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Book's id slot coincides with name by default (useful when the name is occupied
;;; by a "usual" Lisp system definition).
;;; Additional important properties (available via asd:component-property)
;;;   braces (default t)
;;;	Use { } in templates, see html-template:*accept-braces-markers*

(defclass book-system (help-module-mixin book-mixin asd:system)
 ((id :reader id :initarg :id)
  (content-header :initarg :content-header	; cont page header toolbar specification
	:initform '(:prev :up :next :toggle))
  (content-footer :initarg :content-footer	; cont page footer toolbar specification
	:initform '(:prev :up :next))
  (use-templates :reader use-templates		; topics contain <-- TMPL_tag -->
	:initarg :use-templates			; if true, call html:parse-htt
	:initform (if htt:*accept-braces-markers*
                      :braces			; used as default for parsed-files
                      t))
  (charset :reader charset :initarg :charset	; default output charset (canonical)
	:initform nil)				; - string or key as of content-type
  (language :reader language :initarg :language	; default language tag for output pages
	:initform nil)
  (dump :reader dump :initarg :dump		; default dump property for all chapters
	:initform t)				; and topics
  (environment :reader environment		; LHTT environment - alist of book-specific
	:initarg :environment  :initform ())	; variables used for compiling all topics
  (entities :reader entities			; Pseudo-entities - alist ((name . exp)...)
	:initarg :entities  :initform ())	; where expansion is either LHTML or string
  (topics :initform nil)			; cache of all levels topics except navi
  (aliases :initarg :aliases :initform ())))	; virtual anchors

(defun %initialize-anchors (book-system)
 ;;; Initialize the anchors slots - table of HTML anchors and aliases
  (let ((hash-table (anchors book-system)))
    (unless (hash-table-p hash-table)			; no clearing otherwise
      (setf hash-table (make-hash-table :test #'equal)
            (anchors book-system) hash-table))
    (loop for (anchor . origins) in (slot-value book-system 'aliases)	; virtual anchors
          do (setf (gethash anchor hash-table) origins))))

(defun %initialize-book-system (self &key language id 
                                          charset external-format output-html-format
                                          entities
                                     &allow-other-keys)
 ;;; Actually, all system options are set in reinitialize-instance
  ;; Args: language Keyword or string
  ;; 	   charset  Unparsed charset string as of HTTP
  ;; Convert language tag to the keyword if needed
  (cond ((stringp language)
         (setf *language* (intern-upcase language)
               (slot-value self 'language) *language*))
        ((keywordp language)
         (setf *language* language))
        (t
         (when language
           (log-warn #L"Wrong book language attribute value ~s." language))
         (setf (slot-value self 'language) *language*)))
  (unless id
    (setf (slot-value self 'id) (asd:component-name self)))
  (when charset
    (setf charset (html:canonicalize-charset charset)
          (slot-value self 'charset) charset))
  (unless external-format
    (setf (slot-value self 'external-format) (if charset
                                                 (html:charset-ef charset)
                                                 htt:*external-format*)))
  (unless output-html-format
    (setf (slot-value self 'output-html-format) (if charset
                                                    (html:charset-ef charset)
                                                    html:*html-external-format*)))
  (phtml:ensure-entity-table
   (mapcar (lambda (pair &aux (val (cdr pair)))		; if a string defines the entity,
             (if (stringp val)				; parse it
                 (cons (car pair) (first (phtml:parse-html val)))
                 pair))
           entities)
   phtml:*html-entity-to-code*)
  (setf (slot-value self 'topics) nil)
  (%initialize-anchors self)				; prepare HTML anchors and aliases
  (unless (hash-table-p (text-anchors self)) 		; initialize table of text anchors
    (setf (text-anchors self) (make-hash-table :test #'equal)))
  (register-book self))

(defmethod reinitialize-instance :after ((self book-system) &rest args)
  (apply #'%initialize-book-system self args))

;(defmethod title ((self book-system))
;  (asd:component-name self))

(defmethod description ((self book-system))
  (asd:system-description self))

(defmethod ensure-book ((self book-system))
  self)

(defmethod topics ((book-system book-system))
  (orf (slot-value book-system 'topics) (call-next-method)))

(defmethod module-components< ((module null) (x book-system) (y book-system))
  nil)

(defmethod absolute-pathname ((book-system book-system) (component asd:component))
  (asd:component-output-pathname component))

(defmethod absolute-uri ((book-system book-system) (component asd:component))
  (asd:component-output-pathname component))

(defgeneric toc-level (self)
  (:method (self) 0))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  CHAPTER  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass chapter (help-module-mixin asd:module)
 (;; Default dump property for all topics, :parent redirects to parent
  (dump :initarg :dump :initform :parent)
  ;; logical level delta added to toc-level of all the topics of the chapter
  (toc-level :reader toc-level :initarg :toc-level :initform 0)
  ;(numeration :initarg :numeration :initfrom "D")	; D=Decimal,R=Roman,A=Alphabetic
  ;(number)
))

(defmethod id ((self chapter))
  (asd:component-name self))

(defmethod dump ((self chapter))
  (let ((dump (slot-value self 'dump)))
    (if (eq dump :parent) (dump (asd:component-parent self)) dump)))

(defmethod start-topic ((module chapter))
 ;;; The method differs in default: take the start-topic of the parent if toc-level=-1
  ;;         
  (if (slot-boundp module 'start-topic)
      (slot-value module 'start-topic)
      (setf (slot-value module 'start-topic)
            (if-bind (start-topic-name (start-topic-name module))
              (or (asd:find-component module start-topic-name)
                  (error #L"Start topic ~s is missing from ~s." start-topic-name module))
              (if (eql (toc-level module) -1)
                  (start-topic (asd:component-parent module))
                  (first (topics module)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  PARSED-FILE  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Source file that is analysed for multiple purposes:
;;; splitting, anchor substitution etc.

(defclass parsed-file (asd:dynamic-file)
 ((anchors :accessor anchors :initform nil)))	; (string1 string2 ...) or ((anc1 ...) ...)

(defmethod reinitialize-instance :after ((self parsed-file) &key title description)
  (if title
      (setf (asd:component-property self :title) title)
      (remf (asd::component-properties self) :title))
  (if description
      (setf (asd:component-property self :description) description)
      (remf (asd::component-properties self) :description)))

(defmethod title ((self parsed-file))
  (asd:component-property self :title))

(defmethod description ((self parsed-file))
  (asd:component-property self :description))

(defmethod print-object ((self parsed-file) stream)
  (if *print-escape*
      (call-next-method)
      (write-string (or (title self) (asd:component-name self)) stream)))

(defmethod segment-number ((self parsed-file)) 0)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  PARSED-HTML-FILE  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; This is assumed a kind of segment.
;;; Every segment-... funciton should be applicable to a parsed-html-file instance,
;;; except for text-topic.
;;; ASDF NB: Do not mix up with asdf:html-file, which is a subclass of asd:static-file!
;;; Q: Shall the :title in book definition override <title> within the HTML-file?

(defclass parsed-html-file (parsed-file)
 ((segments :accessor segments :initform nil)))		; list of segment instances
  ;(title :initarg :title :initform nil)		; overwrites the HEAD's one in TOC

(defmethod reinitialize-instance :after ((self parsed-html-file)
                                         &key (braces t supplied-p))
  (if supplied-p
      (setf (asd:component-property self :braces) braces)
      (remf (asd::component-properties self) :braces)))

;(defmethod title ((self parsed-html-file))
;  (or (slot-value self 'title) (asd:component-property self :title)))|#

(defmethod (setf headings) (value (c parsed-html-file))
  value)					; for common parser functions compatibility

(defmethod asd:source-file-type ((c parsed-html-file) s)
  "html")

(defmethod segment-property ((self parsed-html-file) property &optional default source)
  (declare (ignore source))
  (asd:component-property self property default))

(defmethod (setf segment-property) (value (self parsed-html-file) property
                                    &optional default)
  (setf (asd:component-property self property default) value))

(defmethod asd:component-output-pathname ((c parsed-html-file))
  (let ((pathname (asd:output-pathname-of-type c)))
    (if (segments c)
        (make-pathname :name (format nil +segment-file-format-string+
                                     (pathname-name pathname) 0)
                       :defaults pathname)
        pathname)))

;;; In theory, :compile should be triggered by
;;; - redefinition of the parent chapter or any grandparent,
;;; - :preprocess of itself or any topic this component has links to.

(defmethod asd:operation-requires ((o (eql :compile)) (c parsed-html-file))
  (declare (ignorable o))
  (cons `(:preprocess ,c ,@(siblings c)) ;'(or chapter (and topic (not text-topic)) popup)
        (call-next-method)))

#|(defmethod asd:operation-caused-by ((o (eql :compile)) (c parsed-html-file))
  (declare (ignorable o))
  (cons `(:preprocess ,c ,@(siblings c)) ;'(or chapter (and topic (not text-topic)) popup)
        (call-next-method)))|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  SEGMENT  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass segment ()
 ((source-component :reader source-component :initarg :source-component)
  (number :reader segment-number :initarg :number)
  (plist :accessor plist :initarg :plist :initform ())))

(defmethod print-object ((self segment) stream)
  (if *print-escape*
      (print-unreadable-object (self stream :type t :identity t)
        #1=(format stream "~d ~s"
                   (segment-number self) (asd:component-name (source-component self))))
      (if-bind (title (title self))
        (write-string title stream)
        #1#)))

(defun make-segment (topic plist)
  (let* ((segments (segments topic))
         (segment (make-instance 'segment :source-component topic
                                 :number (1+ (length segments))
                                 :plist plist)))
    (setf (segments topic) (nconc segments (list segment)))
    segment))

(defun first-segment (topic)
  topic)

(defun last-segment (topic)
  (when topic
    (if-bind (segments (segments topic))
      (first (last segments))
      topic)))

(defun nth-segment (number topic)
  (if (iplusp number)
      (let ((segments (segments topic)))
        (if (<= number (length segments))
            (nth (1- number) segments)
            (error #L"Segment ~d is not found in ~s." number topic)))
      topic))

(defmethod segment-property ((self segment) property &optional default source)
 ;;; Args: default :source = for non-zero segment, return topic property if own is missing
  ;;               Used for :title, :index, :robots      
  (let ((value (getf (plist self) property *unbound-value*)))
    (cond ((neq value *unbound-value*)
           value)
          ((and source
                (neq (setq value (segment-property (source-component self) property
                                                   *unbound-value*))
                     *unbound-value*))
           value)
          (t default))))

(defmethod (setf segment-property) (value (self segment) property &optional default)
  (declare (ignore default))
  (setf (getf (plist self) property) value))

(defmethod title ((self segment))
  (segment-property self :title nil t))			; default from source

(defmethod description ((self segment))
  (segment-property self :description nil t))		; default from source

(defmethod dump ((self segment))
  (dump (source-component self)))

(defun segment< (x y)
  (let ((x-topic (source-component x))
        (y-topic (source-component y)))
    (if (eq x-topic y-topic)
        (< (segment-number x) (segment-number y))
        (component-path< (component-path x-topic) (component-path y-topic)))))

(defmacro do-segments ((segment topic) &body body)
 ;;; Loop over segment designators
  (with-gensyms (tpc)
    `(let ((,tpc ,topic))
       (dolist (,segment (cons ,tpc (segments ,tpc)))
         ,@body))))

(defmethod asd:component-system ((segment segment))
  (asd:component-system (source-component segment)))

(defmethod asd:component-output-pathname ((segment segment))
  (let ((pathname (asd:output-pathname-of-type (source-component segment))))
    (make-pathname :name (format nil +segment-file-format-string+
                                 (pathname-name pathname) (segment-number segment))
                   :defaults pathname)))

(defmethod absolute-uri ((book-system book-system) (segment segment))
  (asd:component-output-pathname segment))

(defun path-segment-match-p (pattern segment)
 ;;; Args: pattern  List representing a uri-parsed-path;
  ;;		    can contain wildcard "*" as any element;
  ;;                if :relative, relative to *current-topic*.
  ;;	   segment  Segment or source component (topic, popup, etc.)
  ;; Value: True if the segment or topic lays on the path specified by pattern
  ;;	    from the root, i.e. the book itself.
  ;; Called by: translate-uri
  ;; NB: Returns true if the shortest common parts match and the pattern is shorter, e.g.
  ;;     (path-segment-match-p '(a b) '(<module a> <module b> <module c>)) => true
  (do* ((component (source-component segment))
        (x (case (first pattern)
             (:absolute (rest pattern))
             (:relative (append (rest (component-path *current-topic*))
                                (rest pattern)))
             (otherwise pattern))			; as if :absulute
           (rest x))
        (y (rest (component-path component))		; skip book at the list head
           (rest y)))
       ((not (and x y))
        (not x))					; => false if the pattern is longer
    (let ((first (first x)))
      (cond ((not (stringp first))			; component before relative part
             (unless (eq first (first y))		; - must coincide
               (return nil)))
            ((string= first "*"))
            ((not (string-equal first (asd:component-name (first y))))
             (return nil))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  TOPIC  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Topic-mixin produces a content page or pages that are placed into
;;; the page sequence of the book.

(defclass topic-mixin ()
 ((headings :accessor headings :initform nil)))	; ((level id . cont)...)

(defmethod reinitialize-instance :after ((self topic-mixin) &key)
  ;; Annul all component operations (actually only :preprocess has to be annulled)
  (dolist (action (asd::actions self))
    (setf (asd::action-timestamp action) nil)))

(defmethod topics ((self topic-mixin))
  (if (asd:component-featurep self)
      (list self)
      ()))

(defclass topic (topic-mixin parsed-html-file)
 (;(headings :accessor headings :initform nil)	; ((level id . cont)...)
  (dump :initarg :dump :initform :parent)))	; redirect to parent by default

(defmethod reinitialize-instance :after ((self topic) &key (index t supplied-p) popups)
  ;; Annul all component operations (actually only :preprocess has to be annulled)
  (if supplied-p
      (setf (asd:component-property self :index) (or index :none))	; overwrites ROBOTS
      (remf (asd:component-properties self) :index))
  (setf (segment-property self :popups) popups))

(defmethod dump ((self topic))
  (let ((dump (slot-value self 'dump)))
    (if (eq dump :parent) (dump (asd:component-parent self)) dump)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  NAVIGATION  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Loaded into the left frame, excluded from page sequence

(defclass navigation (navigation-mixin parsed-html-file) ())

#|(defmethod asd:operation-requires ((o (eql :compile)) (c navigation))
  (declare (ignore o))
  (cons `(:preprocess ,@(siblings c '(or chapter parsed-html-file)))
        (call-next-method)))

(defmethod asd:operation-caused-by ((o (eql :compile)) (c navigation))
 ;;; In fact, :compile should be triggered by :preprocess of every subling also.
  (declare (ignore o))
  (cons `(:preprocess ,c) (call-next-method)))|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  POPUP  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Loaded only in pop-up frame, excluded from page sequence
;;; PDF: Markup annotations, which appear as highlights, underlines, strikeouts,
;;; or jagged ("squiggly" - ) underlines in the text of a document.

(defclass popup (parsed-html-file) ())

(defgeneric popup-p (object)
 (:method (object) (declare (ignore object)) nil)
 (:method ((object popup)) (declare (ignore object)) t)
 (:method ((segment segment)) (popup-p (source-component segment))) )

(defmethod segment-property ((self popup) (property (eql :index)) &optional default source)
  (declare (ignore self property default source))
  :none)			; never index popup content

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  NON-PARSED COMPONENTS  ;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; AUX-FILE - Source HTML/JS/CSS file - just copy
;;; Though aux-file does not need preprocessing, we need the perform method below
;;; because such a component can be embedded within a chapter, and :force t
;;; cause for a preprocess action to be pushed into the plan.
;;;  "help.html"    - master frameset
;;;  "help.js"
;;;  "popupref.js"
;;;  "popup.js"
;;;  "navi.css"
;;;  "cont.css"

(defclass aux-file (asd:source-file) ())

(defmethod asd:operation-done-p ((o (eql :preprocess)) (c aux-file))
  (declare (ignore o c))
  t)

(defmethod asd:perform ((o (eql :preprocess)) (c aux-file))
  (declare (ignore o c)))

(defmethod asd:output-files ((o (eql :compile)) (c aux-file))
  (declare (ignore o))
  (list (asd:component-output-pathname c)))

(defmethod asd:perform ((o (eql :compile)) (c aux-file))
  (mapc #'log-copy-file (asd:input-files o c) (asd:output-files o c)))

;;; AUX-FOLDER - a directory with graphics files or the like - just copy in full.
;;; For :preprocess, both input nor output files are empty, so asd:operation-done-p
;;; is always true.
;;; For :compile, the output-files list is empty (as returned by a default method),
;;; but they are modeled by asd:operation-done-p and asd:perform.

(defclass aux-folder (asd:module)
 (;(output-pathname :initarg :output-pathname  :initform nil)
  (exclude :initarg :exclude  :initform ())))		; list of file-namestrings

(defmethod absolute-uri ((module asd:module) (path cons))
 ;;; Args: module  Book-system, chapter or aux folder
  ;;       path    Relative parsed path to file in aux folder:
  ;; Value: Convert (:absolute ... "") -> (:absolute ... . path)
  (let ((uri (uri:uri (asd:component-output-pathname module))))	; always end with ""
    (setf (uri:uri-parsed-path uri) (nconc (butlast (uri:uri-parsed-path uri))
                                           (if (eq (first path) :relative)
                                               (rest path)
                                               path)))
    uri))

(defmethod asd:input-files ((o (eql :compile)) (c aux-folder))
  (let ((exclude (slot-value c 'exclude)))
    (delete-if (lambda (pathname)
                 (or (directory-pathname-p pathname)
                     (member pathname exclude :test #'pathname-match-p)
                     (intersection (pathname-directory pathname) *exclude-subdirectories*
                                   :test #'string-equal)))
                 (directory (merge-pathnames (make-pathname
                                              :directory '(:relative :wild-inferiors)
                                              :name :wild :type :wild)  ; "**/*.*"
                                             (asd:component-pathname c))))))

(defmethod asd:operation-done-p ((o (eql :compile)) (c aux-folder))
 ;;; Simplified time test:
  ;; - react on deletion of the entire target directory,
  ;; - do not react on addition of a new file.
  (let ((output-dir (asd:component-output-pathname c)))
    (when (probe-file output-dir)			; fad:directory-exists-p
      (if (asd::get-timestamp o c)			; already was copied
          (call-next-method)				; - check when
          (loop with input-dir = (asd:component-pathname c)
                for input-file in (asd:input-files o c)
                for output-file = (merge-pathnames (enough-pathname input-file input-dir)
                                                   output-dir)
                always (and (truename input-file)
                            (probe-file output-file)
                            (<= (file-write-date input-file)
                                (file-write-date output-file))))))))

(defmethod asd:perform ((o (eql :compile)) (c aux-folder))
  (let ((input-dir (asd:component-pathname c))
        (output-dir (asd:component-output-pathname c)))
    (log-info "copy dir: ~a -> ~a" input-dir output-dir)
    (let ((output-files 
           (loop for input-file in (asd:input-files o c)
                 for output-file = (merge-pathnames (enough-pathname input-file input-dir)
                                                    output-dir)
                 unless (and (probe-file output-file)
                             (<= (file-write-date input-file)
                                 (file-write-date output-file)))
                 do (ensure-directories-exist output-file)
                 (copy-file input-file output-file)
                end
                collect output-file)))
      (dolist (output-file (directory output-dir))
        (unless (member output-file output-files :test #'pathname-equal)
          (log-info "deleting ~a" output-file)
          (delete-file output-file nil))))))
    
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  GENERATED COMPONENT  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Virtual component that not based on any source file.
;;; Its slots are usually referred in the HTML title attribute or in TOC
;;; CAUTION: Placing generated components within a chapter is untested.
;;;          If so, we would additionally need methods on :preprocess!

(defclass generated-component (asd:component)
 ((title :reader title :initarg :title :initform nil)
  (description :reader description :initarg :description :initform nil)))

(defmethod asd:input-files (o (c generated-component))
  (declare (ignore o c))
  ())

(defmethod asd:component-output-pathname ((component generated-component))
  (asd:output-pathname-of-type component "html"))	; relative-pathname should be null

(defmethod asd:output-files ((o (eql :compile)) (c generated-component))
  (declare (ignore o))
  (list (asd:component-output-pathname c)))

(defmethod asd:operation-caused-by ((o (eql :compile)) (c generated-component))
 ;;; In theory, should be triggered by
 ;;; - :preprocess of any system topic except frameset.
  (declare (ignore o))
  (cons `(:preprocess ,@(siblings c '(or topic chapter)))	; including text-topic
        (call-next-method)))

(defmethod asd:operation-done-p ((o (eql :compile)) (c generated-component))
 ;;; Assumed done if all output files exist
  ;; NB: We must specialize as the default method always returns NIL.
  (every #'probe-file (asd:output-files o c)))

(defclass frameset (generated-component) ())

;;; TOC and heading lists

(defstruct (heading (:copier nil))
  tag				; tag of defining HTML element
  segment
  level				; final level in the TOC of the book
  id				; id attribute, explicit or auto generatated
  class				; class attribute - CSS class name
  plist				; from *anchor-span-css-class-properties*
  content
  explicit)			; true (actually key) if toc-level property were specified
