;;; -*- Mode: Lisp; -*-
;;; YstokHelp - HTML-based help
;;; Copyright (c) 2009-2014 Dr. Dmitry Ivanov.  All rights reserved.
;;; Website: http://lisp.ystok.ru/yhelp/
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Context-popup - parse and preprocess text files

;;; Text displayed in context-sensitive help pop-up window within the application
;;; The reason for storing such text in a help book - it is too large for tooltip.
;;; Segment headers (separators):
;;; (1) #\Newline #\Page anchor #\Newline
;;; (2) #\Newline <!-- HELP_SEGMENT (:id anchor) --> #\Newline

(in-package :ystok.help)

(defconstant +whitespace-bag+ '(#\Space #\Tab))

(defclass context-popup (parsed-file) ())

(defmethod asd:source-file-type ((c context-popup) s)
  "txt")

(defmethod asd:component-output-pathname ((c context-popup))
  (asd:output-pathname-of-type c "cxt"))

(defmethod asd:output-files ((o (eql :compile)) (c context-popup))
  (list (asd:component-output-pathname c)))

#|(defmethod asd:operation-caused-by ((o (eql :compile)) (c context-popup))
 ;;; :compile is triggered only by :preprocess of itself
  (declare (ignore o))
  `((:preprocess ,c)))|#

;;; Helpers

(defun reverse-trim-lines (lines)
 ;;; Trim completely empty lines from the head and tail of the list
  ;; and trim spaces from the right for remaining lines.
  (flet ((%trim ()
           (loop while (and lines (every #'whitespace-char-p (first lines)))
                 do (setq lines (rest lines)))))
    (%trim)				; from end
    (setq lines (nreverse lines))
    (%trim))				; from start
  (do ((rest lines (rest rest)))
      ((null rest) lines)
    (rplaca rest (string-right-trim +whitespace-bag+ (car rest)))))

(defun write-context (book-system stream lines start
                      &optional (external-format (external-format book-system)))
 ;;; Args: stream Binary output stream
  ;; Value: The number of bytes written.
  (declare (ignore start))
  (multiple-value-bind (octets size)
      (string-to-octets (interleave-seq 'string #\Newline lines)
                        :external-format external-format :null-terminate nil)
    (write-sequence octets stream)
    size))

(defmethod asd:perform ((o (eql :preprocess)) (component context-popup))
 ;;; C,   external format   
  (declare (ignore o))
  (let ((input-file (asd:component-pathname component))
        (external-format (external-format (asd:component-system component)))
        (hash-table (text-anchors (asd:component-system component))))
    (log-info "preprocess: ~a" input-file)
    ;; Clear-anchors: clean up anchors gathered from previous compile
    (loop for alist being each hash-value of hash-table using (hash-key anchor)
          for tuple = (assoq component alist)
          when tuple
            if (deletef alist tuple)
            do (setf (gethash anchor hash-table) alist)
            else do (remhash anchor hash-table))
    (setf (anchors component) ())
    (with-open-file (in input-file :direction :input :external-format external-format)
      (let ((anchors ())
            (state :start)
            (line nil)
            len				; = (length line)
            lines			; list of text strings collected in reverse order
            start			; starting position of an anchor in the line
            anchor)
        (flet ((%read ()
                 (neq (setq line (read-line in nil *unbound-value*)) *unbound-value*))
               (%push ()
                 (when anchor
                   (if (setq lines (reverse-trim-lines lines))
                       (setq anchors (acons anchor lines anchors))
                       (log-warn #L"Empty text for ~s - anchor ignored." anchor))
                   (setq anchor nil
                         lines ())))
               (%anchor-start (&aux pos)
                 (cond ((zerop (setq len (length line)))
                        nil)
                       ((char= (schar line 0) #\Page)
                        1)
                       ((and (string= line "<!--" :end1 (min (setq len (length line)) 4))
                             (setq pos (position-if-not #'whitespace-char-p line
                                                        :start 4))
                             (string= line +segment-tag+ :start1 pos
                                      :end1 (min len (incf pos #.(length +segment-tag+)))))
                        pos))) )
          (loop
           (when (or (eq line *unbound-value*) (not (or line (%read))))
             (return))						; end of file
           (case state
             (:start
              (if (or (zerop (length line))			; skip blank lines at top
                      (every #'whitespace-char-p line))
                  (setq line nil)
                  (setq state :anchor)))
             (:anchor
              (if (and (setq start (%anchor-start))		; anchor start pos found
                       (if (= start 1)
                           ;; #\Page anchor
                           (if (zerop (length (setq anchor (string-trim
                                                            +whitespace-bag+
                                                            (subseq line start)))))
                               (setq anchor nil)
                               t)
                           ;; <!-- HELP_SEGMENT (:id anchor) -->
                           (when-let (plist (read-from-string line nil nil :start start))
                             (setq anchor (getf plist :id)))))
                  (setq state :text)
                  (setq state :start))				; strange
              (setq line nil))
             (:text						; read lines of text for
              (setq lines (list line))				; the anchor just read
              (loop while (and (%read) (not (%anchor-start)))
                    do (push line lines))
              (%push)
              (setq state :anchor))))				; end of loop
          (%push))						; push the last or warn
        (setf (anchors component) (nreverse anchors))))))

(defmethod asd:perform ((o (eql :compile)) (component context-popup))
  (declare (ignore o))
  (let* ((book-system (asd:component-system component))
         (hash-table (text-anchors book-system))
         (output-file (asd:component-output-pathname component)))
    (log-info "compile: ~a" (asd:component-pathname component))
    (with-open-file (out output-file :direction :output :if-exists :supersede
                         :element-type 'octet)
      (loop with offset = 0
            for (anchor . lines) in (anchors component)
            for start = offset
            for size = (write-context book-system out lines start)
            and alist = (gethash anchor hash-table)
            for pair = (cons start (incf offset size))
            and tuple = (assoq component alist)
            if tuple					; (component . whatever)
            do (rplacd tuple pair)			; -> (component . (start . end))
            else do (setf (gethash anchor hash-table) (acons component pair alist))
            end))
    (setf (anchors component) ())			; drop alist of text contents
    (log-info "compile-> ~a" output-file)))
