;;; -*- Mode: Lisp; -*-
;;; YstokHelp - HTML-based help
;;; Copyright (c) 2009-2014 Dr. Dmitry Ivanov.  All rights reserved.
;;; Website: http://lisp.ystok.ru/yhelp/
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Template-based and generating helpers

(in-package :ystok.help)

;;;;;;;;;;;;;;;;;;;;;;;;  ECMAScript handlers and calls  ;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun escape-ecma-string (string)
 ;;; Minimally escape the characters to be used inside a string passed to
  ;; an ECMAScript function and not handled by HTMLGen.
  (if (and (stringp string)
           (or (char-position #\' string)
               (char-position #\\ string)
               (char-position #\Newline string)))
      (with-output-to-string (stream nil :element-type (array-element-type string))
        (loop for char across string
              do (cond ((char-position char "'\\")
                        (write-char #\\ stream) (write-char char stream))
                       ((char= char #\Newline)
                        (write-string "\\n" stream))
                       (t (write-char char stream)))))
      string))

(defun popup-menu-onclick (referrer-uri saplist &key fragment title)
 ;;; Produce ECMAScript handler expression:
  ;;	popup_menu(event,'Menu caption',['item-uri','item-text',item-title,...])
  ;; Args: referrer-uri Absolute uri/pathname of the referencing source file or NIL.
  ;;       saplist      Already sorted according to segment<.
  ;; Value: Expression string with links sorted in "natural" order of their topics.
  ;; NB: We do not place explicit 'return' as the popup_menu function cancels
  ;;	 the default behavior itself.
  (format nil +popup-menu-onclick-js+ ;"popup_menu(event,~@['~a',~][~{~a~^,~}])"
          title
          (loop for (segment . plist) in saplist
                for uri = (absolute-uri (asd:component-system segment) segment)
                when referrer-uri
                  do (setq uri (uri:enough-uri uri referrer-uri))
                end
                unless (popup-p segment)			; exclude popups
                  when fragment
                    do (setf (uri:uri-fragment uri) (getf plist :fragment fragment))
                  end
                  and collect (format nil +popup-menu-item-js+
                                      (uri:render-uri uri)
                                      ;(nth-value 1 (get-properties plist '(:title :text)))
                                      (escape-ecma-string (title segment))
                                      (escape-ecma-string (description segment))))))

(defun ensure-jscript-lhtml (content pattern new)
 ;;; If no manully inserted script with code PATTERN found at the content top-level,
  ;; return LHTML list with calls to the NEW code.
  (unless (find-if (lambda (elem) (and (eq (html:elem-key elem) :script)
                                       (string-equal (getf (html:elem-attrs elem) :type)
                                                     "text/javascript")
                                       (search pattern (first (html:elem-cont elem)))))
          content)
    `((:jscript ,new)
      :newline)))

(defun body-end-lhtml (segment content &key (popups (segment-property segment :popups)))
 ;;; Args: content Old content of BODY element
  (when popups
    (ensure-jscript-lhtml content "make_popup"
                          (format nil "~:[~;make_popup();~]~:[~;make_popup_menu();~]"
                                  (memq :note popups) (memq :menu popups)))))
    #|`((:jscript ,(format nil "~:[~;make_popup();~]~:[~;make_popup_menu();~]"
                         (memq :note popups) (memq :menu popups)))
      :newline)))|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;  HTML document pieces  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ensure-charset-lhtml (head-cont component
                                      &optional (book (asd:component-system component)))
 ;;; Analyse existing content of the HEAD element and update or append
  ;; the meta-tag Content-Type (not Content-Language to comply HTML5)
  (let ((new (charset book))
        (old (asd:component-property component :charset)))
    (when (and new
               (or (not old) (string-not-equal new old)))
      (let ((value (string-append "text/html; charset=" new)))
        (if-bind (elem (html:find-element :meta head-cont :http-equiv :content-type
                                          :test #'string-equal))
          (setf (getf (html:elem-attrs elem) :content) value)
          (push-end (html:make-elem :meta
                                    (list :http-equiv :content-type :content value)
                                    ())
                    head-cont))))
    #|(when (and (language book)
               (not (html:find-element :meta head-cont :http-equiv :content-language
                                       :test #'string-equal)))
      ;; A wrong language is unlikely specified in the source HTML file.
      (push-end (html:make-elem :meta
                                (list :http-equiv :content-language :content lang)
                                ())
                head-cont)))|#
    head-cont))

(defun ensure-no-inverse-elem (key content attr &optional value &rest attrs)
 ;;; Provide an element having no inverse (no content) with attribute and value specified
  ;; Value: LHTML list or NIL if a matching link already exists in the content given.
  (unless (html:find-element key content attr value :test #'string-equal)
    `(((,key ,attr ,value ,@attrs)))))

(defun %in-uri-path-test (uri name)
 ;;; Args: uri   Subject - attribute value.
  ;;       name  Pattern - component name.
  (cond ((uri:uri-p uri) (member name (uri:uri-parsed-path uri) :test #'string-equal))
        ((pathnamep uri) (member name (pathname-directory uri) :test #'string-equal))
        (t (search name uri))))				; string

(defmethod ensure-aux-link-lhtml ((key (eql :script)) name content component
                                  &optional (book (asd:component-system component)))
 ;;; Args: name    Component name
  ;;       content HEAD element content
  ;; Value: LHTML list or NIL if a matching link already exists in the content given.
  (unless (html:find-element key content :src name :test #'%in-uri-path-test)
    (when-let (aux-file (asd:find-component book name))
      (list `((:script :src ,(uri:enough-uri (absolute-uri book aux-file)
                                             (absolute-uri book component))
               :type "text/javascript"))
            :newline))))

(defmethod ensure-aux-link-lhtml ((key (eql :link)) name content component
                                  &optional (book (asd:component-system component)))
 ;;; Insert the link to the stylesheet at the top-level of the book if none
  ;; or amend existing link (needed when importing topics from another book).
  (when-let (aux-file (asd:find-component book name))
    (let ((uri (uri:enough-uri (absolute-uri book aux-file)
                               (absolute-uri book component))))
      (if-bind (elem (html:find-element key content :href name :test #'%in-uri-path-test))
        (progn (setf (getf (html:elem-attrs elem) :href) uri)	; destructively modify
          ())
        (list `((:link :href ,uri :rel "stylesheet" :type "text/css"))
              :newline)))))

(defgeneric head-cont-lhtml (component old-cont &key &allow-other-keys)
 ;;; Return LHTML list for the HEAD element of the file-based component.
  ;; Ensure charset and language meta-tags, links to root aux-files etc.
  ;; Args: old-cont Existing content of the HEAD element.
  ;;       level    Depth of the topic page(s) within physical directory structure
  ;;	   popups   True if the component (topic or segment) refers to popup
  ;;		    windows or menus
 (:method (component old-cont &key)
  old-cont))

(defgeneric toolbar-lhtml (component position segment &key book &allow-other-keys)
 ;;; LHTML list to be appended at the start of BODY content
  ;; Args: component Instance of parsed-html-file (or generated-component?)
  ;;       position ::= :header | :footer
  ;;       content   LHTML Content of the BODY element for the segment
  ;;		     (before html:parse-htt applied)
 (:method (component position segment &key book)
  (declare (ignore component position segment book))
  ()))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  LHTT  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod html:template-environment ((segment segment)
                            &optional (uri (asd:component-output-pathname segment)))
  (html:template-environment (source-component segment) uri))

(defmethod html:template-environment ((topic topic)
                            &optional (uri (asd:component-output-pathname topic)))
  (declare (ignore topic uri))
  ())

;;; Helpers

(defun class-attr-match-p (pattern attr)
 ;;; Args: pattern ::= keyword | string	- match string
  ;;		     | "*" | :all	- match any attr including empty.
  ;;		     | NIL | :none	- match empty (not specified), i.e. NIL or ""
  ;;	             | (not pattern) | (and pattern...) | (or pattern...)
  ;;		   Lispy representation of class
  ;;       attr    String or NIL - the value of HTML class attribute.
  ;; NB: If attr constists of several CSS class names separated by spaces,
  ;;     we check for some (or none) of these names
  (let ((word-list nil))			; cache: splitted attr value or :none
    (labels ((%match-p (p)
             (cond ((or (eq p :all) (and (stringp p) (string= p "*")))
                    t)
                   ((or (null p) (eq p :none))
                    (or (null attr)
                        (and (stringp attr)
                             #1=(eq (orf word-list (or (yl:split-seq #\Space attr
                                                          :keep-empty-subseqs nil)
                                                       :none))
                                    :none))))
                             ;(string= attr +null-string+))))
                   ((or (keywordp p) (stringp p))		; attr must be non-empty
                    (and (stringp attr)				; string and some substr
                         (not #1#)
                         (loop for word in word-list
                               thereis (string-equal p word))))	;lw:find-regexp-in-string?
                   ((consp p)
                    (case (first p)
                      (not
                       (not (%match-p (second p))))
                      (and
                       (loop for p in (rest p) always (%match-p p)))
                      (or
                       (loop for p in (rest p) thereis (%match-p p)))
                      (otherwise
                       ;(template:signal-template-invocation-error)?
                       (log-warn #L"Bad CSS class pattern: ~s." pattern)) )) )))
      (%match-p pattern))))

(defun help-uri (link)
 ;;; Helper for converting string, possibly with anchor, to URI.
  ;; Args: link Flexible link string.
  ;;		If has neither "help:" at head nor the sharp char '#' inside,
  ;;		it is treated as a pure anchor, i.e. fragment only.
  ;; Value: URI with "help:" scheme by default.
  (cond ((uri:uri-p link)					; ready to use
         link)
        ((or (eql (search "help:" link) 0) (char-position #\# link))
         (let ((uri (uri:uri link)))		; full-fledged conversion
           (unless (uri:uri-scheme uri)
             (setf (uri:uri-scheme uri) :help)
             (uri:clear-cache uri))
           uri))
        (t						; "naked" string means fragment
         (uri:make-uri :scheme :help :fragment link)) ))

;(defun item-environment (base segment content &optional fragment class plist item suffix)
(defun %item-environment (item base segment fragment suffix plist
                          &optional (content (%anchor-content plist nil suffix))
                                    (class (getf plist :classes)))
 ;;; Helper
  ;; Args  item     Instance of heading, segment itself, pair (segment . plist) or NIL
  ;;       base     Referrer base uri
  ;;       segment  Instance of segment, topic, or text-topic
  ;;       fragment Heading-id or NIL
  ;;       suffix   If true, category is appended to item-content as
  ;;	   plist    Plist of heading, segment or anchor(?)
  ;;       content  Heading content, segment title, or anchor content
  ;;		    per (anchor-content ... :suffix t)
  (let ((uri (uri:enough-uri (asd:component-output-pathname segment) base)))
    (multiple-value-bind (span-class span-cont)
        (when suffix
          (get-properties plist *anchor-span-css-class-properties*))
      (when fragment					; do we need a fragment?
        (setf (uri:uri-fragment uri) fragment))
      `((item	      . ,item)
        (item-topic   . ,segment)
        (item-uri     . ,uri)
        (item-class   . ,(when-let (classes (if (stringp class)
                                                (intersection
                                                 (split-seq-if #'whitespace-char-p class)
                                                 *anchor-span-css-classes*
                                                 :test #'string-equal)
                                                class))
                           (interleave-seq 'string #\Space classes)))
        (item-title   . ,(title segment))
        (item-content . (:content
                         ,@content
                         ,@(when span-class
                             `(,#\Space ,(html:make-elem :span
                                                         (list :class span-class)
                                                         (listify span-cont))))))
        (item-description . ,(description segment))))))

(defgeneric item-environment (item base &optional fragment suffix)
 ;;; Helper
 (:method ((heading heading) base &optional (fragment (heading-id heading)) suffix)
  (%item-environment heading base (heading-segment heading)
                     fragment suffix (heading-plist heading)
                     (heading-content heading) (heading-class heading)))
 (:method ((segment segment) base &optional fragment suffix)
  (%item-environment segment base segment
                     fragment suffix (plist segment)
                     (list (title segment)) nil))
 (:method ((item topic-mixin) base &optional fragment suffix)
  (%item-environment item base item
                     fragment suffix nil))		;(asd::component-properties item)?
 (:method ((item cons) base &optional fragment suffix)
 ;; Q: Shall we take plist and content from the fisrt heading?
  (%item-environment item base (car item)
                     fragment suffix (cdr item)))
)

(defgeneric item-a (item &key class suffix)
 ;;; Value: LHTML content with A element like
  ;;		((:a :href uri :title title) text-or-content)
 (:method (item &key class suffix)
  (declare (ignore item class suffix))
  nil)
 (:method ((heading heading) &key (class nil supplied-p) (suffix t))
  (let* ((referrer ;(if (eq *current-topic* (start-topic *current-chapter*))
                   ;    *current-chapter* *current-topic*))
                   (nth-segment *current-segment* *current-topic*))
         (segment (heading-segment heading))		; target segment
         (uri (uri:enough-uri (asd:component-output-pathname segment)
                              (asd:component-output-pathname referrer))))
    (when-let (fragment (heading-id heading))
      (setf (uri:uri-fragment uri) fragment))
    (multiple-value-bind (span-class span-cont)		; suffix
        (when suffix
          (get-properties (heading-plist heading) *anchor-span-css-class-properties*))
      `(:content ((:a :href ,uri
                   ,@(when (orf class
                                (and (not supplied-p)	; not suppressed
                                     (when-let* ((class (heading-class heading))
                                                 (classes (if (stringp class)
                                                              (intersection
                                                               (split-seq-if
                                                           #'whitespace-char-p class)
                                                               *anchor-span-css-classes*
                                                               :test #'string-equal)
                                                              class)))
                                       (interleave-seq 'string #\Space classes))))
                       `(:class ,class))
                   :title ,(or (segment-property segment :description) (title segment)))
                  ,@(heading-content heading)
                  ,@(when span-class
                      `(,#\Space ,(html:make-elem :span
                                                  (list :class span-class)
                                                  (listify span-cont))))))))) )
  
(defgeneric %toc-children (referrer heading-classes source &optional depth suffix)
 ;;; Args: referrer Chapter or topic component (not a segment!)
 (:method ((module asd:module) heading-classes source &optional (depth 1) suffix)
  ;; Collecs over
  ;; - non-zero segments of the start topic,
  ;; - all segments of the all other topics.
  (when-let* ((start-topic (start-topic module))
              (self-level (if (eq source :title)
                              (segment-title-toc-level start-topic)
                              (heading-level (first (headings start-topic))))))
    (let ((base (asd:component-output-pathname module))
          (lo (1+ self-level))
          (up (+ self-level depth))
          (acc ()))
      (flet ((%collect-h (item fragment tag level)
               (when (and (<= lo level up)			; verify level and tag
                          (or (null tag) (null source)
                              (if (consp source)
                                  t				;(memq tag source)
                                  (eq tag source))))
                 (push (item-environment item base fragment suffix) acc))))
                       ;(item-environment segment base segment content fragment class plist
                       ;                  item suffix)
        (dolist (topic (remove start-topic (topics module)))
          (if (eq source :title)
              (do-segments (segment topic)
                (%collect-h segment nil nil (segment-title-toc-level segment)))
                            ;(list (title segment))
                            ;nil nil				; fragment class
                            ;(plist segment)))
              (dolist (heading (headings topic))
                (when (class-attr-match-p heading-classes (heading-class heading))
                  (%collect-h heading (heading-id heading)
                              (heading-tag heading) (heading-level heading)))))))
                              ;(heading-content heading)
                              ;(heading-class heading)
                              ;(heading-plist heading)
                              ;heading))))))
      ;; Inspect non-null segments of the start-topic
      (nreconc (%toc-children start-topic heading-classes source depth)
               acc))))

 (:method ((topic topic) heading-classes source &optional (depth 1) suffix)
  ;; Collecs over all segments of this topics
  (when-let (self-level (if (eq source :title)
                            (segment-title-toc-level topic)
                            (heading-level (first (headings topic)))))
    (let ((base (asd:component-output-pathname topic))
          (lo (1+ self-level))
          (up (+ self-level depth))
          (acc ()))
      (flet ((%collect-h (item fragment tag level)
               (when (and (<= lo level up)
                          (or (null tag) (null source)
                              (if (consp source)
                                  t				;(memq tag source)
                                  (eq tag source))))
                 (push (item-environment item base fragment suffix) acc))))
                       ;(item-environment base segment content fragment class plist
                       ;                  item suffix)
        (if (eq source :title)
            (do-segments (segment topic)
              (unless (eq segment topic)
                (%collect-h segment nil nil (segment-title-toc-level segment))))
                            ;(list (title segment))
                            ;nil nil				; fragment class
                            ;(plist segment)
                            ;segment)))
            (dolist (heading (rest (headings topic)))
              (when (class-attr-match-p heading-classes (heading-class heading))
                (%collect-h heading (heading-id heading)
                            (heading-tag heading) (heading-level heading))))))
                            ;(heading-content heading)
                            ;(heading-class heading)
                            ;(heading-plist heading)
                            ;heading)))))
      acc))) )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  API  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; CAUTION: Avoid spreading {form} in braces among several lines
;;;          as newlines are preserved during parse-html!

(defun toc-children (&key (heading-classes :all)
                          (source (if-bind (toc (find-toc *current-book*))
                                    (toc-source toc)
                                    *default-toc-source*))
                          (depth 1)
                          (referrer (if (eq *current-topic*
                                            (start-topic *current-chapter*))
                                        *current-chapter*
                                        *current-topic*))
                          (suffix t))
 ;;; Links to direct children that reside DEPTH levels deeper than the COMPONENT given.
  ;; Args: heading-classes
  ;; Value: Sequence of template environments used by TMPL_LOOP over of the form:
  ;;   ((item             . {heading | segment})	; abstract item
  ;;    (item-topic       . {topic | segment})		; target topic
  ;;	(item-uri         . relative-uri)
  ;;	(item-class       . CSS-class)			; class for element A
  ;;	(item-title       . segment/topic-title-text)
  ;;	(item-description . segment/topic-description) ; segment doesn't inherit from topic
  ;;	(item-content     . heading-inline-content))
  ;;   ...)))
  ;; Topics are sorted in "natural" order according to their appearence in the book.
  ;;
  ;; TODO: Add template variable item-level?
  ;; Q:    , ,   ?
  ;;          ,      
  ;;	toc-children    .
  (sort (%toc-children referrer heading-classes source depth suffix)
        #'segment<
        :key (lambda (alist) (cdr-assoq 'item-topic alist))))

#||
#+unused
(defun popup-menu-a (items &key href title class content (referrer *current-topic*))
 ;;; Produce an HTML element A with onclick attribute to pop-up a menu.
  ;; Args: items   List of anchors or sublists (anchor :text text :title title)
  ;;       href    The value of href attribute;
  ;;		   if null, the first item provides the default.
  ;;       title   The menu window title.
  ;;       class   CSS class name of the resulting A element.
  ;;       content LHTML list or string specifying the content of the resulting A element;
  ;;		   if null, the first item's anchor-text provides the default.
  ;; Value: LHTML content with A element like
  ;;	    ((:a :href href
  ;;	      :onclick "popup_menu(event,title,['item-uri','item-text',item-title,...])"))
  ;; BAD: There is no way to set segment property :popups (:menu) here.
  (let ((acc ()))
    (dolist (item items)
      (multiple-value-bind (uri attrs plist)
          (translate-uri (help-uri (first-or-self item)) nil referrer)
        (when uri ;(and (null attrs))
          (let ((text (or (and (consp item) (getf (rest item) :text))
                          (%anchor-text plist))))
            (unless href
              (setq href uri))
            (unless content
              (setq content text))
            (push (format nil +popup-menu-item-js+
                          (uri:render-uri uri)
                          (escape-ecma-string text)
                          (escape-ecma-string (or (and (consp item)
                                                       (getf (rest item) :title))
                                                  (getf plist :text)
                                                  (nth-value 1 (get-properties
                                                                attrs
                                                                '(:title :description))))))
                  acc)))))
    `(:content ((:a :href ,href
                    ,@(when class `(:class ,class))
                    :onclick ,(format nil +popup-menu-onclick-js+
                                      title (nreverse acc)))
                ,@(listify content)))))
||#
