;;; -*- Mode: Lisp; -*-
;;; YstokHelp - HTML-based help
;;; Copyright (c) 2009-2014 Dr. Dmitry Ivanov.  All rights reserved.
;;; Website: http://lisp.ystok.ru/yhelp/
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Book compiler specials and generics

;;; OPERATIONS
;;; :preprocess
;;;	Preparses topics: collect anchors, headings, segments.
;;;	Affects the image only, does not create any output files.
;;;	Q:  full parse ,  LHTML  ?
;;;	      "help:#anchor"    .
;;; :compile
;;;	Second parse of topics: generates pages and copies auxiliary files 
;;;	into the output directory.
;;; :load
;;;	Unused for now, not applicable as we never load compiled HTML.
;;;
;;; SUBBOOKS
;;; A subbook is implemented by a chapter, which is embedded into the current
;;; (parent) book directory from an external source by means of :pathname #P"..."
;;;
;;; TODO: EXTERNAL BOOK
;;; - May not be mentioned in :depends-on.
;;; - Must be registered before the current book system is compiled.
;;; - Any link to it should lead to opening a new browser window or tab.

(in-package :ystok.help)

;; Users of source code really need this for loading stopwords files
(unless (boundp 'yl::%thys-product-path%)
  (setq yl::%thys-product-path% (pathname-location #.(current-pathname))))

;; CAUTION: Avoid naming variables as URI or TEXT in non-preprocessor (run-time only)
;;          code because that could conflict with html:uri!
(eval-when (:compile-toplevel :load-toplevel :execute) 	; those used in templates
  (import '(html:icon html:icon-html html:text html:uri)))

(defconstant* +heading-keys+ #(:h1 :h2 :h3 :h4 :h5 :h6))

(defconstant* +segment-tag+ "HELP_SEGMENT")
(defconstant* +segment-file-format-string+ "~a_~2,'0d")

(defconstant* +popup-onclick-js+   "popup(event)")
(defconstant* +popup-onkeydown-js+ "popup_okd()")
(defconstant* +popup-onkeyup-js+   "popup_oku()")

;; popup_menu(event,'Menu caption',['using.html','Using help book',null,menuitem2])
(defconstant* +popup-menu-onclick-js+ "popup_menu(event,~@['~a',~][~{~a~^,~}])")
(defconstant* +popup-menu-item-js+ "'~a',~:[null~;~:*'~a'~],~:[null~;~:*'~a'~]")

(defconstant* +target-popup+ "_popup")		; accompanying hrefs by target="_popup"

;; Some authority tools add auxiliary subdirs of its own;
;; we strip them off from aux-folders.
(defparameter *exclude-subdirectories*
  '("_private" "_vti_cnf" "_vti_pvt"))			; FrontPage

;;; Anchors
;;;
;;; Simplification of the content of an anchor element retains only the following
;;; inline elements:
;;; - B CITE CODE EM I KBD SAMP U VAR (what about FONT STRONG SUB SUP?)
;;; - SPAN class="name1 name2..."
;;;   when some CSS class name strings is in *anchor-span-css-classes* (case-insensitive);
;;; - SPAN style="prop1: value1; ..."
;;;   when some CSS property keyword is in *anchor-span-style-properties*

(defparameter *anchor-inline-tags* '(:b :cite :code :em :i :kbd :samp :strong :u :var))
(defparameter *anchor-span-style-properties* '(:font-variant))

;; Such CSS class names are also pushed into
;; - the :classes property of an anchor (to be printed in Index),
;; - the class slot of heading (to be printed into TOC).
(defparameter *anchor-span-css-classes*
  '("code" "small-caps" "menu-command" "widget" "yh_using"))

;; CSS-classes of the inner SPAN elements that are moved the the very end of content
;; (actually dropped from content but converted to property)
(defparameter *anchor-span-css-class-properties* '(:category))	;:widget

;;; Computing a default of the definition property.
(defparameter *anchor-definition-tags* '(:dfn :dt))
(defparameter *anchor-definition-css-classes* '("signature"))

;; Default values for the bbcode property of anchors and headings
(defparameter *bbcode-in-keywords* t)
(defparameter *bbcode-in-text* t)

;; If true, should be a tag keyword with/without attributes/content;
;; the whole element A is replaced by this LHTML element.
;; If null, remove A and leave content only without any indication.
;; The former is useful during developing, the latter - for the final help book.
(defparameter *undefined-anchor-tag* '(:span :class "yh_undefined_anchor"
                                       :title "Undefined anchor"))

;;; Index compiler

;; Determines whether we must register <Hi id=...> in anchors table. One of
;;  NIL       Do not register
;;  :keywords Register and put explicit keywords into index automatically
;;	      but do not put text there.
;;  :text     Register and put into index either
;;	      - the explicit keywords when the :keywords property is present,
;;	      - or all words mentioned in the :text property.
;;  T (true)  Look to what is specified in the following order:
;;            - the :keywords property,
;;            - the :text property,
;;            - the element content.
;; In all cases, the explicit "keywords:none" property is specified, neither
;; the :text property nor the content of the heading will be indexed.
(defparameter *index-heading-anchors* t)

;; max number of chars in the left column of a permuted index
(defparameter *default-index-max-left-chars* 15)

;; CSS class name for master table in index
(defparameter *index-table-css-class* "idx")

;;; TOC compiler

(defparameter *default-toc-source* '(:h1 :h2))
(defparameter *default-toc-max-level* 3)

;; CSS class name for BODY element in TOC
(defparameter *toc-body-css-class* "toc")

;; Exported-from ylib: keyword extracted from the language slot of the book-system
;(defvar *language* nil)

;;; Specials that
;;; - indicate current patht,
;;; - are bound during parse-html as used in templates.

(defvar *current-book* nil)		; instance of book-system (to be) compiled
(defvar *current-chapter*)		; also used in templates
(defvar *current-topic*)		; topic, navigation, or popup.

;; The following two are bound within
;;  :preprocess - to NIL,
;;  :compile    - to 0, then increased.
(defvar *current-segment*)		; position in (segments *current-topic*)
(defvar *current-heading*)		; position in (headings *current-topic*)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  HELPERS  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun log-info (&rest args)
  (apply #'logg :help args))
(defun log-warn (&rest args)
  (apply #'logg :warn args))

(defun log-copy-file (input-file output-file)
  (logg :help "copy: ~a -> ~a" input-file output-file)
  (copy-file input-file output-file))

(declaim (inline nsubseq))
(defun nsubseq (sequence start &optional (end (length sequence)))
  "Like SUBSEQ but the result shares structure with SEQUENCE."
  (make-array (- end start)
              :element-type (array-element-type sequence)
              :displaced-to sequence
              :displaced-index-offset start))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  COMPONENT-BASED  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric topics (component)
 ;;; List of topics that generate linear page sequence
  ;; It contains only featured topics, i.e. those satisfying the features slot value!
  ;; Args: component For module, collect topics of submodules
 (:method (object)
  ())
 (:method ((module asd:module))
  (if (asd:component-featurep module)
      (mapcan #'topics (asd:module-components module))
      ())) )

(defgeneric rel-object (self rel)
 ;;; Topic and segments relationships within final page sequence
 (:method (self rel)
  nil))				; by default, nothing is in relation rel w.r.t. self

(defgeneric module-component< (module x y)
 (:method ((module asd:module) x y)
  (let ((seq (asd:module-components module)))
    (< (position x seq) (position y seq)))))

(defun component-path (component)
 ;;; Value: Absolute path through logical component hierarchy, i.e. the list like
  ;;		(book chapter1 chapter12 ... component)
  (if component
      (nconc (component-path (asd:component-parent component)) (list component))
      ()))

(defun component-path< (x y)
  (do ((parent nil)
       x-first y-first)
      ((not (and x y)) y)		; chapter preceeds topic
   (if (eq (setq x-first (first x)) (setq y-first (first y)))
       (setq parent x-first
             x (rest x)  y (rest y))
       (return (module-component< parent x-first y-first)))))

(defgeneric source-component (self)
 ;;; An instance of asd:component representing an entire source file
 (:method ((self asd:source-file))
  self))

(defgeneric dump (self)
 ;;; Return true if dump-anchors must save anchors resided inside the component.
  ;; For module, the returned value is treated as a default for its components.
 (:method (self) nil))

(defun siblings (component
                 &optional (type '(or chapter (and topic (not text-topic)) popup)))
 ;;; Return children of the parent of the component given excluding itself.
  ;; Args: type Filter the resulting list by components of this type only.
  ;;            The default excludes navigation instances.
  ;; Value: List sorted "naturally". according to the parent module defintion.
  (let ((acc (remove component (asd:module-components (asd:component-parent component)))))
    (if type
        (remove type acc :test-not #'xtypep)
        acc)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  LOGICAL PATHNAMES  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ASDlite allows any component accept logical pathname as the :pathname initarg
;;; CAUTION: Underscore '_' is treated differently from the "usual" pathname.
;;;	     It is not allowed in the name part of a logical pathname!

(defun %directory-translations (directory-pathname)
  (let ((wild (merge-pathnames (load-time-value		; #P"E:/Lisp/Projects/**/*.*"
                                (make-pathname :name :wild :type :wild
                                               :directory #1='(:relative :wild-inferiors)))
                               directory-pathname)))
    `(("**;*" ,(merge-pathnames (load-time-value	; #P"E:/Lisp/Projects/**/*"
                                 (make-pathname :name :wild :directory #1#))
                                directory-pathname))
      ("**;*.*" ,wild)
      (";**;*.*" ,wild))))

(defmacro with-logical-hosts (pairs &body body)
 "Bind hosts with the new translations mentioned in pairs, evaluate the body,
then restore old translations for the hosts.
Arguments:
 pairs List of two-element lists, evaluated."
  (with-gensyms (new old pair)
    `(let* ((,new ,pairs)
            (,old (loop for ,pair in ,new
                        for logical-pathname-translations
                            = (ignore-errors (logical-pathname-translations (first ,pair)))
                        when logical-pathname-translations	; is about to rebind
                        collect (cons (first ,pair) logical-pathname-translations))))
       (unwind-protect (progn
                         (dolist (,pair ,new)			; assign new translations
                           (setf (logical-pathname-translations (first ,pair))
                                 (%directory-translations (second ,pair))))
                         ,@body)
         (dolist (,pair ,old)					; restore old translations
           (setf (logical-pathname-translations (car ,pair)) (cdr ,pair)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  LOCALIZATION  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *yhelp-bundle* #-ys-product (yl:make-bundle)
                       #+ys-product (ys::find-module :yhelp))

(yl:define-label-language *yhelp-bundle* :en :ascii)

;;; Parsing (fragment-char-p) and index compiler

(defgeneric lang-alpha-char-p-function (language)
 ;;; Args: langage Keyword
  ;; Value: Function of signature: (character) => boolean
 (:method (language)
  (declare (ignore language))
  #'latin-1-alpha-char-p))

(defgeneric lang-shorten-span (language)
 ;;; Facilitate removing exsessive SPAN elements:
  ;;	((:SPAN :LANG "primary-sub") . content) -> content
  ;; Value: Either a language tags as of the lang attribute value,
  ;;        or list of tags,
  ;;        or T meaning any language code.
 (:method (lang)
  (declare (ignore lang))
  nil))


;;;;;;;;;;;;;;;;;;  STOPWORDS - words that are never indexed  ;;;;;;;;;;;;;;;;;;
;;; Stopwords files are
;;;  - placed into lang/ subdirectory of the product-pathname directory,
;;;  - named stopwords-{language}.txt
;;; NB: These files are considered components of the YstokHelp system rather than a book!

;; List of tuples: (lang parsed-charset . timestamp)
;; Args: lang Language tag; if contains subtags, its primary language is also loaded.
;; CAUTION: In the contrast to bundle-languages, we store timestamp in cddr, not in caddr.
(defvar *stopword-languages* '((:en :ascii)))

;; Stopwords in all languages are loaded into the single hash table.
;; Each stopword is bound to the corresponding language keyword.
(defvar *stopword-hash-table* (make-hash-table :test #'equal))

(defun load-stopwords-file (pathname &key (language t) (external-format :default))
  (log-info "Loading stopwords file ~a" pathname)
  (with-open-file (stream pathname :direction :input
                          :element-type #-lispworks 'character
                                        #+lispworks 'lw:simple-char
                          :external-format external-format)
    (loop for word = (read-line stream nil nil)
          while word
          when (plusp (length (setq word (string-trim '(#\Space #\Tab #\Newline #\Return)
                                                      word))))
          do (setf (gethash word *stopword-hash-table*) language))))

(defun ensure-stopwords (lang)
 ;;; Load stopword files for the lang itself and its primary language if needed.
  ;; Args: lang Language tag keyword
  (dolist (lang (language-primary-list lang))
    (let* ((tuple (assoc lang *stopword-languages*))
           (pathname (yl:labels-pathname :stopwords lang "txt"))
           (timestamp (asd::safe-file-write-date pathname)))
      (cond (tuple
             (unless (and (cddr tuple) (<= timestamp (cddr tuple)))
               (when (cddr tuple)
                 ;; File was loaded then changed - remove all words from the hash-table
                 (loop for word being the hash-key of *stopword-hash-table*
                       using (hash-value value)
                       when (eq value lang)
                       do (remhash word *stopword-hash-table*)))
               (load-stopwords-file pathname :language lang
                                    :external-format (html:charset-ef (second tuple)))
               (setf (cddr tuple) timestamp)))
            (t
             (load-stopwords-file pathname :language lang)
             (push `(,lang :default . ,timestamp) *stopword-languages*))))))

;;; Index compiler

;;; WORDS: Utility to find word boundaries a la this
;;; (?i) - case-insensitive
;;; (?<=\\W) A zero-width positive look-behind assertion, ..  non-alphanumericp
;;;            non-alphanumericp   
;;; \\b ::= \w\W or \W\w -  
;;; (let ((result (list 0)))
;;;    (do-matches (start end "(?i)(?<=\\W)\\b" symbol)
;;;      (declare (ignore end))
;;;      (push start result))
;;;    (nreverse result)))

(defgeneric lang-word-char-p-function (langauge)
 ;;; Predicate to accept characters in addition to alphanumericp and _
  ;; Args: lang Keyword
 (:method (lang)
  (declare (ignore lang))
  #'latin-1-alpha-char-p))

(defgeneric lang-default-index-letters (langauge)
 ;;; Used as a default when the letters slot in null and supplies the value
  ;; for %index-letters%
 (:method (lang)
  (declare (ignore lang))
  "ABCDEFGHIJKLMNOPQRSTUVWXYZ*"))
 
(defgeneric lang-char-representative (language char)
 ;;; Representative is a character that implement upper- and lowercase variants,
  ;; sometimes for several letters.
  ;; Value: NIL if the char has not got a distinctive representative.
 (:method (lang char)
  (declare (ignore lang))
  (cond ((alpha-char-p char)
         (char-upcase char))
        ((latin-1-alpha-char-p char)
         (locale-char-upcase char)))))


#||;;;;;;;;;;;;;;;;;;  LABELS - phrases on the book pages  ;;;;;;;;;;;;;;;;;;
;;; Labels files are
;;;  - placed into lang/ subdirectory of the product-pathname directory,
;;;  - named labels-{language}.lisp,
;;;  - loaded by means of read in loop.

(defun lang-file-pathname (lang kind &optional (type "txt"))
 ;;; Args: lang - main language keyword
  ;;       kind ::= :stopwords or :labels
  (make-pathname :name (string-append (string-downcase kind)
                                      #\-
                                      (string-downcase lang))
                 :type type 
                 :defaults (product-pathname "lang/")))

;; List of tuples: (language parsed-charset . timestamp)
(defvar *label-languages* '((:en :ascii)))

;; Labels are stored in the forest of "binary" trees, one tree per master language
;; ((:en (:navigation (:toggle (:title . "..."))
;;                    (:sync   (:title . "...")))
;;       (:prev . "...")
;;       (:up   . "...")
;;       (:next . "...")
;;       ...))
(defvar *label-forest* ())

(defun load-labels-file (pathname lang &key (external-format :default))
 ;;; Value: t = success
  (log-info "Loading labels file ~a" pathname)
  (let ((*read-eval* nil)
        (*read-suppress* nil)
        (tree (assoq lang *label-forest*))
        line)
    (with-open-file (stream pathname :direction :input
                            :element-type #-lispworks 'character
                                          #+lispworks 'lw:simple-char
                            :external-format external-format)
      (if tree
          (rplacd tree ())			; cut the old tree
          (setq tree (list lang)
                *label-forest* (cons tree *label-forest*)))
      (labels ((%put (self tree path value)
                 #+debug (assert (consp tree))
                 (let ((alist (cdr tree))
                       (first (first-or-self path))
                       (rest (if (consp path) (rest path) nil)))
                   (if (atom alist)
                       (cond (path
                              (when (and self alist)	; replace atom with ((nil . atom))
                                (rplacd tree (setq alist (acons nil alist nil))))
                              (let ((tuple (list first)))
                                (rplacd tree (cons tuple alist))
                                (%put first tuple rest value)))
                             (alist
                              (warn "Duplicated language label: ~s" line))
                             (self
                              (rplacd tree value)))
                       ;; (consp alist)
                       (let ((tuple (assoq first alist)))
                         (unless tuple
                           (rplacd tree (cons (setq tuple (list first)) alist)))
                         (%put first tuple rest value))))))
        (while (setq line (read stream nil nil))
          (%put nil tree (car line) (cdr line))))
      t)))

(defun ensure-labels (lang)
 ;;; Args: language Master language keyword
  ;; Value: True = success
  (let* ((tuple (assoq lang *label-languages*))
         (pathname (lang-file-pathname lang :labels "lisp"))
         (timestamp (asd::safe-file-write-date pathname)))
      (cond (tuple
             (cond ((and (cddr tuple) (<= timestamp (cddr tuple)))
                    timestamp)
                   ((load-labels-file pathname lang
                                      :external-format (html:charset-ef (second tuple)))
                    (setf (cddr tuple) timestamp))))
            ((load-labels-file pathname lang)
             (push `(,lang :default . ,timestamp) *label-languages*)
             timestamp))))

(defun get-label (path language)
 ;;; Value: String or NIL
  (flet ((%not-found (path)
           ;; In case there is no match, output seqeunce starting from an unknown key
           (if (consp path)
               (interleave-seq 'string #\Space (mapcar #'string-capitalize path))
               (string-capitalize path))))
    (if-bind (lang (cdr-assoq3 language *sublanguage-alist* language))
      (when (or (cddr (assoq lang *label-languages*))	; already loaded
                (ensure-labels lang))
        (when-let (tree (assoq lang *label-forest*))
          (labels ((%get (self tree path)
                     #+debug (assert (consp tree))
                     (let ((alist (cdr tree)))
                       (if (atom alist)
                           (cond (path (%not-found path))
                                 (alist)					; success!
                                 (self (%not-found self)))
                           (let* ((first (first-or-self path))
                                  (rest (if (consp path) (rest path) nil))
                                  (tuple (assoq first alist)))
                             (cond (tuple (%get first tuple rest))
                                   (rest (%not-found path))
                                   (first (%not-found first))
                                   (self (%not-found self))))))))
            (%get nil tree path))))
    (%not-found path))))
||#               
