;;; -*- Mode: Lisp; -*-
;;; YstokHelp - HTML-based help
;;; Copyright (c) 2009-2014 Dr. Dmitry Ivanov.  All rights reserved.
;;; Website: http://lisp.ystok.ru/yhelp/
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Basics used by both compiler and run-time API

(in-package :ystok.help)

(defparameter *anchor-file-type* "yha")

;; List of book instances, looked up by id (not coerced name), :test #'string-equal
(defvar *registered-books* ())

;; Chapter can return the start topic of the parent if toc-level=-1
(defgeneric start-topic (module))

(define-condition simple-help-error (simple-error)
  #+ywidgets ((ywi::icon :initform :query)))

;; ANCHORS SLOT
;; In compiler, map: anchor -> (({segment | topic} . plist) ...)
;; where plist collects
;;  (1)	:content  LHTML list extracted from the content of an element.
;;  (2)	:text	  String specified explicitly within the help attribute.
;;  (3) :keywords Phrases specified explicitly within the help attribute;
;;		  if :none, index entries for the anchor are not generated.
;;  (4)	:bbcode   If specified explicitly withing the help attribute and true, both
;;		  the :text and :keywords properties are considered with BBcode markup;
;;		  if equals to :none, these properties are plain strings.
;;		  If not specified, *bbcode-in-text* and *bbcode-in-keywords* are used
;;		  accordingly.
;; TODO :page	  PDF page number.
;;
;; In run-time, map: anchor -> {uri-path | ((uri-path . title) ...)}
;;
;; Text Anchors
;; In compiler, map: anchor -> ((context-popup start . end) ...)
;; In run-time, map: anchor -> ((uri-path start . end) ...)
;;
;; The output-html-format
;;   book-system - computed from charset if null
;;   book        - hard-wired into the application or read from the anchors file
;;
;; NB: Specifying the :external-format and :output-html-format initargs are useful
;;     for a book. For a book-system, they are usually calculated from the charset slot.

(defclass book-mixin ()
 ((id :reader id :initarg :id)
  (title :reader title :initarg :title		; book title for <title> and help-menu item
	:initform nil)
  (external-format :reader external-format	; universal external format for
	:initarg :external-format		; - input HTML
        :initform :default)			; - input and output anchor/context files
  (output-html-format :reader output-html-format ; for resulting HTML pages of the book
	:initarg :output-html-format)
  (anchors      :accessor anchors       :initform nil)	 ; assined to hash-table in
  (text-anchors :accessor text-anchors  :initform nil))) ; reinitialize-instance/load-book

;;; Off-the-shelf book
;;; 1) registered in the application defsys file and invoked in run-time,
;;; 2) registered in the definition file of a "master" book system for convetring links.

(defclass book (book-mixin)
 ((description :reader description :initarg :description :initform nil)
  (start-topic :reader start-topic :initarg :start  :initform "help.html") ; namestring
  ;; Subpath relative to *default-book-path* replaced by an absolute path in run-time,
  (location :accessor location :initarg :location :initform "help/")
  ;; If specified, should be an absolute URI of the book located in the Internet
  (book-uri :accessor book-uri :initarg :uri :initform nil)))

(defmethod print-object ((self book) stream)
  (if *print-escape*
      (print-unreadable-object (self stream :type nil)
        (format stream "BOOK ~s ~a" (id self) (title self)))
      (format stream "~s~% ~a" (id self) (title self))))

(defmethod start-topic-name ((book book))
  (start-topic book))

;; Merged pathname or uri - the final location of the component page or file
(defgeneric absolute-uri (book component))

(defun register-book (id &rest args)
  "Create (optionally) and register a help book instance.
Arguments:
 id   - Book-mixin instance or book id (usually when called in defsys.lisp).
 args - Keyword argument list passed to make-instance.
Value:
 Existing instance of book-mixin or newly created book instance."
  (let* ((book (if (typep id 'book-mixin)
                   id
                   (apply #'make-instance 'book :id id args)))
         (id (id book)))
    (if-bind (rest (member id *registered-books* :key #'id :test #'string-equal))
      (unless (eq (first rest) book)
        (warn "Reregistering book: ~s -> ~s." (first rest) book)	; logg :warn
        (rplaca rest book))
      (push book *registered-books*))
    book))

(defgeneric find-book (id &optional errorp)
 ;;; Q: Run-time or parse/deliver-time search?
 (:method (id &optional (errorp t))
  (if errorp
      #1=(cerror #L"Ignore context help or book topic request."
                 'simple-help-error
                 :format-control #L"Help book is not found~@[: ~a~]."
                 :format-arguments (list id))
      nil))
 (:method ((self book-mixin) &optional errorp)
  (declare (ignore errorp))
  self)
 (:method ((id string) &optional (errorp t))
  (cond ((find id *registered-books* :key #'id :test #'string-equal))
        (errorp #1#))) )

(defgeneric ensure-book (object))

(defun latin-1-alpha-char-p (char)
  (declare (character char)
           #-debug (optimize (speed 3) (safety 0) #+lispworks (hcl:fixnum-safety 0)))
  (let ((code (char-code char)))
    (declare (fixnum code))
    (or (<= #xC0 code #xD6)
        (<= #xD8 code #xF6)
        (<= #xF8 code #xFF))))

