;;; -*- Mode: Lisp; -*-
;;; YstokHelp - HTML-based help
;;; Copyright (c) 2009-2014 Dr. Dmitry Ivanov.  All rights reserved.
;;; Website: http://lisp.ystok.ru/yhelp/
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; CAPI/YstokWidgets stuff

;;; CAPI widget anchor computation, menu construction and settings
;;; For a widget, returns the anchor chain in the order of relevance.
;;; By default, the anchor of a top-level interface finishes the chain.

(in-package :ystok.help)

(defconstant* +missing-context+
              #L"Context help is unavailable for the element~@[~%'~a'~].")

(defvar *show-what-is-this-mi* t "Show What's This? command in context menus.")

;; NIL = do nothing/beep, :tooltip = show tooltip/"unavailable", :topic = view book topic
(defvar *if-context-missing* :tooltip "Call view-topic if there is no context info.")

(defparameter *topic-file-types* '("html" "htm"))
(defparameter *context-file-types* '("txt"))

(defun help-key-id (help-key &optional context)
  (cond ((consp help-key)
         (if context
             (let ((rest (cdr help-key)))
               (if (consp rest)
                   (car rest)		; (topic-id context-id . tooltip)
                   (car help-key)))	; (topic-id . tooltip) => context-id = topic-id
             (car help-key)))
        ((not (stringp help-key))	; not a tooltip
         help-key)))

(defgeneric help-locator (object &optional context)
 (:documentation "Args: object - CAPI object.
Value: Either list designators of HTML/text resources,
       or immediate context help string.")
 (:method (object &optional context)
  (declare (ignore object context))
  nil))

;; WORKAROUND: LW4.4 works on capi:menu but returns nil on capi:menu-component/item
(let ((lw:*handle-warn-on-redefinition* #+lispworks4 :quiet #-lispworks4 :warn))
 (defmethod CAPI:TOP-LEVEL-INTERFACE ((object capi:menu-component))
   (capi:top-level-interface (capi:element-parent object)))
 (defmethod CAPI:TOP-LEVEL-INTERFACE ((object capi:menu-item))
   (capi:top-level-interface (capi:item-collection object))))

(defun %pane-locator (pane arg &optional context fragment)
 ;;; Args: arg      Part of the pane's help-key
  ;;	   fragment If true, the result is going to be embedded into a fragment,
  ;;		    so we should not result in path!
  ;; Value: string, symbol, or list of such (may be NIL)
  ;; CAUTION: CAPI built-ins interpret help-key=T as :name, i.e. use capi-object-name!
  ;; Q: Define "reasonble parent" concept based on :parent, :collection, and :group?
  (case arg
    ((t :default)					; :default from where?
     nil)
    (:name
     (capi:capi-object-name pane))			; name symbol must be kept
     ;(when-let (name (capi:capi-object-name pane))	; name symbol must be kept
     ;  (string name)))
    (:interface
     (let ((interface (capi:top-level-interface pane)))
       (if (and fragment (consp (capi:help-key interface)))
           (type-of interface)				; path unacceptable-coerce to :type
           (help-locator interface context))))
    (:package
     (package-name (symbol-package (capi:capi-object-name pane)))) ; uppercased!
    ;(:parent						; due to intermediate
    ; (help-locator (capi:element-parent pane)))	; layouts, not very useful
    (:collection
     (help-locator (capi:item-collection pane) context)); 
    (:group
     (let* ((group (CAPI::BUTTON-BUTTON-GROUP pane))	; for check/radio-button
            (value (help-locator group context)))	; instead of useless
       (if (and group context fragment)			; context string
                ;(eq (first-or-self value) ?group-items-only?))
           (capi:capi-object-name group)		; return group name
           value)))
    (:tab						; ys-product only
     (let ((object (ywi::ambient-layout-visible-child pane))) ; both tab- and switchable-
       (help-locator (if (eq object pane) nil object)	; prevent recursion
                     context)))
    (:frame						; ys-product only
     (help-locator (ywi::%pane-frame pane) context))
    (:type						; useful for top-level interfaces
     (type-of pane))					; type symbol name must be kept
    (otherwise
     (when (eq arg :data)
       (setq arg (ywi:object-data pane)))
     (cond ((stringp arg) arg)
           ((numberp arg) (princ-to-string arg))
           ((symbolp arg) arg)))))			; symbol name must be kept

(defun %union-locator (&rest args)
 ;;; Flatten and append arguments excluding duplicates
  (remove-duplicates (remove nil (yl:mappend #'listify args)
                     :test #'string= :from-end t)))

(defun coerce-to-anchor (arg)
 ;;; Map a symbol encountered in ID to an HTML anchor
  (cond ((stringp arg) arg)
        ((symbolp arg) (substitute #\_ #\* (symbol-name arg)))))	; NIL -> "NIL"

(defun coerce-path-component (arg)
 ;;; A la asd::coerce-name
  (cond ((stringp arg) arg)
        ((and arg (symbolp arg)) (string-downcase (symbol-name arg)))))

(defun %locator (arg %atom-locator &optional context)
 ;;; Args: arg		 Designator; list designates path, vector - fragment.
  ;;       %atom-locator Function: (designator &optional in-fragment-p)
  ;;				-> { string | symbol | NIL }
  ;; Value: string or NIL
  ;; NB:   path: symbol -> string-downcase (a la ASDF),
  ;;       fragment: symbol -> symbol-name (as is with * replaced by _).
  (cond ((consp arg)					; ([:book book-id] path-comp...)
         (let* ((book-id (if (eq (first arg) :book)
                             (prog1 (second arg) (setq arg (cddr arg)))
                             nil))
                (atom (last arg 0))				; last after dot
                (type (when (and (or (stringp atom) (symbolp atom))
                                 (member atom
                                         (if context
                                             *context-file-types*
                                             *topic-file-types*)
                                         :test #'string-equal))
                        (coerce-path-component atom)))		; last is type
                (path (interleave-seq 'string #\/
                        (mapcan (lambda (arg)			; downcase symbol
                                  (when (setq arg (coerce-path-component
                                                    (funcall %atom-locator arg)))
                                    (list arg)))
                                (if atom (butlast arg 0) arg)))))
                        #|(remove nil (mapcar (lambda (arg)
                                               (first-or-self (funcall %atom-locator arg)))
                                              (if atom (butlast arg 0) arg)))|#
           (string-append* (append (when book-id
                                     (list book-id))
                                   (list #\/ path)
                                   (cond (type
                                          (list #\. type))
                                         ((and atom		; last is a fragment
                                               (setq atom (funcall %atom-locator atom t)))
                                          (list "/#" (coerce-to-anchor atom)))
                                         (t (list #\/)))))))
        ((setq arg (funcall %atom-locator arg t))
         (coerce-to-anchor arg))))

(defun %object-locator (object &optional context (help-key (capi:help-key object))
                                         top-level)
 ;;; Convert atom (excluding tooltip string), list, or vector representation to string
  ;; vector: #(a b c) -> "a_b_c"
  ;; atom:   help-key -> (%pane-locator help-key atom)
  ;; list:
  ;;	([:book book-id] a b)		 => "[book-id]/a/b/"	 	; directory assumed
  ;; 	([:book book-id] a b . :html)	 => "[book-id]/a/b.html"	; file, downcased
  ;; 	([:book book-id] "path" . #(c "d"))=> "[book-id]/path/#C_d"
  ;; 	([:book book-id] "path" . symbol)=> "[book-id]/path/#SYMBOL"	; symbol-name as is
  ;; 	([:book book-id] "path" . "atom")=> "[book-id]/path/#atom"
  ;; Value: List of strings (locator1 locator2 ...)
  ;; NB: When object is a top-level interface, help-key can only be a topic designator
  ;; Q: Is there any sense to union with interface locator for context?
  (if (eq help-key :parent)
      (help-locator (capi:element-parent object) context)
   (labels ((%atom-locator (arg &optional fragment)	; -> string or symbol (m.b. NIL)
              (cond ((stringp arg)
                     arg)
                    ((numberp arg) 
                     (princ-to-string arg))
                    ((and arg (symbolp arg))
                     (let ((value (first-or-self (%pane-locator object arg
                                                                context fragment))))
                       (if fragment (coerce-to-anchor value) value)))
                    ((vectorp arg)
                     (interleave-seq 'string #\_		; symbol names
                                     (remove nil (map 'list
                                                      #'(lambda (arg)
                                                          (%atom-locator arg t))
                                                      arg)))))))
     (if top-level
         ;; help-key is designator, not tooltip
         (%union-locator (%locator help-key #'%atom-locator context))
         (%union-locator (cond ((consp help-key)
                                (%locator (help-key-id help-key context)
                                          #'%atom-locator context))
                               ((not (stringp help-key))	; not a tooltip
                                (when-let (atom (%atom-locator help-key))
                                  (string atom))))
                         (help-locator (capi:top-level-interface object) context))))))
                         ;(mapcar (lambda (arg) (%locator arg #'%atom-locator context))
                         ;        (interface-default-locator
                         ;         (capi:top-level-interface object))))))))

(defmethod help-locator ((object capi:capi-object) &optional context)
  (%object-locator object context))

(defmethod help-locator ((interface capi:interface) &optional context)
 ;;; Like for pane but without tooltip or defaulting
  (%object-locator interface context
                   (capi:help-key interface)
                   (capi:top-level-interface-p interface)))

;;; Defaulting parent

(defmethod help-locator ((object capi:item) &optional context)
 ;;; Arg: object Instance of menu-item, item-pinboard-object, or stand-alone user-button.
  (let ((help-key (capi:help-key object))
        parent)
    (if (and (or (null help-key) (stringp help-key))		; own anchor not specified
             (setq parent (capi:item-collection object)))	; and collection exits
        (help-locator parent context)				; - borrow from there
        (call-next-method))))

(defmethod help-locator ((object CAPI::SIMPLE-BUTTON) &optional context)
 ;;; Arg: object Button in a button group, not an instance of capi:item!
  (let ((help-key (capi:help-key object))
        parent)
    (if (and (or (null help-key) (stringp help-key) (eq help-key t)) ; own anchor unspecif.
             (setq parent (CAPI::BUTTON-BUTTON-GROUP object)))	     ; and group exists -
        (help-locator parent context)  				     ; borrow from there
        (call-next-method))))					     ; -> on item

(defmethod ywi:make-editing-pane :around (type pinboard displayer)
 ;;; For pane within editing-pinboard, redirect help anchor requests to parent
  ;; by setting help-key to :parent or (:parent . column-help-key)
  (declare (ignore type pinboard displayer))
  (let* ((pane (call-next-method))
         (help-key (capi:help-key pane)))
    (when (eq help-key t)
      (setf (slot-value pane 'capi:help-key) :parent))
    pane))

;;; Defaulting child

(defmethod help-locator ((object capi:tab-layout) &optional context)
  (when-let (child (capi:tab-layout-visible-child object))
    (help-locator child context)))
(defmethod help-locator ((object capi:switchable-layout) &optional context)
  (when-let (child (capi:switchable-layout-visible-child object))
    (help-locator child context)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  HELP-POPUP-INTERFACE  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Implementation of "What's this?" pop-up window
;;; As popup does not have top-level-hook,
;;; - avoid errors as much as possible,
;;; - or execute with the owner interface process.
;;;
;;; Difference from tooltip:
;;; 1) displays an additional '?' button,
;;; 2) captures focus
;;; 3) does not return focus to the origin CAPI element even if Cancel button is pressed.
;;;    tooltip:
;;; 1)  . ,
;;; 2)  ,
;;; 3)      ,    Cancel
;;;
;;; BAD:    layout   :background 'win32:color_infobk -
;;;	 , dithered-     256.

(gp:register-image-translation 'book
  #.(gp:read-external-image (lw:current-pathname "images/book-dark.bmp"); or book-light
                            :transparent-color-index 13))

(capi:define-interface help-popup-interface (ywi:popup)
 ((topic :initarg :topic :initform nil)			; topic locator
  (ywi:chooser :reader ywi:chooser :initarg :chooser :initform nil) ; pane or pinboard
  (dx :initarg :dx :initform nil)			; offsets from chooser's top left
  (dy :initarg :dy :initform nil))			; received by popup-context
 (:panes
  (message capi:title-pane
	#+win32 #+win32 :foreground 'win32:color_infotext
  	#+win32 #+win32 :background 'win32:color_infobk	; works fine
        :visible-max-width t
	:visible-max-height t)
  (more-button capi:push-button :default-p t :data nil
	:image 'book	;:text ">>" :visible-max-width 26 ;:text-width
	:callback '%view-topic
	:callback-type :interface
	:help-key nil))
 (:layouts
  (main-layout capi:row-layout '(message ywi::cancel-button more-button)
	:adjust :bottom
	:internal-border ywi:+padding+
	#+win32 #+win32 :background 'win32:color_infobk	; doesn't work in 256-color mode
		;(color:get-color-spec 'win32:color_infobk)
		;'win32:color_3dface #(:RGB 0.5 .5 .5) :yellow	; works fine
	:gap #.(/ ywi:+h-gap+ 2)))	; as there one invisible button in the interim
 (:default-initargs
  :layout 'main-layout
  :internal-border 1)
  ;:enable-tooltips nil
  ;:help-callback '%view-topic)				; is not invoked on pressing F1
 (:extra-initargs '(:text)))


(defmethod shared-initialize-interface ((self help-popup-interface)
                                        &key topic ((:text %text%) nil)
                                        &allow-other-keys)
 ;;; Args: text String or list of strings representing context help text.
  ;; NB: This code cannot be placed into shared-initialize :after as it is called
  ;;     before creating panes
  (with-slots (main-layout message more-button) self
    (setf (capi:title-pane-text message) (if (stringp %text%)
                                             %text%
                                             (interleave-seq 'string #\Newline %text%))
          (ywi:enabled more-button) topic)
    #|(if (capi::pane-created-p more-button)
        (setf (ywi:shown more-button) topic)
        ;; BAD: The button representation looses image on removing/restoring in the layout
        (if topic
            (lw:push-end-new more-button (capi:layout-description main-layout))
            (removef (capi:layout-description main-layout) more-button)))|#
    (let ((cols (ywi:string-hint-length
                 (if (stringp %text%)
                     (ywi::max-width-in-characters %text%)
                     (loop for elt in %text% maximize (length elt)))))
          (rows (if (stringp %text%)
                    (ywi::max-height-in-characters %text%)
                    (length %text%))))
      (capi:set-hint-table message
       `(:visible-min-width (:character ,cols)
         :visible-min-height (:character ,(max 2 rows)))))))

(defmethod initialize-instance :after ((self help-popup-interface) &rest args)
  (apply #'shared-initialize-interface self args))

(defmethod reinitialize-instance :after ((self help-popup-interface) &rest args)
  (apply #'shared-initialize-interface self args))

(defmethod ywi:popup-rectangle ((popup help-popup-interface) displayer &optional subscript)
  (let* ((owner (ywi:interface-owner popup))
         (screen (capi:element-screen owner))
         (dx (slot-value popup 'dx))			; if non-NIL, relative to chooser
         (dy (slot-value popup 'dy)))
    (ywi:rect-bind (left top nil displayer-height)
        (ywi:displayer-rectangle (if dx (ywi:chooser popup) displayer) subscript screen)
      (with-slots (message more-button main-layout) popup
        ;(multiple-value-bind (x y width height) (capi:top-level-interface-geometry popup)
        (capi:with-geometry popup			; => external size and constraints
          ;(let* ((created (CAPI::GET-HINT-VALUE popup :visible-min-width nil))
          (let* ((created (CAPI::HINT-TABLE popup))	; T = not called by
                 (width (if created 			; :after interface-on-create
                            (capi:with-geometry message
                              (or capi:%width% capi:%min-width%))
                            (or capi:%width% capi:%min-width%)))
                 (delta (if created
                            (+ (capi:simple-pane-visible-width more-button)
                               (* 2 ywi:+padding+)
                               ywi:+h-gap+		; between title and more-button
                               -6)			; no explaination
                            -8))			; no explaination
                 ;;     height  ,  y 
                 ;;    owner,    client area.
                 (height (or capi:%height% capi:%min-height%))  ;(unless created
                 (x (if dx (+ left (max (- dx 50) 0)) (- left 2)))
                 (y (if dy (+ top dy 10) (+ top displayer-height -2)))
                 (right (capi:screen-width screen))
                 (bottom (capi:screen-height screen)))
            (declare (fixnum x y right bottom))
            (when height
              ;(incf height 4)				; 2*(1=padding + 1=thin border)
              (unless (<= (+ y height) bottom)		;   
                (setq y (- top height))))		; -  
            (when width
              (incf width delta)			; ,  
              (capi:set-geometric-hint popup :visible-min-width width)
              (unless (<= (+ x width) right)		;    
                (setq x (- right width))))		; -  
            (ywi:make-rect x y width height)))))))	; 0 - don't change width/height
                           

(defmethod capi:find-interface ((type (eql 'help-popup-interface)) &rest args
                                &key chooser
                                     (owner (ywi::chooser-top-level-interface chooser)))
 ;;; Args: chooser Must be specified
  ;(setq args (remove-properties args '(:owner)))
  (let ((popup (capi:locate-interface 'help-popup-interface :owner owner)))
    (if popup
        (progn (apply #'reinitialize-instance popup args)
          (ywi:interface-on-create popup)			; -> %resize-popup
          ;(ywi::popup-before-activate owner popup)
          (capi:activate-pane popup))
        (progn (setq popup (apply #'make-instance 'help-popup-interface args))
           ;(ywi::popup-before-activate owner popup)
           (capi:display popup :owner owner :process nil)))	; -> interface-on-create
                         ;:window-styles '(:always-on-top #-win32 :borderless))))
    popup))

(defun popup-context (chooser &key (displayer (ywi:current-displayer chooser))
                                   (interface (ywi:chooser-top-level-interface chooser))
                                   (help-key (capi:help-key displayer))
                                   ((:text %text%))
                                   (context (or (help-locator displayer t)
                                                (and (neq displayer chooser)
                                                     (help-locator chooser t))))
                                   (topic (or (help-locator displayer nil)
                                              (and (neq displayer chooser)
                                                   (help-locator chooser nil))))
                                   (book (or *default-book* *registered-books*))
                                   dx dy)
 ;;; Args: :text Usually equals to (ywi:help-key-tooltip help-key), but NIL can be 
  ;;		 explicitly passed when we jupm direcly to topic.
  ;;	   x,y	 Mouse coordinates relative to the top left corner of the chooser.
  (unless %text%
    (flet ((%get-text (locator)
             (cond ((and (stringp locator) (plusp (length locator))
                         (char= (char locator 0) #\?))
                    (subseq context 1))
                   (locator
                    (when-let (book-pages (collect-paths book locator t))
		      ;; Silently take the first pair (string/context-popup start . end)
                      (let* ((book (caar book-pages))
                             (page (cdar book-pages))
                             (start (cadr page)))
                        (if (numberp start)			; compiled component
                            (read-context book (absolute-pathname book (car page))
                                          start (cddr page))
                            page)))))))				; page ::= (line1 line2...)
      (setq context (first-or-self context))			; only the first is of use
      (case (yl:thys-preference '*if-context-missing*)
        (:tooltip
         (setq %text% (or (%get-text context)
                          ;(and topic (not (equal context topic)); shall we interpret
                          ;     (%get-text topic))		; shared disignator here?
                          (and (eq help-key :parent)		; mustnot be for column!
                               (when-let* ((ywi:*show-tooltips* t)
                                           (parent (capi:element-parent displayer))
                                           (%text% (or (ywi:help-callback
                                                        interface parent :tooltip)
                                                       (ywi:help-key-tooltip
                                                        (capi:help-key parent)))))
                                 (capi:wrap-text %text% 80)))
                          (ywi:help-key-tooltip help-key)
                          (format nil +missing-context+ (ywi:object-title displayer)))))
        (:topic
         (setq %text% (%get-text context)))
        (otherwise ;(nil)
         (unless (setq %text% (%get-text context))
           (setq topic nil))))))
  (cond (%text%
         (apply #'capi:find-interface 'help-popup-interface :text %text% :topic topic
                :chooser chooser :owner interface
                (when (or dx dy (multiple-value-setq (dx dy)
                                    (ywi::keyboard-post-menu-x-y chooser)))
                  (list :dx dx :dy dy))))
        (topic
         (view-topic topic :interface interface :displayer displayer))
        (t
         (ywi:beep (format nil #L"No~:[ context~;~] help information available~@[ for ~s~]."
                           (yl:thys-preference '*if-context-missing*)
                           (ywi:object-title displayer))
                   interface)))
  (values))

(defun %view-topic (popup &rest args)
 ;;; Used as :callback for more-button
  (declare (ignore args))
  (when-let (topic (slot-value popup 'topic))
    (let ((interface (ywi:interface-owner popup)))
      (capi:execute-with-interface interface
       'view-topic topic :interface interface
                         :displayer (ywi:current-displayer (ywi:chooser popup))))))

;;; displayer -> string
;(defvar *context-help-hash-table* (make-hash-table :weak-kind :key))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  HELP-CALLBACK  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Real help methods specialized on type (eql :help)
;;; 1.  For F1 compatibility, we ignore null help-key
;;; 2.  "What's this" mode should accept clicking on the title panes as well as
;;;	the controls (e.g. buttons or text panes). To get help, some users click
;;;	the control while others click the title.

(defmethod ywi:help-callback ((interface capi:interface) (pane capi:capi-object)
                              (type (eql :help))
                              &optional (help-key (capi:help-key pane)))
 ;;; CAUTION: Specialize on an interface subclass instead of capi:interface for safety!
  (when help-key					; for compatibility with F1
    (popup-context pane :interface interface :help-key help-key)))

(defmethod ywi:help-callback ((interface capi:interface) (pane capi:capi-object)
                              (type (eql :topic))
                              &optional (help-key (capi:help-key pane)))
 ;;; The :topic type means jump directly to help book page, no context popup.
  (when help-key					; for compatibility with F1
    (if-bind (locator (help-locator pane nil))		; get topic locator
      (view-topic locator :interface interface :displayer pane)
      (call-next-method))))

;;; Specializations

(defmethod ywi:help-callback ((interface capi:interface) (pane capi:menu-item)
                              (type (eql :help))
                              &optional (help-key (capi:help-key pane)))
  (ywi:help-callback interface pane :topic help-key))

(defmethod ywi:help-callback ((interface CAPI::CONFIRMER) pane (type (eql :help))
                              &optional help-key)
 ;;; Args: pane  NIL comes from ywi:beep-pane and other confirmers via ywi:help-function.
  (when help-key
    (popup-context pane :interface interface
                   :help-key help-key
                   :topic (if (and (typep pane 'capi:button)
                                   (equal (capi:item-data pane) #L"Help"))
                              (coerce-to-anchor 'ywi:*show-help-button*)
                              (%object-locator interface nil help-key (null pane))))))
    ;(view-topic (%object-locator interface nil help-key (null pane))
    ;              :interface interface)))

(defmethod ywi:help-callback ((interface ywi::confirmer) pane (type (eql :help))
                              &optional help-key) 	;(capi:help-key pane)
 ;;; Args: pane Can be NIL when clicking on the help-button itself - help-on interface
  (when help-key					; for compatibility with F1
    (popup-context (or pane interface)			; => displayer
                   :interface interface
                   :help-key help-key
                   :text (when pane (ywi:help-key-tooltip help-key))
                   :context (if pane (help-locator pane t) nil)	
                   :topic (if (and (typep pane 'capi:push-button)
                                   (eq (capi:item-data pane) :help))
                              (coerce-to-anchor 'ywi:*show-help-button*)
                              (%object-locator interface nil help-key (null pane))))))

(defmethod ywi:help-callback ((interface capi:interface) (pane null) (type (eql :topic))
                              &optional help-key)
 ;;; Args: interface Instance of ywi::confirmer or CAPI::CONFIRMER
  ;; Called by: help-function, neither context popup nor next method is invoked
  (when help-key					; for compatibility with F1
    (when-let (locator (%object-locator interface nil help-key t))
      (view-topic locator :interface interface :displayer interface))))

  #|(if-bind (locator (if pane
                          (help-locator pane)
                          (%object-locator interface help-key t)))
      (view-topic type :interface interface
                    :tooltip (when pane (ywi:help-key-tooltip help-key)))
      (ywi:beep-box #L"~@[~a~2%~]Detailed context help is not implemented for ~s."
             :format-arguments (list (ywi::help-key-tooltip help-key)
                                     (or (ywi:object-title pane) help-key))
             :icon :info  :owner interface))))|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  HELP-MENU  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun help-menu-items (interface)
 ;;; Value: List of items context help, help on window, and all the books menu-component 
  ;; Warning: F1 is not a valid user accelerator on Windows and not displayed on the right
  ;; Usage:
  ;;  (defmethod ywi:help-menu ((interface interface-subclass-or-mixin) &key)
  ;;    (let ((menu (call-next-method)))
  ;;      (dolist (item (nreverse (help:help-menu-items face)) menu)
  ;;        (ywi:insert-menu item menu :before 0))))
  (declare (ignore interface))
  `(,(make-instance 'capi:menu-item
	:mnemonic-title (load-time-value (string-append #L"Context &Help" #\Tab "F1"))
        :data :help					; :contexthelp would be better
	;:accelerator "F1"				; but F1 -> :help is hard-wired
        :callback-type '(:interface :focus :data)
        :callback 'ywi:help-callback
	:help-key #L"Pop-up 'What's This?' help about the current user interface element.")
    ,(make-instance 'capi:menu-item
	:mnemonic-title #L"Help on &Element"
        :data :topic					; means skipping context pop-up
        :callback-type '(:interface :focus :data)
        :callback 'ywi:help-callback
	:help-key #L"Open the help book page about the current GUI element.")
    ,(make-instance 'capi:menu-item
	:mnemonic-title #L"Help on &Window"
	:callback-type :interface
	:callback (lambda (interface) (view-topic (help-locator interface)
                                                  :interface interface))
        :enabled-function 'help-locator
	:help-key #L"Open the help book chapter about the current window.")
    ,(make-instance 'capi:menu-component
	:callback-type :interface-data
	:selection-callback (lambda (interface book)
                              (view-page :book book :interface interface)) ;:new nil
	:items-function (lambda (interface)
                          (declare (ignore interface))
                          (let ((acc ())
                                (i 0))
                            (dolist (book *registered-books*)
                              (let ((description (description book)))
                                (push (make-instance 'capi:menu-item
                                       :data book
                                       :title (format nil "~d ~a"
                                                      (incf i)
                                                      (or (title book) description))
                                       :help-key description)
                                      acc)))
                            (nreverse acc)))) ))
	;:items *registered-books*
	;:print-function (lambda (book) (or (title book) (description book))))))
     ;,@(when help-books-menu
     ;    `(,help-books-menu))))

(let ((lw:*handle-warn-on-redefinition* #+lispworks4 :quiet #-lispworks4 :warn))
 (defmethod capi:pane-popup-menu-items :around ((pane capi:simple-pane)
                                                (interface capi:interface))
  ;; Put What's This? item last or one but last before Properties (:name :properties)
  (when-let (items (call-next-method))
    (if (yl:thys-preference '*show-what-is-this-mi*)
        (let ((help-key (capi:help-key pane)))
          (nconc items
          `(,(make-instance 'capi:menu-component :items
             `(,(make-instance 'capi:menu-item
		:mnemonic-title #L"&What's This?"
		:name :help  :data :help		; :contexthelp would be better
                :enabled-function (if help-key #'lw:true #'lw:false)
		:callback-type '(:interface :collection)
		:callback (lambda (interface component &aux (menu (capi:element-parent
                                                                   component)))
                            ;(capi:with-geometry menu
                            ;  (format t "~%What's this: x ~s y ~s" capi:%x% capi:%y%))
                            (popup-context pane :interface interface :help-key help-key
                                           :dx (capi:capi-object-property menu :x)
                                           :dy (capi:capi-object-property menu :y)))
                           ;(ywi:help-callback interface pane #1#))
		:help-key nil))))))
        items))))
