;;; -*- Mode: Lisp; -*-
;;; YstokHelp - HTML-based help
;;; Copyright (c) 2009-2014 Dr. Dmitry Ivanov.  All rights reserved.
;;; Website: http://lisp.ystok.ru/yhelp/
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Index compiler

;;; Index compiler creates a single idx*.html file or a family of idx_*.html files.
;;; 
;;; Any source file can include the following meta tag for the topic or
;;; segment property:
;;;  <meta name="ROBOTS" content="NOINDEX">
;;;	Specifies that the segment should be ignored by Index compiler.
;;;	Even anchors residing in this file are not collected for index.
;;;	You should not index a page that contains only links or graphics,
;;;	or a page on which you don't want the user to land as a result of a search.

(in-package :ystok.help)

;(setf (fdefinition 'not-word-char-p) (complement #'word-char-p))

(defun greek-char-p (char)
 ;;; From "Alpha" (913) to "piv" (982)
  (<= 913 (char-code char) 982))

(eval-when (:compile-toplevel :load-toplevel :execute) (ystok.meta:enable-meta-syntax))
(defun string-boundaries (string)
 ;;; Seek for word start positions and collect them into the list acc
  (let ((acc ())
        (lang-word-char-p-function (lang-word-char-p-function *language*)))
    (labels ((%word-char-p (char)
               (or (alphanumericp char)
                   (greek-char-p char)
                   (char= char #\_)
                   (funcall lang-word-char-p-function char)))
             (%not-word-char-p (char)
               (not (%word-char-p char))))
      (ystok.meta:with-string-meta (string)
        (ystok.meta:match $[ $@(%not-word-char-p) @(%word-char-p)
                             !(push (1- ystok.meta:%index%) acc)
                             $@(%word-char-p) ])))
    (nreverse acc)))
(eval-when (:compile-toplevel :load-toplevel :execute) (ystok.meta:disable-meta-syntax))

(defun content-boundaries (cont)
 ;;; Args: path       ,   
  ;;       left, right ,    ,     path
  ;; Values: 1) left 2) right
  (let ((acc ()))
    (labels ((%collect (list)
               (dolist (child list)
                 ;; If the child is not found in list, everything goes to left
                 (cond ((consp child)
                        (%collect (html:elem-cont child)))
                       ((stringp child)
                        (dolist (position (string-boundaries child))
                          (push (cons child position) acc)))))))
      (%collect cont))
    (nreverse acc)))

(defun split-lthml-content-by (cont elem &optional position)
 ;;; Args: cont     LHTML list
  ;;       elem     String or keyword element (or even tree?) within the cont
  ;;       position Position to cut the string
  ;; Values: 1) left 2) right - both are LHTML lists.
  (labels ((%split (tree)
             (do ((head ())
                  (tail tree (rest tail)))
                 ((null tail)
                  (values tree nil))			; elem is not found in the tree
               (let* ((child (first tail)))
                 (cond ((eq child elem)			; elem is found - put it to right
                        (return (if (and (stringp elem) (iplusp position))
                                    (values (nreconc head (list (nsubseq elem 0 position)))
                                            (cons (nsubseq elem position) (rest tail)))
                                    (values (nreverse head)
                                            tail))))
                       ((consp child)
                        (let ((key (html:elem-key child))
                              (attrs (html:elem-attrs child))
                              (cont (html:elem-cont child)))
                          (multiple-value-bind (left right) (%split cont)
                            (if (eq left cont)		; content is not changed, hence
                                (push child head)	; elem is not found in the clild
                                (return (values (nreconc head
                                                         (if left
                                                             (list (html:make-elem
                                                                    key attrs left))
                                                             ()))
                                                (if right
                                                    (cons (html:make-elem key attrs right)
                                                          (rest tail))
                                                    (rest tail))))))))
                       (t
                        (push child head)))))))		; elem is not found in the clild
    (%split cont)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  INDEX-LETTERS  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Letters that define the index page header and footer

(defvar %index-letters%)

(defun char-representative (char)
 ;;; Uppercased version of CHAR if alphabetic or #\* otherwise
  ;; Used as:    *chars-hash-table*  header/footer
  (or (lang-char-representative *language* char)
      #\*))					; represents all other characters

(defun char-fragment (char)
 ;;; ID and fragment part of URI
  (if (char= char #\*) "star" (char-representative char)))

(defun char-title (char &optional (lang *language*))
  (if (char= char #\*)
      (gettext "Non-alphabetic" *yhelp-bundle* lang)
      nil))

(defun char-output-name+type (prefix char)
  (make-pathname :name (format nil "~a~@[_~a~]" prefix
                               (case char
                                 ((nil) nil)
                                 (#\* "star")
                                 (otherwise (transliterate-char char))))
                 :type "html"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  INDEX Class  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass index (navigation-mixin generated-component)
 ((index-type :initarg :type :initform :permuted)
  (max-left-chars :reader max-left-chars :initarg :max-left-chars
	:initform *default-index-max-left-chars*)
  (multipage :reader multipage :initarg :multipage :initform nil)
  (letters   :reader letters :initarg :letters :initform nil)

  ;; Keyword or list of keywords, book language by default
  (stopwords-language :reader stopwords-language :initarg :stopwords-language)

  ;; Hash table mapping keywords, i.e. phrases, to a list of all anchors.
  ;; Map: phrase -> (content (anchor1 segment1 . plist1) (anchor2 segment2 . plist2) ...)
  ;;      phrase ::= plain string representing anchor or keyword text,
  ;;      content ::= NIL or LHTML list representing anchor or keyword content
  ;;		      may be a result of BBcode conversion.
  ;; The tuples are sorted according to the segment< predicate.
  (keywords-hash-table :reader keywords :initform (make-hash-table :test #'equal))

  ;; Hash table mapping characters to a list of all permuted index entries
  ;; starting with the corresponding character.
  ;; Map: char -> ((phrase position content (anchor1 segment1 . plist1)...)...)
  ;;    phrase ::= string              ; anchor or keyword text
  ;;		 | list                ; content
  ;;	position ::= integer           ; position of char in phrase (i.e. word start)
  ;;		   | (string . integer)
  ;;    -> content (anchor1 segment1 . plist1)... ; identical to val in keywords-hash-table
  (chars-hash-table :reader chars :initform (make-hash-table :test #'equal)))
 (:default-initargs
  :tab "Index"
  :title "Index"
  :description "Permuted Index"))

(defmethod reinitialize-instance :after ((self index)
                                         &key (stopwords-language nil supplied-p))
                                              ;type description
  (if supplied-p
      (cond ((stringp stopwords-language)
             (setf (slot-value self 'stopwords-language)
                   (intern-upcase stopwords-language)))
            ((consp stopwords-language)
             (setf (slot-value self 'stopwords-language)
                   (mapcar (lambda (arg) (if (stringp arg) (intern-upcase arg) arg))
                           stopwords-language))))
      (setf (slot-value self 'stopwords-language)
            (or (language (asd:component-system self)) :en))))
;  (unless description
;    (case type
;      (:permuted (setf (slot-value self 'description "Permuted Index")))

(defun find-index (module)
  (find 'index (asd:module-components module) :test #'xtypep))

(defun index-start-pathname (index)
 ;;; Single or the first index file
  (merge-pathnames (char-output-name+type
                    (asd:component-name index)	;(index-file book)
                    (when (multipage index)
                      (find-if (lambda (char) (gethash (char-representative char)
                                                       (chars index)))
                                                ;*chars-hash-table*
                               %index-letters%)))
                   (pathname-location (asd:component-output-pathname index))))

(defun %downcase-first (tree &optional plist)
 ;;; Lowercase the first letter of the very first string within the elem
  ;; Args: tree String or content list
  ;; Values: 1) tree itself or its copy
  ;;         2) T = a string was found within the elem
  (let ((classes (getf plist :classes)))
    (cond ((and classes (intersection classes *anchor-span-css-classes*
                                      :test #'string-equal))
           tree)
          ((stringp tree)
           (let* ((src (schar tree 0))
                  (dest (locale-char-downcase src)))
             (unless (char= src dest)
               (setf tree (copy-seq tree)
                   (schar tree 0) dest)))		; not destructive
           (values tree t))
          ((consp tree)
           (do ((head ())
                (tail tree (rest tail)))
               ((null tail)
                (values tree nil))			; no string found in tree
             (let* ((child (first tail))
                    (cont (html:elem-cont child)))
               (multiple-value-bind (new found) (%downcase-first (or cont child))
                 (if found
                     (return (values (if (or (eq new child)	; nothing has been copied
                                             (eq new cont))
                                         tree
                                         (nreconc head (cons (if cont	; new - content
                                                                 (html:make-elem
                                                                  (html:elem-key child)
                                                                  (html:elem-attrs child)
                                                                  new)
                                                                 new)	; new - string
                                                             (rest tail))))
                                     t))				; a string found
                     (push child head))))))		; no string found in child
          (t tree))))

(defun anchor-keywords (plist)
 ;;; Value: List (phrase1 ...), where phrase ::= string or (string . content)
  (let (val)
    (cond ((null (setq val (getf plist :keywords *unbound-value*)))
           ())					; indexing is suppressed expilicitly
          ((neq val *unbound-value*)
           ;; Explicit keywords are specified, %downcase-first conversion for them
           (if (getf plist :bbcode *bbcode-in-keywords*)
               (loop for phrase in (split-seq #\, val)
                     collect (if (char-position #\[ phrase)	; BBcode string
                                 (let ((cont (html:bbcode-lhtml phrase)))
                                   (cons (interleave-seq 'string #\Space
                                                         (lhtml-word-list cont))
                                         cont))
                                 phrase))
               (split-seq #\, val)))
           ;((getf plist :noindex)		; indexing is suppressed
           ; ())
          ((setq val (getf plist :text))			; expicit text
           (if (and (getf plist :bbcode *bbcode-in-text*)
                    (char-position #\[ val))			; BBcode string
               (let ((cont (%downcase-first (html:bbcode-lhtml val) plist)))
                 (list (cons (interleave-seq 'string #\Space (lhtml-word-list cont))
                             cont)))
               (list (%downcase-first val plist))))		; plain string
          ((setq val (getf plist :content))			; distilled LHTML list
           (let ((cont (%downcase-first val plist)))
             (list (cons (interleave-seq 'string #\Space (lhtml-word-list cont))
                         cont)))) )))

(defun fill-keywords (index &optional (book (asd:component-system index)))
 ;;; Fills keywords hash-table with all entries found in anchors hash-table
  ;; and topic/segment keywords
  (loop with keywords = (keywords index)
        for anchor being the hash-keys of (anchors book)
            using (hash-value saplist)
        do (dolist (pair saplist)				; exclude origin strings
             (when (consp pair)					; (segment . plist)
               (let* ((segment (car pair))
                      (value (segment-property segment :index *unbound-value* t)))
                 (unless (if (eq value *unbound-value*)		; not specified explicitly
                             (member (segment-property segment :robots nil t) ;:source
                                     '("NOINDEX") ;"KEYWORDS"	; try :robots (from meta)
                                     :test #'string-equal)
                             (or (eq value :none) (null value))) ; explicitly prohibited
                             ;(eq (segment-property segment :index nil t) :none))
                   (dolist (phrase (anchor-keywords (cdr pair))) ; <- plist
                     (let* ((text (first-or-self phrase))
                            (content (if (consp phrase) (cdr phrase) nil))
                            (old (gethash text keywords)))
                       (cond (old
                              ;; Take content from the first source by chance
                              (when (and content (not (first old)))
                                (setf (first old) content))
                              (setf (rest old)
                                    (merge 'list 
                                           (list (cons anchor pair))
                                           (rest old)
                                           #'segment<
                                           :key #'second)))
                             (t
                              (setf (gethash text keywords)   ; => (anchor segment . plist)
                                    (list content (cons anchor pair)))))))))))))
                           #|(merge 'list
                                  (list (cons anchor pair))   ; => (anchor segment . plist)
                                  ;(list (cons segment anchor))
                                  (gethash phrase keywords)
                                  #'segment<
                                  :key #'second))))))))|#
  ;; TODO:   meta name="keywords"

(declaim (inline stopword-p))
(defun stopword-p (string index)
  (when-let (language (gethash string *stopword-hash-table*))
    (memq language (listify (stopwords-language index)))))

(defun fill-chars (index &optional (exclude-numbers t))
 ;;; Break all the phrases from the keywords hash-table into words and
 ;;; spread according to first letters of these.
  ;; Fill the chars hash-table with a list of tuples of the form
  ;;	(phrase position-designator content (segment1 anchor1 . plist1) ...)
  ;;    where phrase - entire string,
  ;;          position-designator ::= start or (chunk . start)
  ;; Args: exclued-numbers Exclude words completely consisted of digits
  ;; NB: Always exclude the following words
  ;;	 - within parantheses (), brackets [], or braces {}.
  #-debug (declare (optimize (speed 3) (safety 0) #+lispworks (hcl:fixnum-safety 0)))
  (dolist (lang (listify (stopwords-language index)))
    (ensure-stopwords lang))
  (loop with chars = (chars index)
        for phrase of-type simple-string being the hash-keys of (keywords index)
            using (hash-value tuple)
        ;; tuple ::= (content (anchro1 seqment1 . plist1) ...)
        do
        (loop with embraced = nil				; embraced mode
              and content = (first tuple)
              for rest on (if content
                              (content-boundaries content)
                              (string-boundaries phrase))
              for pair = (first rest)
              for chunk = (if content (car pair) phrase)
              and start of-type fixnum = (if content (cdr pair) pair)
              for end = (position-if-not #'whitespace-char-p chunk
                                         :from-end t
                                         :start start
                                         :end (if content
                                                  (if (eq chunk (caadr rest))
                                                      (cdadr rest)
                                                      nil)
                                                  (second rest)))
              for word = (nsubseq chunk start (if end (1+ end) nil))
              do
              (cond (embraced
                     ;; Already in braces - ignore the word and
                     ;; turn off the embraced mode if a closing is encountered.
                     (setq embraced (not (char-position (char word (1- (length word)))
                                                        ")]}"))))
                    ((or (setq embraced (and (plusp start)
                                             (char-position (schar chunk (1- start))
                                                            "([{")))	; open found
                         (stopword-p word index)
                         (and exclude-numbers
                              (loop for char of-type character across word
                                    always (digit-char-p char 10)))))
                    (t
                     (let ((representative (char-representative (char chunk start))))
                       (push (list* phrase pair tuple)
                             (gethash representative chars))))))))

(defun sort-chars (index)
 ;;; Sort tuple lists that are the hash values of chars hash-table alphabetically
  ;; (starting at the position denoted by the `permutation' index).
  (loop with chars = (chars index)
        for char being the hash-keys of chars
        do (setf (gethash char chars)
                 (stable-sort 	
                  (sort (gethash char chars)
                        #'locale-string-lessp		; sort
                        :key #'first)			; first sort by the entire phrases
                  #'locale-string-lessp
                  :key (lambda (tuple)			; second sort by first words
                         (let ((pair (second tuple)))
                           (if (consp pair)		; (chunk . position)
                               (nsubseq (car pair) (cdr pair))
                               (nsubseq (first tuple) pair))))))))

;;; Templates and generation

(defun index-html-letters (index &optional current-char bottom)
 ;;; Make DIV element for index page:
  ;; - either header (bottom is false)
  ;; - or footer (bottom is true).
  (html:div (:class (format nil "toolbar ~:[header~;footer~]" bottom))
    (loop with hash-table = (chars index)
          and lang = (language (asd:component-system index))
          for char across %index-letters%
          for representative = (char-representative char) ;(yl:transliterate-char char)
          do (html:princ-http #\Space)
             (cond ((eql char current-char)
                    (html:span (:style :current)
                      #1=(html:princ-http representative)))
                   ((gethash representative hash-table)	;*chars-hash-table*
                    (html:a (:href (string-append #\# (char-fragment char))
                             :target "_self"
                             :title (:optional (char-title char lang)))
                      #1#))
                   (t #1#) )))
  (html:htm :newline))

(defun html-print-first (string &key letter word all)
 ;;; Helper to make the starting part of the string formatted differently:
  ;; - first letter, CSS equivalent td:first-letter (works for Firefox)
  ;; - first word (not implemented actually)
  ;; - entire string (CSS equivalent td:first-line?).
  ;; Args: letter, word, all
  ;;		Keywords denoting HTML character formatting tags.
  (cond ((or word all)
         (html:html-print (if letter
                              `(,(or word all)
                                (,letter (:princ-safe ,(char string 0)))
                                (:princ-safe ,(nsubseq string 1)))
                              `(,(or word all)
                                (:princ-safe ,string)))))
        (letter
         (html:html-print-list `((,letter (:princ-safe ,(char string 0)))
                                 (:princ-safe ,(nsubseq string 1)))))
        (t (html:princ-safe-http string))))


(defun index-html-letter-rows (index char &optional th
                                                    (book (asd:component-system index)))
  ;;; Output rows of an HTML table with the permuted index for the character CHAR.
   ;; Args: th  If true, output a heading row with <th>C</th>
  (let* ((representative (char-representative char))
         (max-left-chars (max-left-chars index))
         (entries (gethash representative (chars index))))
   (when entries
     (when th					; output the letter for single-file index
       (terpri html:*html-stream*)
       (html:tr (:id (char-fragment char))
         (html:th () (html:princ-http html:+nbsp+))
         (html:th () (html:princ-safe-http representative)))
       (terpri html:*html-stream*))
     (loop with base = (asd:component-output-pathname book)
           for (phrase pair content . tuples) in entries
           do
          (let ((chunk (if content (car pair) phrase))
                (position (if content (cdr pair) pair))
                (class nil)
                left right left-title word-list length)
            (cond (content
                   (loop for (nil nil . plist) in tuples
                         do (when-let (classes (getf plist :classes))
                              (setq class (interleave-seq 'string #\Space classes))))
                   (multiple-value-setq (left right)
                       (split-lthml-content-by content chunk position))
                   (when (< max-left-chars
                            (setq length (reduce #'+
                                                 (setq word-list (lhtml-word-list left))
                                                 :key #'length
                                                 :initial-value (1- (length left)))))
                     (let ((text (string-append #+lispworks #.(code-char #x2026) ;:ellipsis
                                                #-lispworks "..."
                                                (nsubseq (interleave-seq 'string
                                                                         html:+nbsp-char+
                                                                         word-list)
                                                         (- length -1 max-left-chars)))))
                       ;; If the left consisted of one element, copy its formatting
                       (setq left (if (and (consp (car left)) (not (rest left)))
                                      (list (list (caar left) text))	; html:make-elem
                                      text)
                             left-title phrase))))
                  (t
                   (setq left (nsubseq phrase 0 position)
                         right (nsubseq phrase position))
                   (when (< max-left-chars (setq length (length left)))
                     (setq left (string-append #+lispworks #.(code-char #x2026) ;:ellipsis
                                               #-lispworks "..."
                                               (nsubseq phrase
                                                        (- length -1 max-left-chars)
                                                        position))
                           left-title phrase))))
             (html:tr ()
	       ;; The lefmost column is right-aligned
               (html:td (:class (:optional class)
                         :title (:optional left-title))
                 (if (listp left)
                     (html:html-print-list left)
                     (html:princ-safe-http left)))
	       ;; The second column is left-aligned
               ;; As CSS td:first-letter does not work on many browsers,
               ;; try to bold the first char even in list-like content.
               (html:td (:class (:optional class))
                 (cond ((stringp right)
                        (html-print-first (substitute html:+nbsp-char+ #\Space right)
                                          :letter :b))
                       ((consp right)
                        (let ((first (first right)))
                          (if (stringp first)
                              (progn (html-print-first (substitute html:+nbsp-char+
                                                                   #\Space first)
                                                       :letter :b)
                                (html:html-print-list (rest right)))
                              (html:html-print-list right)))))	; too complex to bold
                 ;; Output numeral links
                 (html:princ-http html:+nbsp+)			; additional non-break
                 (loop for (anchor segment . plist) in tuples
                       and j upfrom 1
                       do
                       (let ((uri (uri:enough-uri (asd:component-output-pathname segment)
                                                  base))
                             (title (title segment)))
                         (html:princ-http #\Space)
                         (when anchor
                           (setf (uri:uri-fragment uri) anchor))
                         (html:a (:href uri :title (:optional title)) ;:target target
                           (html:princ-http #\[)
                           (multiple-value-bind (span-class span-cont)
                               (get-properties plist *anchor-span-css-class-properties*)
                             ;; Check for suffix like "category"
                             ;; It is printed in the Index anyway, even if
                             ;; [...] is specified with the keyword property.
                             (when span-class
                               (html:span (:class span-class)
                                 (html:princ-safe-http
                                  (string-trim "[]"
                                               (interleave-seq 'string html:+nbsp-char+
                                                 (lhtml-word-list (listify span-cont))))))
                                               ;(html:html-princ-to-string span-cont))))
                               (html:princ-http html:+nbsp+)))
                           (if (getf plist :definition)
                               (html:b () (html:princ-http j))	; in bold font
                               (html:princ-http j))
                           (html:princ-http #\]))) ))))))))
                                 ;(format html:*html-stream* "[~d]" j)))))))))))

                   
(defmethod toolbar-lhtml ((component index) (position (eql :header)) segment
                          &key (book (asd:component-system component)))
  (declare (ignore segment position))
  (navi-tabs-lhtml component book))

(defun index-html-page (index
                        &optional char
                                  (pathname (merge-pathnames
                                             (char-output-name+type
                                              (asd:component-name index) char)
                                             (pathname-location
                                              (asd:component-output-pathname index)))))
 ;;; Creates the index page for the character CHAR or the start page if CHAR is NIL.
  (let* ((book (asd:component-system index))
         (*language* (language book))
         (max-left-chars (max-left-chars index)))
    (html:with-html-file (pathname :external-format (output-html-format book)
                                   :mode (or (charset book) :xml)
                                   :prologue t)
      (log-info "creating index (permuted)-> ~a" pathname)
      (html:html (:lang (:optional *language*)
                  :xml\:lang (:optional *language*)
                  :xmlns html::+xmlns+)
        (navi-html-head book :title (format nil "~a~@[ - ~a~]"
                                            (description index)
                                            (case char
                                              ((nil) nil)
                                              (#\* (gettext "Non-alphabetic"
                                                            *yhelp-bundle* *language*))
                                              (otherwise (locale-char-upcase char)))))
        (html:body ()
          (html:html-print-list (toolbar-lhtml index :header nil))
                           ;(navi-toolbar-lhtml (asd:component-system index) index))
          (index-html-letters index char)
          (terpri html:*html-stream*)
          (html:table (:class (:optional *index-table-css-class*))
            (html:col (:align :right
                       :style (format nil "width:~dem; text-align:right" max-left-chars)))
            (html:col (:align :left
                       :style "text-align:left;first-letter:"))
            (if char
                (index-html-letter-rows index char nil)
                (loop for char across %index-letters%
                      do (index-html-letter-rows index char t))))
          (terpri html:*html-stream*)
          ;(html:hr ())
          (index-html-letters index char t))))))

(defmethod %create-index ((index index) (type (eql :permuted))
                          &optional (multipage (multipage index)))
 ;;; Create index file(s) in the book output directory.
  (clrhash (keywords index))	;*keywords-hash-table*)
  (clrhash (chars index))	;*chars-hash-table*)
  (fill-keywords index)
  (fill-chars index)
  (sort-chars index)
  (if multipage
      (loop for char being the hash-keys of (chars index)	; create bunch of files
            do (index-html-page index char))		; index-file_<transliterated>.html
      (index-html-page index))				; create single index-file.html
  (values))

;;; Top-level

(defun create-index (book-system &key type (multipage nil supplied-p))
 ;;; Create index file(s) in the book output directory.
  ;; Args: type, multipage
  ;;		Can only be explicitly passed for testing the index compiler alone
  ;;            as other components rely on the values of the index object slots.
  (when-let* ((book-system (if (typep book-system 'book-system)
                               book-system
                               (asd:find-system book-system t)))	;find-book
              (index (find-index book-system)))
    (let* ((*language* (language book-system))
           (%index-letters% (or (letters index) (lang-default-index-letters *language*))))
      (%create-index index
                     (or type (slot-value index 'index-type))
                     (if supplied-p multipage (multipage index))))))
  
(defmethod asd:output-files ((o (eql :compile)) (c index))
  (declare (ignore o))
  (let* ((book (asd:component-system c))
         (pathname (asd:component-output-pathname book))
         (name (asd:component-name c)))
    (if (and (multipage c)
             (plusp (hash-table-count (chars c))))
        ;; Create bunch of files idx_<transliterated>.html
        (loop for char being the hash-keys of (chars c)
              collect (merge-pathnames (char-output-name+type name char) pathname))
	;; single idx.html
        (list (merge-pathnames (char-output-name+type name nil) pathname)))))

(defmethod asd:perform ((o (eql :compile)) (c index))
  (declare (ignore o))
  ;(when (getf (asd::operation-original-initargs :index t))
  (let* ((*language* (language (asd:component-system c)))
         (%index-letters% (or (letters c) (lang-default-index-letters *language*))))
    (%create-index c (slot-value c 'index-type))))
