;;; -*- Mode: Lisp; -*-
;;; YstokHelp - HTML-based help
;;; Copyright (c) 2009-2014 Dr. Dmitry Ivanov.  All rights reserved.
;;; Website: http://lisp.ystok.ru/yhelp/
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Parsing callbacks and method for ASDlite operations

(in-package :ystok.help)

(setq phtml:*end-tag-mismatch-action* 'yl:logg)

;; External format for the parsed-html-file in process.
;; Both load-source-op and compile-op bind this variable with a value that is
;; - eihter extracted from the charset or,
;; - equal to (external-format book) if no <meta> element is present in the source file.
(defvar *current-external-format* nil)

;;;;;;;;;;;;;;;;;;;; COMMENT-CALLBACK and TMPL_INCLUDE  ;;;;;;;;;;;;;;;;;;;;;;;;
;;; We do not collect comments during load-source-op, but do it during compile-op.
;;;
;;; When we encounter
;;;	<!-- TMPL_INCLUDE filename -->,
;;; we must parse the included file immediately for both load-source-op and compile-op.
;;; Cf. html:include, which is invoked only during generating.
;;;
;;; LIMITATIONS
;;;  1. All included files must be in the same external-format as of the including
;;;	topic file.
;;;  2.	Files of type .htm and .html are only supported for now.
;;;	Do we need including and parsing LHTML files?
;;; TODO: Shall we use a hash-table to cache the results of parsing included files,
;;;       on which operation?

;; The variable *include-function* holds a function that is invoked from the comment
;; callback, signature: (pathname) => <comment-callback values>
(defvar *include-function* nil)

(defun comment-callback (elem)
 ;;; With the comment element, do one of
  ;; - On :preprocess pass,
  ;;   -- replace <!-- HELP_SEGMENT ...--> by a segment instance,
  ;;   -- parse recursively a file mentioned  <!-- TMPL_INCLUDE pathname -->.
  ;; - On :compile pass,
  ;;   -- either retreive the segment instance,
  ;;   -- parse recursively a file mentioned  <!-- TMPL_INCLUDE pathname -->, 
  ;;   -- or leave the element untouched if templates are possibly used,
  ;; - Otherwise, if the element is not a segment separator or inclusion,
  ;;   remove it at all.
  ;; The *include-function* should provide the values returned from this callback.
  (when (eq (html:elem-key elem) :comment)
    (when-let (string (first (html:elem-cont elem)))		; entire comment text
      (let* ((length (length string))
             ;; a la (position #\Space string :test-not #'char=)
             (start (or (position-if-not #'htt::whitespacep string) length))
             (end (min length (+ start #.(length +segment-tag+)))))
        (cond ((string= string +segment-tag+ :start1 start :end1 end)
               (values (cond (*current-segment*			; compile pass
                              (nth-segment (incf *current-segment*)
                                           *current-topic*))
                             ((neq (first-or-self phtml:*callback-parent-tag*) :body)
                              (phtml:syntax-error
               #L"Separator <!-- ~a ...--> is only allowed on top-level of <BDDY> content."
                               +segment-tag+))
                             (t					; preprocess pass
                              (make-segment *current-topic*
                               (if (find-if-not #'htt::whitespacep ; Lisp form present,
                                                string :start end) ; must be a plist
                                   (html:template-read-from-string string
                                    :start end
                                    :package (load-time-value (find-package :ystok.help)))
                                   ()))) )
                       t))
              ;((string= string "HELP_NUMBER" :start1 start :end1 end)
              ((and (string-equal string #2="TMPL_INCLUDE" :start1 start
                                  :end1 (setq end (min length (+ start (load-time-value
                                                                        (length #2#))))))
                    *include-function*)
               (let ((pathname (merge-pathnames		; a la html::include
                                (html:template-read-from-string string
                                     :kind :attribute
                                     :start end
                                     :package (load-time-value (find-package :ystok.help)))
                                (pathname-location
                                 (asd:component-pathname *current-topic*)))))
                                 ;template:*default-template-pathname*)))
                 (if (member pathname html::%included-files% :test #'pathname-equal)
                     (values (phtml:syntax-cerror #L"Ignore inclusion of this file."
                              #L"Infinite recursion - file ~A includes itself."
                              pathname)
                             :append)
                     (let ((html::%included-files% (cons pathname html::%included-files%)))
                       (funcall *include-function* pathname)))))
              ((and *current-segment*				; compile-op pass
                    (use-templates (asd:component-system *current-topic*)))
               elem)
              (t
               (values nil :append)))))))			; delete comment at all

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  PREPROCESS  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; We are accumulating something in the properties slot of an asd:component.
;;;  :!doctype  Represents *html-mode* and is pertinent to the entire topic
;;;  :charset   String value pertinent to the entire topic and extracted from
;;;		<meta http-equiv="Content-Type" content="...; charset=value">
;;;  others	Pertinent to the first segment (segments are indexed started from zero)
;;;
;;; Segment dividers are comments like <-- HELP_SEGMENT plist -->
;;; and must reside on top-level just inside BODY.

(defun collect-metas-callback (elem)
  (case (html:elem-key elem)
    (:!doctype			; set *html-mode*
     (setf (asd:component-property *current-topic* :mode)
           (if phtml::*xhtml-mode* :xml :sgml)))
               ;(search "XHTML" (first (html:elem-cont elem))
    (:title
     (setf (asd:component-property *current-topic* :title)
           (first (html:elem-cont elem))))
    (:meta
     (let ((attrs (html:elem-attrs elem))
           name)
       (cond ((setq name (getf attrs :name))
              (cond ((member name '(:description :keywords :robots)
                             :test #'string-equal)
                     (setf (asd:component-property *current-topic* (intern-upcase name))
                           (getf attrs :content)))
                    ((string-equal name :braces)
                     ;; Convert the value from string a la phtml:parse-declarations
                     (let ((content (getf attrs :content)))
                       (setf (asd:component-property *current-topic* (intern-upcase name))
                             (if (or (zerop (length content))
                                     (string-equal content "nil")
                                     (string-equal content "none"))
                                 nil
                                 content)))) ))
             ((setq name (getf attrs :http-equiv))
              (cond ((string-equal name :content-type)
                     (when-let (charset (phtml:parse-content-type-charset (getf attrs
                                                                                :content)))
                       (setf (asd:component-property *current-topic* :charset) charset)
                       ;; If the charset mismatches the external-format set on opening,
                       ;; reopen the source.
                       ;; Rationale: Only a few implementaions support
                       ;;            (setf stream-external-format)
                       (unless (html:charset-ef-match-p charset *current-external-format*)
                         (setq *current-external-format* (html:charset-ef charset))
                            ;(html:get-external-format (phtml:parse-charset charset)))
                         (log-warn "Reopening source file ~a with external-format ~s"
                                   (asd:component-pathname *current-topic*)
                                   *current-external-format*)
                         (throw '%parse nil)))) ))
) ) )) )

(defun collect-anchors-callback (elem)
 ;;; Called on load-source-op only
  ;; BAD: We are ignoring comments with template conditional pseudo-tags in
  ;;      the loading-source phase.
  ;;      Even if an anchor is encountered in a false branch of code,
  ;;      it gets collected nevertheless, and listed in the index finally.
  ;; WORKAROUND: Avoid defining anchors in conditionalized parts of source files!
  (let* ((attrs (html:elem-attrs elem))
         (key (html:elem-key elem))
         (index (not (popup-p *current-topic*)))		; do not index popups
         (anchor (case key
                   (:a
                    (getf attrs :name))
                   ((:dfn :dt)
                    (getf attrs :id))
                   ((:h1 :h2 :h3 :h4 :h5 :h6 :caption)
                    (when index
                      (setq index *index-heading-anchors*))
                    ;(when (and index (eq *index-heading-anchors* :keywords))
                    ;  (setq index :keywords))
                    (getf attrs :id)))))
    (when anchor
      (when (char-position #\% anchor)
        (setq anchor (uri:percent-decode anchor
                                         :external-format *current-external-format*)))
      (register-anchor anchor elem attrs index)
      (pushnew anchor (anchors *current-topic*) :test #'string=))
    (when (getf attrs :href)					; check for explicit calls
      (when-let (onclick (getf attrs :onclick))			; of popup or popup_menu
        (cond ((string= onclick +popup-onclick-js+)
               (pushnew :note (segment-property (last-segment *current-topic*)
                                                :popups)))
              ((search "popup_menu" onclick)
               (pushnew :menu (segment-property (last-segment *current-topic*)
                                                :popups))))))
    nil))

(defun span-callback (elem)
 ;;; Shorten ((:SPAN :LANG "en-US") content) -> content
  ;; Q: Do we need a more complex %test?
  (let ((attrs (html:elem-attrs elem))
        (lang-shorten-span (lang-shorten-span *language*)))
    (when (and (eq (first attrs) :lang)
               (null (cddr attrs))				; single attribute
               (flet ((%test (subject pattern)
                        (string-equal subject pattern)))
                 (cond ((eq lang-shorten-span t))
                       ((consp lang-shorten-span)
                        (member (second attrs) lang-shorten-span :test #'%test))
                       (t
                        (%test (second attrs) lang-shorten-span)))))
      (values (html:elem-cont elem) :append))))

(defmethod asd:perform ((o (eql :preprocess)) (*current-topic* parsed-html-file))
  ;; lean up anchors gathered from the source file of the component earlier
  (clear-anchors *current-topic*)
  (clear-headings *current-topic*)
  (setf (segments *current-topic*) nil)
  (let* ((book-system (asd:component-system *current-topic*))
         (*language* (language book-system))
         (*current-external-format* (external-format book-system))
         (*current-segment* nil)			; load-source pass indication
         (*current-heading* nil)
         (callbacks (nconc (h-callback-alist *current-topic*)
                           (when (lang-shorten-span *language*)
                             (list '(:span . span-callback)))
                           '((:a	. collect-anchors-callback)
                             (:dfn	. collect-anchors-callback)
                             (:dt	. collect-anchors-callback)
                             (:caption	. collect-anchors-callback)
                             (:comment	. comment-callback)
                             (:!doctype	. collect-metas-callback)
                             (:meta	. collect-metas-callback)
                             (:title	. collect-metas-callback) ))))
    (flet ((%parse (pathname &optional included)
             (log-info "preprocess~:[:~;+~] ~a" included pathname)
             (let ((lhtml (phtml:parse-html pathname
                                :callbacks callbacks
                                :collect-data :callbacks	;:callback-only t
                                :parse-entities t
                                ;:unknown-tag-action 'logg	;:collect-rogue-tags t
                                :external-format *current-external-format*)))
               (unless (or lhtml included)	; don't signal if included file is empty
                 (error 'asd:operation-error :component *current-topic* :operation o))
               t)))					; returned from catch
      (let ((*include-function*
             (lambda (pathname)
               ;(log-info "load-source-op+ ~a" pathname)
               (case (file-type-format pathname)
                 (:html
                  (%parse pathname t)))
               (values nil :append))))			; <- returned from callback
        ;; Give a second try if NIL is thrown from collect-metas-callback
        ;; ASSUMPTION: The throw occurs on the very beginning of the file, so there is
        ;;             no need to reinitialized anchors and segment list.
        (unless (catch '%parse
                  (%parse (asd:component-pathname *current-topic*)))
          (%parse (asd:component-pathname *current-topic*)))))))
         ;(template:*default-template-pathname* (pathname-location input-file)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  COMPILE  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
(defun translate-uri-callback (elem)
 ;;; :href "..." => :href #u"help:..." => :href #u"..." :onclick ...
  ;; CAUTION: The attrs of the elems are destructively modified!
  (when-let* ((attrs (html:elem-attrs elem))
              (desc (html:elem-desc elem))
              (indicators (html:desc-uri-attrs desc)))
    (let ((tail attrs)
          (new-attrs ())			; attributes added during translation
          (target nil)
          attr value)
      (loop
       (multiple-value-setq (attr value tail) (get-properties tail indicators))
       (if attr
           (let ((uri (if (stringp value)
                          (handler-case (uri:parse-uri value
                                         :decode *current-external-format*
                                         :fragment-char-p (lang-alpha-char-p-function
                                                           *language*))
                            ;; Ignore errors in parse-uri due to template pseudo-tags
                            (uri:uri-parse-error (condition)
                              (let ((use-templates (use-templates (asd:component-system
                                                                   *current-topic*))))
                                (unless (and use-templates
                                             (or (and (segment-property
                                                       (nth-segment *current-segment*
                                                                    *current-topic*)
                                                       :braces
                                                       (eq use-templates :braces) t)
                                                      (char-position #\{ value))
                                                 (search htt:*template-start-marker*
                                                         value)))
                                  (yl:logg :error condition)))
                              nil))
                          value)))
             (when uri
               (multiple-value-bind (uri attrs)
                   (translate-uri uri
                                  (setq target (getf attrs :target))
                                  (nth-segment *current-segment* *current-topic*))
                 (unless uri				; undefined - put or not indication
                   (return-from translate-uri-callback
                     (if *undefined-anchor-tag*		; a la html:make-elem
                         (values (cons *undefined-anchor-tag* (html:elem-cont elem))
                                 t)
                         (values (html:elem-cont elem)	; remove A and replace by contents
                                 :append))))
                 (setf (second tail) uri
                       new-attrs (append attrs new-attrs))))
             (setq tail (cddr tail)))
           (return)))
      (when (getf attrs :help)
        (remf attrs :help))			; delete pseudo-attribute
      (when (and target (string-equal target +target-popup+))
        (if (popup-p *current-topic*)		; if inside popup, replace the non-standard
            (setf (getf attrs :target) "_self")	; attribute value by the standard
            (remf attrs :target)))		; if outside, delete attribute at all
      (setf (html:elem-attrs elem) (nconc attrs new-attrs))))
  elem)

(defun delete-help-attribute-callback (elem)
 ;;; Delete :help pseudo-attribute if any
  (when-let (attrs (html:elem-attrs elem))
    (when (getf attrs :help)
      (remf attrs :help)
      (setf (html:elem-attrs elem) attrs)
      (values elem t))))

(defun entity-callback (elem)
 ;;; CAUTION: Pseudo-entities must not be used within headings text and anchor keywords.
  ;;          So we are not calling this during preprocess.
  #+debug (assert (eq (html:elem-key elem) :entity))
  (let ((value (first (html:elem-cont elem))))
    (cond ((stringp value)
           (values value			; strings can be used in keywords?
                   t))
          ((consp value)			; LHTML - make deep copy and translate
           (values (html::transform-elem (lambda (key attrs cont parents)
                                           (declare (ignore parents))
                                           (cond ((eq attrs t)
                                                  key)
                                                 ((memq key '(:a :img))
                                                  (translate-uri-callback
                                                   (html:make-elem key
                                                                   (copy-list attrs)
                                                                   cont)))
                                                 (t (html:make-elem key attrs cont))))
                                         value)
                   t))
          (t
           (values nil :append)))))		; delete strange expansion at all

(defmethod asd:output-files ((o (eql :compile)) (c parsed-html-file))
 ;;; We must have known all segments after compiling
  (declare (ignore o))
  (let ((acc ()))
    (do-segments (segment c)
      (push (asd:component-output-pathname segment) acc))
    (nreverse acc)))

(defmethod asd:perform ((o (eql :compile)) (*current-topic* parsed-html-file))
 ;;; Correct input format must be applied during preprocess
  (let* (;(html:*lhtml-not-safe* t)
         (*current-segment* 0)
         (*current-heading* 0)
         ;(template:*default-template-pathname* (pathname-location input-file))
         (book-system (asd:component-system *current-topic*))
         (*language* (language book-system))
         (charset (asd:component-property *current-topic* :charset))
         (*current-external-format*
          (if charset						; input html format
              (html:charset-ef charset)				; native format of the page
              (external-format book-system)))			; default for the book
         (use-templates (use-templates book-system))		; only body is expanded
         (callbacks (nconc (h-callback-alist *current-topic*)
                           (when (lang-shorten-span *language*)
                             (list '(:span . span-callback)))
                           '((:a        . translate-uri-callback) ; also delete :help
                             (:img      . translate-uri-callback)
                             ;(:link    . translate-uri-callback)
                             (:dfn      . delete-help-attribute-callback)
                             (:dt       . delete-help-attribute-callback)
                             (:comment	. comment-callback)
                             (:entity	. entity-callback)) ))
         (phtml:*end-tag-mismatch-action* nil))			; suppress same warnings
   (flet ((%parse (pathname &optional included)
            (log-info "compile~:[:~;+~] ~a" included pathname)
            (phtml:parse-html pathname
                              ;:parse-entities html:*lhtml-not-safe*
                              :callbacks callbacks
                              ;:collapse-whitespace nil
                              :external-format *current-external-format*)))
    (let* ((*include-function*
            (lambda (pathname)
              (case (file-type-format pathname)
                (:html
                 (values (%parse pathname t) :append)))))
           (lhtml (%parse (asd:component-pathname *current-topic*)))
           ;(navi-p (typep *current-topic* 'navigation-topic))
           (html (html:find-element :html lhtml))
           (head (html:find-element :head (html:elem-cont html)))
           (head-cont (html:elem-cont head))
           (body (html:find-element :body (html:elem-cont html)))
           (body-cont (html:elem-cont body))
           (charset (charset book-system))			; output charset
           (output-html-format (output-html-format book-system))
           (mode (or (asd:component-property *current-topic* :mode) :sgml))
           (prologue (not (asd:component-property *current-topic* :mode)))) ; prolog needed
      (unless (and lhtml (consp html))
        (error 'asd:compile-error :component *current-topic* :operation o))
      ;(when (asd::operation-original-initargs op :load)
      ;  (setf (lhtml *current-topic*) lhtml))

      (html:set-attribute html :lang *language*)
      ;; Move along the body, select a piece of its content - the segment body:
      ;;   [prev-divider]<segment body content>[next-divider]
      ;; - and replace the whole page body by this segment body.
      (when body
       (loop with tail = body-cont
             for output-file in (asd:output-files o *current-topic*)
             and rest = (segments *current-topic*) then (rest rest)
             for next = (first rest)
             and number upfrom 0
             do
             (html:with-html-file (output-file
                                   :mode (if (and (eq mode :xml) charset prologue)
                                             charset
                                             mode)
                                   :prologue prologue
                                   :external-format output-html-format)
               (let* ((member (if next (memq next tail)))	; look up for next divider
                      (cont (if member				; segment part of BODY 
                                (prog1 (ldiff tail member) (setq tail (rest member)))
                                tail))				; the last segment
                      (segment (nth-segment number *current-topic*))
                      (popups (segment-property segment :popups)))
                 ;; Append system aux-files references to head and body
                 (setf (html:elem-cont head) (head-cont-lhtml *current-topic* head-cont
                                                              :popups popups)
                       (html:elem-cont body)
                       `(,@(toolbar-lhtml *current-topic* :header segment
                                          :content cont)
                         ,@(if use-templates
                               (html:parse-htt cont
                                               :package (load-time-value
                                                         (find-package :ystok.help))
                                               :braces (segment-property segment :braces
                                                        (eq use-templates :braces)
                                                        t))
                               cont)
                         ,@(toolbar-lhtml *current-topic* :footer segment)
                         ,@(body-end-lhtml segment cont :popups popups)))
                 (if use-templates
                     (let* ((*current-book* book-system)
                            (*current-chapter* (asd:component-parent *current-topic*))
                            (environment (nconc (html:template-environment segment
                                                                          output-file)
                                                (environment book-system))))
                         #|(*include-function*
                            (lambda (stream type)
                              (log-info "compile-op+ ~a" (pathname stream))
                              (if (string-equal type "lhtt")
                                  (loop for lhtml = (read stream nil *unbound-value*)
                                        until (eq lhtml *unbound-value*)
                                        do (html:lhtt-print lhtml environment))
                                  (html:lhtt-print-list		; htm (prefered) or html
                                   (html:parse-htt (phtml:parse-html stream
                                                         :parse-entities nil
                                                         :callbacks callbacks)
                                                   :package (load-time-value
                                                             (find-package :ystok.help)))
                                   environment))))|#
                       (html:lhtt-print-list lhtml environment))
                     (html:html-print-list lhtml))))
             (log-info "compile-> ~a" output-file)
             finally (setf (html:elem-cont head) head-cont	; restore the originals
                           (html:elem-cont body) body-cont)) )))))
