;;; -*- Mode: Lisp; -*-
;;; YstokHelp - HTML-based help
;;; Copyright (c) 2009-2014 Dr. Dmitry Ivanov.  All rights reserved.
;;; Website: http://lisp.ystok.ru/yhelp/
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Run-time top-level API

;;; BAD:      YstokHelp, -
;;;           YSTOK.HTML.GENERATOR ?

(in-package :ystok.help)

(defvar *default-book* nil "Default instance of off-the-shelf book")

;; Directory where help sources reside
(defvar *default-book-path* nil "Default help books directory")

(defun load-book (pathname &optional book location)
 "Load compiled anchors file \"book-id.yha\" and replace
a relative value of the location slot by an absolute."
  (logg :help "Loading anchors file: ~a" pathname)
  ;; Avoid vague error message when the book has not installed at all.
  (if (probe-file pathname)
      (with-open-file (stream pathname :direction :input
                              :external-format (if book (external-format book) :default))
        (let* ((plist (read stream nil nil))				; read header
               (id (getf plist :id)))
          (cond ((and id (imp book (string-equal (id book) id)))
                 ;; Replace relative directory specification by an absolute pathname or URI
                 (unless location
                   (setq location (or (getf plist :location)		; external uri
                                      (pathname-location pathname))))
                 (setq plist (remove-properties plist '(:location)))
                 (if book
                     (apply #'reinitialize-instance book :location location plist)
                     (setq book (apply #'make-instance 'book :location location plist)))
                 (let ((anchors (make-hash-table :test #'equal))
                       (text-anchors (make-hash-table :test #'equal))
                       (%text% nil)
                       tuple)
                   (setf (anchors book) anchors
                         (text-anchors book) text-anchors)
                   (loop
                    (cond ((eq (setq tuple (read stream nil *unbound-value*))
                               *unbound-value*)
                           (return))
                          ((eq tuple :text)			; switch to context-enter
                           (setq %text% t))			; mode
                          (%text%
                           (setf (gethash (first tuple) text-anchors) (rest tuple)))
                          (t
                           (setf (gethash (first tuple) anchors) (rest tuple))))))
                 (register-book book))
                (t
                 (cerror #L"Ignore context help or book topic request."
                         'simple-help-error
           :format-control #L"Help book IDs are not equal - incorrect anchors file:~% ~a."
           :format-arguments (list pathname))))))
      (cerror #L"Ignore context help or book topic request."
              'simple-help-error :format-control #L"Help book is not found~@[: ~a~]." 
                              :format-arguments (list (or book (namestring pathname))))))

(defmethod ensure-book ((self book))
  (if (hash-table-p (anchors self))
      self
      (let ((pathname (make-pathname :name (id self) :type *anchor-file-type*
                                     :directory (pathname-directory
                                                 (pathname (location self)))))
            (default (thys-preference '*default-book-path*)))
        (load-book (if default (merge-pathnames pathname default) pathname)
                   self))))

(defmethod ensure-book ((id string))
  (ensure-book (find-book id)))

(defmethod absolute-pathname ((book book) (pathname string))
 ;;; Used for context
  (merge-pathnames (pathname pathname) (location book)))

(defmethod absolute-uri ((book book) arg)
 ;;; Args: arg  String or uri, which includes a filename.
  (uri:merge-uris arg (orf (book-uri book) (uri:uri (location book)))))

(defun book-page-princ-to-string (pair)
  (let (;(book (car pair))
        (page (cdr pair)))
    (if (and (consp page) (stringp (cdr page)))
        (cdr page)					; title
        (princ-to-string (first-or-self page)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  LOCATOR  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; NB: Directory-only locator "/path/" is unreliable because index.html may be
;;;	- either missing,
;;;	- or renamed and split into segments,
;;;	- or not considered by the browser a default page.

(defun path-match-p (pattern path)
 ;;; Similar to (pathname-match-p path pattern)
  ;; Args: pattern  String or list representing a parsed path from the locator;
  ;;		    can contain wildcard "*" as any element.
  ;;	   path     URI path of the page file where the anchor resides;
  ;;		    no book-id or :relative at head, contains only directory names;
  ;;		    the path is assumed relative to the book folder.
  ;; Value: True if the file lays on the path specified by pattern
  ;;	    from the root, i.e. the book folder itself.
  ;; NB: Returns true if the shortest common parts match and the pattern is shorter, e.g.
  ;;     (path-segment-match-p "a/b" '(a b c)) => true
  ;; CAUTION: On the contrary to path-segment-match-p, which takes a logical path
  ;;	      into account, this function interprets the path as physical!
  (do* ((x (if (listp pattern) pattern (split-seq #\/ pattern :keep-empty-subseqs nil))
           (rest x))
        (y (split-seq #\/ path :keep-empty-subseqs nil)
           (rest y)))
       ((not (and x y))
        (not x))			;   ..   
    (let ((first (first x)))
      (cond ((string= first "*"))
            ((not (string-equal first (first y)))
             (return nil))))))

(defun collect-paths (books locator &optional context)
 ;;; Args: books   Book instance or list of registered books
  ;;       locator URI designator or list of designators.
  ;;	       ::= "anchor" | "#anchor" | "//book#anchor" |
  ;;		   "//book#" | "/path" | "//book/path" |
  ;;		   "path#anchor" | "/path#anchor" | "//book/path#anchor"
  ;;       context If true, retrieve context pages
  ;; Values: 1) alist ::= ((book1 . page1) ...),
  ;;		where the page is either a string with a path relative to book
  ;;                  or uri, relative to book or absolute (when w.
  ;;         2) first anchor matched successfully
  ;; NB: Absolute and relative paths are usually treated the same
  ;;	 - relative to the book location, except for an absolute HTTP URI
  ;; Localization NB:
  ;;     When some locator is an unparsed URI string, we should know the language
  ;;     to parse fragment. In run-time, the application is assumed language-specific
  ;;     and we rely on features like :Russian rather than on book language.
  (let ((alist ())
        (external-format (and books (external-format (first-or-self books))))
        book-id path anchor)
    (dolist (elt (listify locator))
      ;; Q: Can elt be a pathname?
      ;; (multiple-value-setq (book-id path anchor) (parse-locator elt))
      (flet ((%from-uri (uri)
               (setq book-id (uri:uri-host uri)
                     path (uri:uri-parsed-path uri)
                     anchor (uri:uri-fragment uri))))
        (cond ((uri:uri-p elt)
               (if (or (null (uri:uri-scheme elt))
                             (eq (uri:uri-scheme elt) :help))
                   (%from-uri elt)
                   (nilf book-id path anchor)))
              ((not (or (char-position #\# elt)			; no delimeters
                        (char-position #\/ elt)))		; - treat as fragment only!
               (setq book-id nil
                     path nil
                     anchor (if (and external-format (char-position #\% elt))
                                (uri:percent-decode elt :external-format external-format)
                                elt)))
              (t
               (%from-uri
                (setq elt (uri:parse-uri elt ;:errorp nil
                           :decode external-format
                           :fragment-char-p #-Russian #'latin-1-alpha-char-p
                                            #+Russian #'russian-alpha-char-p))))))
      ;(when (eq (first path) :relative)
      ;  (logg 'help #L"Relative path is not acceptable in locator: ~a" elt)
      ;  (return))
      (dolist (book (mapcar #'find-book (listify (or book-id books))))
        (cond ((null (setq book (ensure-book book))))
              (anchor
               (dolist (page (listify (gethash anchor
                                               (if context
                                                   (text-anchors book)
                                                   (anchors book)))))
                 (when (or (null path)
                           (path-match-p (rest path) (first-or-self page)))
                   (setq alist (acons book page alist)))))
              ((not context)	       				; collect locator itself 
               (when (and (uri:uri-p elt)			; treat absolute and
                          (or (null (uri:uri-scheme elt))	; relative paths the same
                              (eq (uri:uri-scheme elt) :help))	; - relative to the book
                          (eq (first path) :absolute))		; location
                 (setq elt (uri:make-uri :%parsed-path% (cons :relative (rest path)))))
               (setq alist (acons book elt alist)))))
      (when alist						; the most relevant anchor
        (return)))						; found - loop exit
    (values (nreverse alist) anchor)))

(defun read-context (book pathname start end
                     &optional (external-format (external-format book)))
 ;;; Values: 1) string or NIL 2) length of the string
  ;; Q: Is it of use a single-context file, i.e. start and end are not required?
  ;;    It could be read by means of hcl:file-string or the like.
  (when (and (not (directory-pathname-p pathname)) (probe-file pathname))
    (logg :help "Reading context file: ~a (~d . ~d)" pathname start end)
    (with-open-file (stream pathname :direction :input :element-type 'octet)
      (when (or (zerop start) (file-position stream start))
        (let* ((octets (make-array (- end start) :element-type 'octet))
               (size (read-sequence octets stream)))
          (declare (type octet-vector octets) (dynamic-extent octets))
          (when (iplusp size)
            (octets-to-string octets :end size :external-format external-format)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  TOP-LEVEL FUNCTIONS  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun view-topic (locator &key (book (or *default-book* *registered-books*))
                                interface displayer)
 "Open a help book to the page identified by the locator given.
When several pages match the locator, ask the user to select.
Arguments:
 locator
   Locator designator or list of designators in descending relevance order.
   Identifies help topic by mentioning an anchor and, optionally,
   a book id and/or chapter/folder name, see parse-locator.
 book
   Should only be passed if it is known exactly;
   ignored if a book id is hard-wired within the locator argument.
 interface
   Top-level interface the user has asked for help from."
  ;; NB1: Allows to access a help topic with or without the anchor name.
  ;;	  In most cases, using anchors is easier and more flexible than
  ;;	  tracking the address of the topic file required by view-page.
  ;; NB2: If the locator designates several pages, the function prompt what exactly.
  ;; NB3: Using only directory locator /path/ is unreliable because index.html may be
  ;;	  - either missing,
  ;;	  - or renamed and split into segments index_00.html, index_01.html etc.
  (multiple-value-bind (book-pages anchor) (collect-paths book locator)
    (if book-pages
        (when-let* ((pair (if (rest book-pages)
                              (progn (ywi:beep :query interface)
                                (capi:prompt-with-list book-pages
                                   #L"Select help book topic"
                                   :pane-args '(:print-function book-page-princ-to-string)
                                   :popup-args `(:cancel-button #L"Cancel"
                                                 ,@(when interface
                                                     `(:owner ,interface
                                                       ,@ywi:+popup-args+)))))
                              (first book-pages)))
                    (book (car pair))
                    (page (cdr pair))
                    (u (uri:uri (absolute-uri book (first-or-self page)))))
             ;(case (uri:uri-scheme uri:uri)
             ;  (:file (setq uri:uri (pathname (subseq (uri:uri-path uri:uri) 1))))
             ;  (otherwise ))
          (when anchor
            (uri:clear-cache u)
            (setf (uri:uri-fragment u) anchor))
          (logg :help "Viewing help locator ~s -> ~a" locator u)
          (run interface 'viewer-open u))
        (ywi:beep-box #L"No help book topics found~@[ for ~s~],~%locator:~2% ~a."
                      :format-arguments (cons (ywi:object-title displayer)
                                              (listify locator))
                      :icon :query  :owner interface)))
  (values))

(defun view-page (&key ((:uri u) nil) (book *default-book*) interface new)
 "Open a help book page at a known URI."
 ;;; Args:
 ;;;  book  Book identifier or book instance, *default-book* by default.
  ;;  uri   Can include path and fragment. The path is
  ;;       - either relative if book is not null,
  ;;	   - or absolute URI or string if book is null.
  ;;	   If the "#fragment" specified in the URI, help viewer opens
  ;;       the page and scrolls to the section identified by the anchor.
  ;;  new  It true, open the page in a fresh window (a la target="_blank").
  ;;
  ;; NB: This function requires you to know the full or partial path to the HTML file
  ;;	 describing the desired help topic. You can specify the location of the page
  ;;	 using either a full file://URL, or http://URL,
  ;;	 or a combination of a book ID and a relative URI.
  ;;	 Relative paths should be specified relative to the help book's folder.
  (when-let (pathname
             (cond (u
                    (if book
                        (absolute-uri (ensure-book book) u)
                        u))		; the only case where pathname type is possible!
                   (book
                    (setq book (ensure-book book))
                    (absolute-uri book (start-topic book)))
                   (t
                    (cerror #L"Ignore context help or book topic request."
                       'simple-help-error
                       :format-control #L"Neither URI nor BOOK is given to VIEW-PAGE."))))
    (logg :help "Viewing help page ~s -> ~a" u pathname)
    (run interface 'viewer-open pathname :new new)))

#||
#+unused
(defgeneric parse-locator (locator)
 (:documentation
 "Arguments: 
  locator String or URI instance as defined for the help: schema.
    \"anchor\" | \"#anchor\"| \"//book#anchor\" |
    \"//book#\" | \"/path\" | \"//book/path\" |
    \"/path#anchor\" | \"//book/path#anchor\"
Values:
 1) book-id or NIL,
 2) path as string without leading '/' or NIL,
 3) anchor or NIL.")
 ;; The path and anchor returned are not percent encoded.
 (:method ((string string))
  (let* ((sharp (char-position #\# string))
         (slash (char-position #\/ string))
         (fragment (cond (sharp (subseq string (1+ sharp)))
                         ((not slash) string)))			; entire string
         (book-id (cond (slash (if (plusp slash) (subseq string 0 slash) nil))
                        ((iplusp sharp) (subseq string 0 sharp))))
         (path (if slash (subseq string (1+ slash) sharp) nil)))
    (values book-id path (if (zerop (length fragment)) nil fragment))))
 (:method ((arg uri:uri))
  (let* ((path (uri:uri-parsed-path arg))
         (book-id (case (first path)
                    (:relative (prog1 (second path) (setq path (cddr path))))
                    (:absolute (prog1 nil (setq path (cdr path))))))
         (fragment (uri:uri-fragment arg)))
    (values book-id
            (uri::%render-parsed-path path)
            (if (zerop (length fragment)) nil fragment)))) )

#+unused
(defun search-text (pattern &optional (book *default-book*))
 "Search the help book for a given text string.")
||#