;;; -*- Mode: Lisp; -*-
;;; YstokHelp - HTML-based help
;;; Copyright (c) 2009-2014 Dr. Dmitry Ivanov.  All rights reserved.
;;; Website: http://lisp.ystok.ru/yhelp/
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Table-of-contents (sitemap) compiler
;;; It can produce the TOC of an entire book or a chapter (not tested).

(in-package :ystok.help)

(defun h-level-key (level)
  (svref +heading-keys+ (1- level)))

(defun h-key-level (key)
  (- (char-code (schar (string key) 1)) #.(char-code #\0)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  TOC Class  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Only TOC for the entire book within the navi frame is tested

(defclass toc (navigation-mixin generated-component)
 (;; The source of items collected in TOC
  ;; :title - from topic/segment titles
  ;; (:h1 ... [(tags-with-explict-toc-level-only)]) - list of heading tag keywords
  (toc-source :reader toc-source :initarg :source :initform *default-toc-source*)
  ;; If null, TOC is not created
  (toc-max-level :reader toc-max-level :initarg :max-level
                 :initform *default-toc-max-level*)
  ;; CSS class pattern
  (heading-classes :reader heading-classes :initarg :heading-classes  :initform :all))
 (:default-initargs
  :tab "Sitemap"
  :title "Sitemap"
  :description "Detailed table of contents"))

(defun find-toc (module)
  (find 'toc (asd:module-components module) :test #'xtypep))

(defmethod toc-source ((component topic))
  (toc-source (asd:component-parent component)))
(defmethod toc-source ((module asd:module))
  (let ((toc (find-toc module)))
    (cond (toc (toc-source toc))
          ((setq toc (asd:component-parent module)) (toc-source toc)))))

(defmethod toc-max-level ((component topic))
  (toc-max-level (asd:component-parent component)))
(defmethod toc-max-level ((module asd:module))
  (let ((toc (find-toc module)))
    (cond (toc (toc-max-level toc))
          ((setq toc (asd:component-parent module)) (toc-max-level toc)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;   toc-source :title  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;      :
;;;  (1)       = 1 (   ),
;;;  (2)    / =  ,
;;;  (3)     / =   + 1
;;;
;;;      
;;;  -    =  
;;;  -   
;;;  -- (1,2) =   + 1
;;;  -- (3) =   =   + 1

(defun segment-title-toc-level (self)
 ;;; Compute level i for Hi-element to put into TOC when toc-source :title.
  ;; Args: self Instance of segment, topic, or text-topic
  ;; Value:    
  (let* ((topic (source-component self))
         (segment-number (segment-number self))
         (parent (asd:component-parent topic)))
    (if (eq parent (asd:component-system topic))
        ;; for top-level topic, H1 or H2 no mater start or not but depends on segment
        (if (iplusp segment-number) 2 1)
        (min 6 (max 1 (- (length (component-path topic))
                         (if (eq topic (start-topic parent))
                             (if (iplusp segment-number)
                                 1	; for non-first segment, chapter-level + 1
                                 2)	; for start topic, chapter-level
                             1)))))))	; for others topics, chapter-level + 1

;;;;;;;;;;;;;;;;;;;;;;;;;;;   toc-source headings  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 1.  ,     1 ( ),
;;;       - h<i>  .
;;; 2.      (   1,  1.1  )
;;;           :
;;;	-  ,   :	h<i> -> body.toc h<i>
;;;	-    :		h<i> -> body.toc h<i>
;;;     -    :		h<i> -> body.toc h<i+1>
;;; 3.         toc-level
;;;     help  ""     .
;;;
;;; :
;;;	H1 -     
;;;	   -    .
;;;	H2 help="toc-level:-1"
;;;	   -      /
;;;	     (   "").
;;;	H2 -     
;;;	   -      
;;;	   -      /.
;;;	H2 help="toc-level:+1"
;;;	   -      (
;;;	        "", ,  ).
;;;	H3 help="toc-level:n" (  toc-source '(:h1 :h2 (:h3))
;;;	   -    ,   
;;;	          (  );
;;;	Hi help="toc-level:+0"
;;;	   -  ,     ""  .
;;;	H4 -     ,    
;;;	H5 - ,    
;;;	H6 - , , 
;;;	     
;;; TODO:   heading-cont inline-, : <code>, <span>  ..

(defun segment-toc-level (topic segment-number)
 ;;; Args: segment-number NIL during :compile text topic or :preprocess any
  (let* ((path (component-path topic))
         (length (length path)))
    (if (<= length 2)						; top level topic
        1							; always 1
        (- length
           (- (dsum #'toc-level path :default 0))		; count chpaters delta
           (if (or (eq topic (start-topic (asd:component-parent topic)))
                   (zerop (or segment-number (length (segments topic)))))
               2
               1)))))

(defun clear-headings (topic-mixin)
  ;; Q: Unregister anchors?
  ;; We don't know from where the heading-id has come: from explicit id or inside A
  (setf (headings topic-mixin) nil))

(defun collect-headings-callback (elem &optional explicit)
 ;;; Used when toc-source = headings-list during both load-source-op and compile-op
  ;; Args: explicit  Collect only if help="toc-level:[+|-]number" attribute is present.
  ;; TODO: Attribute :class ::= "figure" or "table"
  (let* ((key (html:elem-key elem))
         (h-key-level (h-key-level key))
         (segment-toc-level (segment-toc-level *current-topic* *current-segment*))
         (attrs (html:elem-attrs elem))
         (id (getf attrs :id))					; explicit anchor
         (plist (phtml:parse-declarations (getf attrs :help)))
         (toc-level (getf plist :toc-level *unbound-value*))	; nil if "none" or empty
         (replace nil))
   (when (and *current-segment* plist)				; :compile
     (remf attrs :help)						; delete pseudo-attribute
     (setf (html:elem-attrs elem) attrs))
   (when (and id *index-heading-anchors* (not *current-heading*))
     ;; preprocess: register explicit id as anchor, put it into topic's anchor list
     (register-anchor id elem attrs
                      (if (null (getf plist :keywords *unbound-value*))
                          nil					; explicitly suppress index
                          *index-heading-anchors*)
                      ;((eq *index-heading-anchors* :keywords) :keywords)
                      ;(t))
                      plist)
     (pushnew id (anchors *current-topic*) :test #'string=))
   (when toc-level						; not suppressed via none
    (let ((level (min 6 (max 1					; compute final level
                  (cond ((eq toc-level *unbound-value*)		; unspecified
                         (if explicit				; do not collect
                             (return-from collect-headings-callback)
                             #1=(+ h-key-level segment-toc-level -1)))
                        ((digit-char-p (schar toc-level 0) 10)	; absolute
                         #2=(parse-integer toc-level :radix 10 :junk-allowed t))
                        (t (+ #1# #2#))) ))))			; relative
     (when t ;(<= level (toc-max-level *current-topic*))
       (if *current-heading*
	   ;; compile: provide id if none
           (let ((heading (nth *current-heading* (headings *current-topic*))))
             (unless (or id
                         (getf (html:elem-attrs (html:find-element :a
                                                                   (html:elem-cont elem)
                                                                   :name))
                               :name))
               (setf (getf attrs :id) (heading-id heading)	; supply with id attribute
                     (car elem) (cons key attrs)		; destructively modify
                     replace t))
             (incf *current-heading*))
           ;; preprocess
           (let* ((segment (if *current-segment*
                               (nth-segment *current-segment* *current-topic*) ; compile-op
                               (last-segment *current-topic*)))		; load-source-op
                  (headings (headings *current-topic*))
                  (cont (html:elem-cont elem))			; canonicalize heading text
                  (text (getf plist :text))			; explicit title text
                  (bbcode (and text
                               (getf plist :bbcode *bbcode-in-text*)
                               (char-position #\[ text))))
             (unless id						; no explicit anchor found
               (if-bind (a (html:find-element :a cont :name))	; search for inside anchor
                 (setq id (getf (html:elem-attrs a) :name))	; found - just use it
                 (setf id (format nil "_h~d" (length headings))	; not found - generate id
                       (getf attrs :id) id			; and equip with attribute
                       (car elem) (cons key attrs)		; destructively modify
                       replace t)))
             ;; If no title specified in HTML and this is the very first heading
             ;; in the segment, use the heading text as a title
             ;; NB:        ,
             ;;          .
             (unless (or (segment-property segment :title)
                         (find segment headings :key #'heading-segment))
               (setf (segment-property segment :title)
                     (cond ((not text)
                            (interleave-seq 'string #\Space
                                            (lhtml-word-list (distil-content cont))))
                           (bbcode
                            (interleave-seq 'string #\Space
                                            (lhtml-word-list (html:bbcode-lhtml text))))
                           (t text))))
             (if text					; if text property is given,
                 (setq cont nil				; neither element content
                       plist nil)			; nor category property is used
                 (multiple-value-bind (content new-props) (distil-content cont)
                   (setq cont content
                         plist (append (retain-properties
                                        plist *anchor-span-css-class-properties*)
                                       new-props))))
             (setf (headings *current-topic*)
                   (nconc headings (list (make-heading :segment segment
                                           :tag key
                                           :level level
                                           :id id		; no text - save :class
                                           :class (if text nil (getf attrs :class))
                                           :plist plist
                                           :content (cond ((not text) cont)
                                                          (bbcode (html:bbcode-lhtml text))
                                                          (t (list text)))
                                           :explicit (neq toc-level *unbound-value*)))))))
       ;; Return the element and replace flag
       (values elem replace))))))

(defun collect-headings-explicit-callback (elem)
 ;;; For elements grabbed only when accompanied by help="toc-level:n"
  ;; Called by: For elements in the last element of toc-source, e.g. (:H3)
  (collect-headings-callback elem t))

(defgeneric h-callback-alist (parsed-html-file &optional toc-source)
 (:method (self &optional toc-source)
  (declare (ignore self toc-source))
  ()))

(defmethod h-callback-alist ((topic topic) &optional (toc-source (toc-source topic)))
                                           ;(toc-max-level (toc-max-level book)))
 ;;; Compute the alist of callbacks passed to parse-html.
  ;; Args: *current-heading*  NIL means :preprocess
  ;;			      non-NIL - :compile
  ;; Q:            ?
  (when (or (consp toc-source) *index-heading-anchors*)
    (loop ;for level upfrom 1 upto toc-max-level for key = (h-level-key level)
          with last = (first (last toc-source))
          with explicit = (if (consp last) last)
          for key across +heading-keys+
          if (memq key toc-source)
            collect `(,key . collect-headings-callback)
          else if (memq key explicit)
            collect `(,key . collect-headings-explicit-callback)
          else if (and *index-heading-anchors* (not *current-heading*))	; :preprocess
            collect `(,key . collect-anchors-callback))))

(defmethod toolbar-lhtml ((component toc) (position (eql :header)) segment
                          &key (book (asd:component-system component)))
  (declare (ignore segment position))
  (navi-tabs-lhtml component book))

;;; Top-level

(defmethod create-toc ((toc toc) &key (source (toc-source toc))
                                      (max-level (toc-max-level toc))
                                      (pathname (asd:component-output-pathname toc))
                                      ;(pathname (merge-pathnames
                                      ;             (make-pathname :name (toc-file book)
                                      ;                            :type "html")
                                      ;             (asd:component-output-pathname book)))
                                      (frame :navi)
                                      (heading-classes (heading-classes toc)))
 ;;; Create toc file in the book output directory.
  ;; Args: frame ::= :navi TOC is created for the navi frame (default),
  ;;		  or :cont TOC is created for the cont frame
  ;;       heading-classes
  ;;		  CSS-class pattern: string, keyword or list (see class-attr-match-p).
  (when max-level
   (let* ((book (asd:component-system toc))
          (*language* (language book)))
    (html:with-html-file (pathname :external-format (output-html-format book)
                                   :mode (or (charset book) :xml)
                                   :prologue t)
     (log-info "creating TOC-> ~a" pathname)
     (html:html (:lang (:optional *language*)
                 :xml\:lang (:optional *language*)
                 :xmlns html::+xmlns+)
      (case frame
        (:navi (navi-html-head book :title (description toc)))
        (otherwise (error "Not-implemented TOC frame ~s" frame)))
      (html:body (:class (:optional *toc-body-css-class*))
        (html:html-print-list (toolbar-lhtml toc :header nil))
        ;(html:hr ())
        (terpri html:*html-stream*)

        (loop with base = (pathname-location pathname)
              ;and target = (if (eq frame :navi) "navi" nil)	; target for hrefs
              for topic in (topics book)
              do
              (flet ((%output-h (segment level content
                                 &optional fragment class plist explicit)
                       (when (or explicit (<= level max-level))
                         (let ((uri (uri:enough-uri (asd:component-output-pathname segment)
                                                    base))
                               (description (description segment)))
                           ;; Suffix span
                           (multiple-value-bind (span-class span-cont)
                               (get-properties plist *anchor-span-css-class-properties*)
                             (when fragment
                               (setf (uri:uri-fragment uri) fragment))
                             (html:html-print
                              `(,(h-level-key level)	;:id id -  "  "
                                ((:a :href ,uri
                                  ,.(when (stringp class)
                                      (when-let (classes (intersection
                                                          (split-seq-if #'whitespace-char-p
                                                                        class)
                                                          *anchor-span-css-classes*
                                                          :test #'string-equal))
                                        `(,:class ,(interleave-seq 'string
                                                                   #\Space classes))))
                                  ,.(when description `(,:title ,description)))
                                  ;,.(when target `(:target ,target)))
                                 ,@content
                                 ,@(when span-class
                                     `(,#\Space
                                       ,(html:make-elem :span
                                                        (list :class span-class)
                                                        (listify span-cont)))))))
                             (terpri html:*html-stream*))))))
                (if (eq source :title)
                    (do-segments (segment topic)
                      (%output-h segment
                                 (segment-title-toc-level segment)
                                 (list (title segment))))
                    (dolist (heading (headings topic))
                      (when (class-attr-match-p heading-classes (heading-class heading))
                        (%output-h (heading-segment heading)
                                   (heading-level heading)	;+ topic-level
                                   (heading-content heading)
                                   (heading-id heading)
                                   (heading-class heading)
                                   (heading-plist heading)
                                   (heading-explicit heading)))))))
	;; Add "Index" at the very bottom
        (when-let (index (find-index book))
          (html:htm
           (:h1 ((:a :href (uri:enough-uri (index-start-pathname index)
                                           (asd:component-output-pathname toc))
                     :target "_self")
                 (:princ-safe (title index))))
           :newline))
  )))))
  (values))

(defmethod create-toc ((book book-system) &rest args &key &allow-other-keys)
  (when-let (toc (find-toc book))
    (apply #'create-toc toc args)))

(defmethod create-toc (name &rest args &key &allow-other-keys)
  (when-let (book-system (asd:find-system name t))		;find-book
    (apply #'create-toc book-system args)))


;(defmethod asd:output-files ((o asd:compile-op) (c toc))
;  (list (asd:component-output-pathname c)))

(defmethod asd:perform ((o (eql :compile)) (c toc))
  (declare (ignore o))
  (create-toc (asd:component-parent c)
              :source (toc-source c)
              :max-level (toc-max-level c)
              :pathname (asd:component-output-pathname c)))
