;;; -*- Mode: Lisp; -*-
;;; YstokHelp - HTML-based help
;;; Copyright (c) 2009-2014 Dr. Dmitry Ivanov.  All rights reserved.
;;; Website: http://lisp.ystok.ru/yhelp/
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Help Viewer - Web browser automation via DDE

#|
Q160957 - INFO: DDE Support in Internet Explorer Versions
Last Reviewed: July 23, 1999
Copyright 1999 Microsoft Corporation. All rights reserved. Terms of Use.

NB: All strings passed to DDEML using DdeCreateStringHandle are limited
    to 255 characters. When passing item names to any of the topics,
    strings greater than 255 characters long are truncated.

In every DDE transaction, you need to specify the following: 
   ApplicationName : iexplore
   DDE Transaction Type: XTYP_EXECUTE or XTYP_REQUEST

Topics (not commands!)
------
WWW_OpenURL
   itemName: 	URL  or  URL,[FileSpec],WindowKind
   returnValue:	-2 for acknowledgment, -3 for error 

   FileSpec   is ignored.
   WindowKind 0  open a new window,
	     -1 open the URL in the most recently active browser window.
   Parameters following the WindowKind are ignored.

  Retrieves a Web page at URL displays it
  Example: "http://www.microsoft.com",,0
  DI 2009-Jan-14: In Registry HKCR/htmlfile/shell/open/ddeexec "file://%1",,-1,,,,,

WWW_ShowFile
   itemName:	FileSpec
   returnValue: -2 for acknowledgment, -3 for error 

  Passes FileSpec to be rendered in a given WindowID: 

WWW_Activate
   itemName:	WindowID  - any value, ignored
   returnValue:	WindowID 

  Brings the most recently active browser window to the foreground
  NB: WindowID returned is synthetic; that is, it can only be used in other DDE 
      transactions with Internet Explorer. It should not be assumed as an HWND.

WWW_Exit
   itemName: WWW_Exit
   returnValue: 0 

  Tells the browser to shut down and exit

WWW_RegisterURLEcho
   itemName: ApplicationName (enclosed in double quotes)
   returnValue: Non-Zero for success, 0 for failure 

  Requests that ApplicationName be notified of URLEcho events whenever a URL
  gets loaded. Multiple servers can be registered to get these notifications: 
  ApplicationName is a DDE server that is registered to receive WWW_URLEcho
  notifications from the browser with the following information:

   topicName (hsz1): WWW_URLEcho
   itemName  (hsz2): URL,MIMEType,WindowID
   Example: "http://www.home.com","text/html",-1
   Transaction: XTYP_POKE 

WWW_UnRegisterURLEcho
   itemName:	ApplicationName (enclosed in double quotes)
   returnValue:	success/failure. 

  Terminates associated request for WWW_URLEcho events notification

WWW_ListWindows
   itemName:	WWW_ListWindows
   Values:	NumItemsInTheList,ArrayOfWindowIDs, i.e. [count,winID,winID,winID]

  Returns a list of window IDs for windows currently used by the browser.

  Currently Internet Explorer does not return a list of window IDs,
  but rather a list corresponding to the number of instances of the
  browser that are currently opened. For additional information, see:
  Q191508 - PRB: Confusing Return Values from DDE topic WWW_ListWindows

WWW_GetWindowInfo
   itemName:	WWW_GetWindowInfo , "DoGetWindowInfo"
   Values:	URL,WindowText

  Returns URL and WindowText currently being displayed in the browser window
  Example: "http://www.microsoft.com/","Microsoft Home Page" 

  NB: The strings are quoted; therefore, double quotes are denoted with backslashes.
      For example, if the title consists of the three characters !"!,
      the title portion of the returned string as the six characters "!\"!".

SUMMARY TABLE                        16bit     16bit      SP1 Beta
                                3.02 3.03 4.0  4.0  4.01 4.01  5.0
     DDE Topic                        *   1712 PP2  2106 3110  626
   ___________________________  __________________________________
   WWW_Activate                  X    X    X    X    X    X     X
   WWW_Exit                      X    X    X    X    X    X     X
   WWW_GetWindowInfo             X    X        4.x (Q178879)
   WWW_ListWindows               X    X    X    X    X    X     X
   WWW_OpenURL                   X    X    X    X    X    X     X
   WWW_RegisterURLEcho           X    X    X    X    X    X     X
   WWW_ShowFile                  X    X    X    X    X    X     X
   WWW_UnRegisterURLEcho         X    X    X    X    X    X     X
   
For additional information, please see in the Microsoft Knowledge Base: 
Q160976 - How to Control the Currently Running Instance of IE3 via DDE 
Q191508 - Confusing Return Values from DDE topic WWW_ListWindows 
|#

;;; MS Internet Explorer
;;;   - The page is open either in a new or the most recently used browser window.
;;;   - A fragment works when is opened in a new window, but is effectively
;;;	ignored when open in an existing browser window.
;;;
;;; Firefox
;;;   - The page is always opened in a new tab.
;;;   - A fragment always works.
;;;
;;; We hope win32:dde-execute-command ensures these:
;;; - Passing NULL in argument list substitute empty string,
;;; - Quotation marks are needed only if a parameter contains white space.
;;;
;;; NB1: If Viewer exits normally, it is likely to notify the client,
;;;      and the conversation is marked as :destroyed. If Viewer terminates abnormally,
;;;      it obviously cannot notify and LW signals:
;;;	    #<win32::simple-ddeml-error win32::error-code 16396>
;;;		DDE: DDE execute #<DDE conversation (Client) VIEWER|SERVICE 0xDESTROYED>
;;;		 "[WWW_OpenURL(\"uri\")]" failed
;;; NB2: Unfortunately, the timeout argument of dde-execute-command specifies a command
;;;      timeout, not a connection timeout. This leads to hanging on the dde-connect stage
;;;      when Viewer is not running or even when it is running (due to Windows bugs?).
;;;	 dde-execute-command => T,   if the command accepted but did not return any result,
;;;			     => NIL, if the command is illegal.


(in-package :ystok.help)

(defvar *viewer-pathname* nil)
(defvar *viewer-service* nil)
(defvar *dde* t)		; Use DDE

(defparameter *browsers* '("Internet Explorer/iexplore.exe"	; list of strings
                           "Mozilla Firefox/firefox.exe"))	; or pathnames

(defun default-viewer-pathname ()
 ;;; Value: #P"D:\Program Files\Internet Explorer\iexplore.exe"
  ;; See also: lw:*browser-location*
  (loop with base = (yl:get-folder :program-files)		; Windows only
        for path in *browsers*
        thereis (probe-file (merge-pathnames path base))))

(defun default-viewer-service (&optional (pathname(yl:thys-preference '*viewer-pathname*)))
  (pathname-name pathname))					; iexplore or firefox


(defvar *viewer-conversation* nil)				; DDE conversation

(defmacro with-conversation ((conversation service &optional topic) &body body)
 ;;; Args: conversation Special variable
  ;; Value: nil = error, and caller should launch the viewer again.
  ;; NB1: All commands should be executed within the same thread as the dde-connect.
  ;; NB2: Usually, we are unlikely to be able to reconnect, so return nil.
  ;; NB3: This code runs in a dedicated process, it must handle all errors.
  (with-gensyms (block)
   `(when (if ,conversation
              (progn (when ,topic
                       (setf (win32::dde-conversation-topic-name ,conversation) ,topic))
                t)
              (setq ,conversation (win32:dde-connect ,service ,topic :errorp nil)))
      (block ,block
        (restart-case
            (handler-bind
                ((fli:foreign-type-error
                  (lambda (c)			; Browser has been closed normally
                    (if (and ,conversation
                             (eq (win32::dde-conversation-hconv ,conversation)
                                 :destroyed))      ;(eq (type-error-datum c) :destroyed))
                        (progn
                          (win32:dde-disconnect ,conversation)
                          (return-from ,block (setq ,conversation nil)))
                        (ywi:prompt-for-restart c))))
                 (WIN32::DDEML-ERROR
                  (lambda (c)
                    (case (slot-value c 'WIN32::ERROR-CODE)
                      ((16385 16389 16393)
		       ;; unexplainable signal - ingore it as if OK
                       (return-from ,block t))
                      (otherwise
                       (ywi:prompt-for-restart c)))))
                 (WIN32::DDE-ERROR
                  #'ywi:prompt-for-restart)
                 (error
                  #'ywi:prompt-for-restart))
                     ;(lambda (c) (ywi:prompt-for-restart c :quit t))))
              ,@body)
          (continue ()
            :report #L"Retry to start Internet browser once again."
            (when ,conversation (win32:dde-disconnect ,conversation))
            (setq ,conversation nil)))))))

(defun viewer-execute-command (topic &rest args)
 ;;; Value: nil = error, and caller should launch the viewer again.
  (with-conversation (*viewer-conversation* (yl:thys-preference '*viewer-service*) topic)
    ;(win32:dde-execute-command *viewer-conversation* command args
    (win32:dde-execute-string *viewer-conversation* (interleave-seq 'string #\, args)
                              :errorp t
                              :timeout (thys-preference 'dde-timeout))))

(defun viewer-request (topic item &key (format :text) (type nil supplied-p))
 ;;; Args: type ::= :string-list
  ;; Values:
  ;;  1) result
  ;;  2) successp, nil = error, and caller should launch the viewer again.
  (with-conversation (*viewer-conversation* (yl:thys-preference '*viewer-service*) topic)
    (apply 'win32:dde-request *viewer-conversation* item :format format ;:type type
           :errorp t
           :timeout (thys-preference 'dde-timeout)
           (when supplied-p (list :type type)))))

(defmethod viewer-open (arg &key new (hwnd 0) (encode t))
 ;;; Args: arg    Instance of uri or string
  ;;	   new    True = open in a new browser window
  ;;       encode Usually the value of output-html-format.
  ;; NB: win32:shell-execute => value greater than 32 if successful
  (let ((s (if (stringp arg)
               arg
               (uri:render-uri arg nil encode))))
    (or (and (thys-preference '*dde*)
             (viewer-execute-command "WWW_OpenURL" s +null-string+ (if new 0 -1)))
        (if-bind (pathname (thys-preference '*viewer-pathname*))
 #+win32  (> (win32:shell-execute hwnd "open" (namestring pathname)
                                  (string-append #\" s #\") nil win32:sw_show)
             32)
 #-win32  (sys:call-system (format nil "~a \"~a\"" (namestring pathname) s))
 #+win32  (> (win32:shell-execute hwnd "open" s nil nil win32:sw_show) 32)
 #-win32  nil))))

(defmethod viewer-open ((pathname pathname) &key (hwnd 0))
 ;;; Args: pathname Means that we call ShowFile.
  ;; NB: win32:shell-execute => value greater than 32 if successful
  (let ((s (namestring pathname)))
    (or (and (thys-preference '*dde*)
             (viewer-execute-command "WWW_ShowFile" s))
        (if-bind (pathname (thys-preference '*viewer-pathname*))
 #+win32  (> (win32:shell-execute hwnd "open" (namestring pathname)
                                  (string-append #\" s #\") nil win32:sw_show)
             32)
 #-win32  (sys:call-system (format nil "~a \"~a\"" (namestring pathname) s))
 #+win32  (> (win32:shell-execute hwnd "open" s  nil nil win32:sw_show)
             32)
 #-win32  nil))))

(defun viewer-exit ()
  (and (thys-preference '*dde*)
       *viewer-conversation*
       (viewer-execute-command "WWW_Exit")
       (win32:dde-disconnect *viewer-conversation*)
       (setq *viewer-conversation* nil)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  RUNNER  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The help runner loops over requests calling mp:mailbox-read
;;; and invokes the underlying function.

(defvar %viewer-mailbox% nil)			; Dedicated process mailbox

(defun viewer-runner ()
  (loop
   (destructuring-bind (interface function . args) (mp:mailbox-read %viewer-mailbox%)
     (yl:debug-format :help "~s: ~s ~:s" %viewer-mailbox% function args)
     (restart-case (handler-bind ((error #'ywi:prompt-for-restart))
                     (let* ((title function) ;(report-title report args)
                            (value (apply function args)))
                       ;(declare (ignore value))
                       (when (and interface (capi:interface-message-area interface))
                         (capi:execute-with-interface interface 'ywi:message-area-format
                           interface
                           #1=#L"Help function~@[ ~s~] ~:[was canceled~;is complete~]."
                           (if (stringp title) title nil)
                           value))))
       (abort () :report (list #L"Cancel help function ~s." function)
                               ;(report-title report args)	;(ignore-errors ?)
              (when (and interface (capi:interface-message-area interface))
                (capi:execute-with-interface interface 'ywi:message-area-format
                  interface #1# function nil))
              t)))
) )

(defun ensure-viewer-mailbox ()
 ;;; Ensure per thread viewer conversation handle
  (unless %viewer-mailbox%
    (setq %viewer-mailbox% (mp:make-mailbox)))
  (unless (and (mp:process-alive-p (mp:mailbox-reader-process %viewer-mailbox%))
               (equal (mp:process-name (mp:mailbox-reader-process %viewer-mailbox%))
                      #1="Help viewer runner"))
    (setf (mp:mailbox-reader-process %viewer-mailbox%)
          (let ((mp:*process-initial-bindings* (acons '*viewer-conversation* nil
                                                      mp:*process-initial-bindings*)))
            (mp:process-run-function #1# (list :mailbox %viewer-mailbox%)
                                     'viewer-runner))))
  %viewer-mailbox%)

(defun run (interface function &rest args)
 ;;; Top-level function to execute help functions
  ;; Args: interface  CAPI interface invoking the help function
  (when (and interface (capi:interface-message-area interface))
    (capi:execute-with-interface interface 'ywi:message-area-format interface
                                 (if (eq function 'viewer-open)
                                     #L"Opening help book page ~a..."
                                     #L"Executing help function ~s...")
                                 (if (eq function 'viewer-open)
                                     (first args)
                                     function)))
  (mp:mailbox-send (ensure-viewer-mailbox) (list* interface function args)))

;;; LispWorks seems to disconnect every conversation on quitting itself.
;;; BAD: When dde-connect hangs, the image tries to kill the runner proccess, sets its
;;;      priority to much, but hangs with CPU activity close to 100%

#|(defun %dde-disconnect-viewer ()
  (flet ((%disconnect (&rest dummy)
           (declare (ignore dummy))
           (when (and *viewer-conversation*
                      (integerp (win32::dde-conversation-hconv *viewer-conversation*)))
             (win32:dde-disconnect *viewer-conversation*))
           (setq *viewer-conversation* nil)))
    (%disconnect)					; global conversation
    (when (and %viewer-mailbox%				; dedicated process binding
               (mp:mailbox-reader-process %viewer-mailbox%)
               (mp:process-alive-p (mp:mailbox-reader-process %viewer-mailbox%)))
      (mp:mailbox-send %viewer-mailbox% (list nil #'%disconnect)))))

(lw:define-action "When quitting image" "Disconnect help DDE conversations"
  '%dde-disconnect-viewer)|#
